;;; database.el --- EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>
;; Keywords: EDB, database, forms
;; Version: 1.25
;; Release-Date: 2005-01-24

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; EDB is a flexible, customizable database program for Emacs.

;;; Code:

(defconst edb-version "1.25")
(defconst edb-date "2005-01-24")

(defun edb-version ()
  "Return a string describing the version of EDB that is running.
When called interactively, display the string in the echo area."
  (interactive)
  (let ((v (format "EDB %s of %s" edb-version edb-date)))
    (when (interactive-p)
      (message v))
    v))


(eval-when-compile (require 'cl))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Global variables
;;;

(defvar db-databases nil
  "Assoc list of database names and databases.")

(defvar db-displaytypes (make-hash-table)
  "Hash associating type names (symbols) and displayspecs.")

(defvar db-recordfieldtypes (make-hash-table)
  "Hash associating type names (symbols) and recordfieldspecs.")

(defvar db-inform-interval 100
  "When doing a lengthy computation, inform the user of progress every this
many records.  If nil, don't inform.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Database messages
;;;

(defun db-message (fmtstr &rest args)
  ;; Format message, display it, and log it in buffer *Database-Log*.
  (let ((formatted (apply 'format fmtstr args)))
    (db-in-buffer "*Database-Log*"
      (save-excursion
        (goto-char (point-max))
        (insert formatted "\n")))
    (db-best-fit-message formatted)))

(defmacro db-warning (fmtstr &rest args)
  ;; Like `db-message', but prepends "Warning: ".
  `(db-message (concat "Warning: " ,fmtstr) ,@args))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Loading
;;;

;; How to add a file to EDB...
;; * add its name to one of these lists
;; * add autoloads below, if necessary
;; * add a one-line description to the HACKING file

;;; Files comprising EDB.

(defconst edb-essential-file-names
  '("db-rep" "db-format" "db-file-io" "db-interfa"
    "db-types"
    ;; Strictly speaking db-summary.el is not necessary.  We include it,
    ;; however, to avoid an autoload during data-display-buffer setup, which
    ;; causes problems in the presence of `universal-coding-system-argument'.
    "db-summary")
  "List of EDB source files loaded by `load', sans extensions.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Autoloads
;;;

;; Declare some variables to be buffer-local here, so that they can
;; be set differently in different data display buffers even though the
;; packages that use them have not been loaded yet.

;;; db-convert.el
(autoload 'db-convert "db-convert"
  "Convert DATABASE's field structure.  To be autoloaded." t)

;;; db-rdb.el
(autoload 'db-rdb-setup "db-rdb"
  "Prepare EDB to read files in RDB format.  To be autoloaded." t)

;;; db-sort.el
(autoload 'database-sort "db-sort"
  "Sort and return DATABASE, which is also side-effected.  To be autoloaded." t)
(autoload 'database-sort-interface "db-sort")
(make-variable-buffer-local 'dbf-field-priorities)
(make-variable-buffer-local 'dbf-hidden-to-end-p)

;;; db-two-dbs.el
(autoload 'db-process-two-databases "db-two-dbs")
(autoload 'db-merge "db-two-dbs"
  "Merge two read-in databases.  To be autoloaded." t)
(autoload 'databases-compatible "db-two-dbs")

;;; db-search.el
(autoload 'db-parse-match-pattern "db-search") ; should be called first
(autoload 'db-print-match-pattern "db-search")
(autoload 'db-match "db-search")
(make-variable-buffer-local 'dbf-field-search-defaults)

;;; db-tagged.el
(autoload 'db-tagged-setup "db-tagged"
  "Prepare EDB to read files in tagged format.  To be autoloaded." t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Compiling
;;;

(defun edb-byte-compile-all ()
  (let ((dir (car load-path))
        (els (mapcar (lambda (file-name)
                       (concat file-name ".el"))
                     (append
                      ;; required
                      '("db-util")
                      ;; essential
                      edb-essential-file-names
                      ;; autoloaded
                      '("db-convert" "db-isbn" "db-rdb"
                        "db-search" "db-sort"
                        "db-tagged" "db-two-dbs")
                      ;; types
                      '("edb-t-human-names"
                        "edb-t-places-usuk"
                        "edb-t-timedate1")))))
    (unless (file-directory-p dir)
      (error "%S is not a directory." dir))
    (unless (file-exists-p (expand-file-name "db-util.el" dir))
      (error "Could not find db-util.el in %S" dir))

    (mapc 'load els)

    (let ((count 0)
          source dest fake)
      (dolist (file (cons "database.el" els))
        (setq source (expand-file-name file dir))
        (setq dest (concat (file-name-sans-versions source) "c"))

        ;; Compile unless a newer .elc file exists.
        (if (file-newer-than-file-p dest source)
            (push dest fake)
          (byte-compile-file source)
          (incf count)))

      ;; If any compilation happens, touch all the old .elc
      ;; files to fool -- aka "interoperate with" -- make(1).
      (unless (= 0 count)
        (apply 'call-process "touch" nil nil nil fake))

      (message "Done (Total of %d file%s compiled)"
               count (if (= count 1) "" "s")))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; One-time setup
;;;

;;; Actually load the database.
;; Perform requires
(require 'cl)
(require 'easymenu)
(condition-case nil
    (require 'db-util)
  (file-error
   (error "Can't find db-util on `load-path'.")))
;; Load EDB files.
(mapc 'load edb-essential-file-names)
(when (string-match "Lucid" emacs-version)
  (put 'emacs-version 'edb-running-lemacs t)
  (load "db-lemacs"))

(unless (assq 'dbc-hide-p minor-mode-alist)
  (push '(dbc-hide-p " Hide") minor-mode-alist))
(add-hook 'kill-buffer-hook 'db-kill-buffer-hook)

;; At the end of the file in case this load aborts.
(provide 'database)                     ; provide before running hooks

(defvar db-load-hooks nil
  "Function or list of functions run after loading EDB.
You can use this to load extensions, redefine EDB functions,
customize key bindings, etc.")

(run-hooks 'db-load-hooks)

;;; database.el ends here
