;;; db-convert.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Functions for changing the field structure of a database.

;; This takes a database and rearranges the field order, converts fields
;; from one type to another, adds or removes fields, etc.

;;; Code:


(eval-when-compile (require 'cl))
(provide 'db-convert)


(defvar db-computed-functions)          ; used dynamically


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Creation methods
;;;

;; To create a new database from an existing one, specify a creation method
;; for each field of the new database.  This can be done from a program by
;; supplying a fieldnames-creation list, which is a list of two-element lists
;; of (fieldname creation-method), or interactively.

;; A creation method (for field F, say) is one of the following:
;; * A field reference.  Field F is set to the value of the specified field
;;   in the corresponding old record.
;; * A literal value.  Field F of every record is set to that value.
;; * Result of a function call.  Specify the function, what arguments should
;;   be passed to it, and which of the results is desired.  Each argument
;;   specifier (one for each argument accepted by the function) specifies a
;;   field of the old record, the entire old record, or a literal.  The result
;;   specifier indicates that the function's result should be used as is or
;;   that the function returns a list, one element of which should be used as
;;   the value for field F.  This permits one function to do the computation
;;   for several fields in the new record; since the results are memoized, the
;;   function need only be called once per record.  For instance, suppose the
;;   original database stored addresses as a single field, but the new one
;;   stored street, city, state, and ZIP code separately.  The parsing of the
;;   address would only have to be done once.

;;; Represenation:
;; * Field reference: a number (the field number in the old record) or a
;;   symbol (the field name in the old database).
;; * Literal value: list of the symbol 'literal and the value; for
;;   strings, just the string itself.  Nil means the empty string.
;; * Function call: The function is a symbol whose function cell should be
;;   bound.  Each argument specifier is a fieldnumber (or -1, for the
;;   entire record) or fieldname (or 'original-record, for the entire
;;   record), or a literal as above.  The result specifier is nil (for the
;;   entire result) or an integer (in which case the value of interest is
;;   obtained via (nth result-number result).  The full specification is a
;;   list of two elements, the first of which is a list of the function and
;;   the argument specifiers and the second of which is the result number.
;;   Just the first element -- the list of the function and argument
;;   specifiers -- may be used if the result number is nil.

;;; Canonical representation:
;; The user is encouraged to use the fieldname representations, because
;; they are easier to read and write, but internally those are all
;; converted to the canonical representation, which is always the first one
;; listed in the descriptions above.  Thus, each creation method is a
;; number, a list whose car is 'literal, or a function specification.  This
;; permits the code which manipulates creation methods internally to be
;; more efficient.


;;;###autoload
(defun db-convert (database fcreate-specs)
  "Convert DATABASE's field structure according to FCREATE-SPECS.
This function rearranges the field order, converts fields from one type to
another, adds or removes fields, and so forth.  Note that this does not
create a new database; it destructively changes the format of DATABASE.

FCREATE-SPECS is a list of two-element lists of \(fieldnames
creation-method\); if creation-method would be nil, then fieldname alone
may be used in place of the two-element list."
  (let ((methods (mapcar (lambda (spec)
                           (db-canonicalize-creation-method
                            (if (symbolp spec)
                                ""
                              (car (cdr spec)))
                            database))
                         fcreate-specs))
        (n (length fcreate-specs)))
    (db-lmap-macro
     (link-set-record db-lmap-link
                      (db-convert-record (link-record db-lmap-link)
                                         methods n))
     database)
    ;; Clear the way for `database-set-fieldnames-to-list',
    ;; which does nothing if the `fieldnames' slot is already set.
    (setf (database-fieldnames database) nil)
    (database-set-fieldnames-to-list database (mapcar 'car fcreate-specs))))

;; Create and return a new record from RECORD using METHODS, a list
;; of canonical creation methods.  Optional argument NO-OF-FIELDS gives the
;; length of METHODS and the number of fields in the result record.
;; Dynamically binds db-computed-functions.
(defun db-convert-record (record methods &optional no-of-fields)
  (let ((new-record (make-vector (or no-of-fields (length methods)) nil))
        (this-field-no 0)
        db-computed-functions)
    (dolist (m methods)
      (aset new-record this-field-no
            (db-convert-compute-field-value m record))
      (incf this-field-no))
    new-record))

;; Uses and sets the "db-computed-functions" dynamic variable.
(defun db-convert-compute-field-value (m old-record)
  (cond ((numberp m)
         (aref old-record m))
        ((and (consp m) (eq (car m) 'literal))
         (car (cdr m)))
        ((consp m)
         (let* ((spec (car m))          ; function spec
                (result-no (car (cdr m)))
                (result-found (when result-no
                                (assoc spec db-computed-functions))))
           (if result-found
               (nth result-no (cdr result-found))
             (let* ((args (mapcar (lambda (a) ; arg spec
                                    (cond ((numberp a)
                                           (aref old-record a))
                                          ((and (consp a) (eq (car a) 'literal))
                                           (car (cdr a)))
                                          (t (error "Bad argument spec %s."
                                                    a))))
                                  (cdr spec)))
                    (result (apply (car spec) args)))
               (if result-no
                   (progn
                     (push (cons spec result) db-computed-functions)
                     (nth result-no result))
                 result)))))
        (t (error "Unknown creation method %s" m))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Canonicalization
;;;

(defun db-canonicalize-creation-method (m database)
  (cond
   ;; Literals
   ((null m)
    '(literal ""))
   ((stringp m)
    (list 'literal m))
   ((and (consp m) (eq (car m) 'literal))
    m)
   ;; Field references
   ((symbolp m)
    (or (fieldname->fieldnumber m database)
        (error "%s isn't a fieldname in database %s."
               m (database-print-name database))))
   ((numberp m)
    (if (and (> m 0)
             (< m (database-no-of-fields database)))
        m
      (error "%d isn't a valid field number in database %s."
             m (database-print-name database))))
   ;; Functions
   ((not (consp m))
    (error "Ill-formed creation method %s." m))
   (t
    (let ((spec (car m))                ; function spec
          function args result-no)
      (if (consp spec)
          (setq function (car spec)
                args (cdr spec)
                result-no (car (cdr m)))
        (setq function spec
              args (cdr m)
              result-no nil))
      (unless (fboundp function)
        (error "%s has no function definition." function))
      (setq args (mapcar (lambda (field-ref)
                           (if (numberp field-ref)
                               field-ref
                             (fieldname->fieldnumber field-ref database)))
                         args))
      (list (cons function args) result-no)))))

;;; db-convert.el ends here
