;;; db-file-io.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Read and write database files.

;;; Code:


(require 'pp)
(require 'db-rep)
(eval-when-compile (require 'cl))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DB variables
;;;

(defvar db-format-file-path nil
  "List of directories (strings) to search, in order, for format files not
found in the directory with their associated databases.")

(defvar db-aux-file-path nil
  "List of directories (strings) to search, in order, for auxiliary files not
found in the directory with their associated databases.")

(defvar db-io-error-p nil
  "Non-nil if there has been an error while reading or writing the database.")

(defvar db-before-read-hooks nil
  "Function or list of functions run immediately before a database is first read
but after all local variables are set.
The hooks are run in the data display buffer with variable `database' bound.
Variable `db-buffer' is bound to a buffer containing the database file.

This is a global variable.  If you set it to be specific to a particular
database \(for instance, in the format or auxiliary file\), then consider
having its last action be to reset the variable to nil.")

(defvar db-after-read-hooks nil
  "Function or list of functions run after a database is completely read.
The hooks are run in the data display buffer with variable `database' bound.
For databases with nonregular layouts, you might put a call to
`database-stored->actual' here, for instance.

This is a global variable.  If you set it to be specific to a particular
database \(for instance, in the format or auxiliary file), then consider
having its last action be to reset the variable to nil.")

(defvar db-format-file-suffixes '(".dbf" ".fmt" "f")
  "List of format file suffixes; the basename is that of the database file.
The suffixes are tried in order; the default is \(\".dbf\" \".fmt\" \"f\").
The . that may precede the extension must be specified explicitly.")

(defvar db-aux-file-suffixes '(".dba" ".aux" "a")
  "List of auxiliary file suffixes; the basename is that of the database file.
The suffixes are tried in order; the default is \(\".dba\" \".aux\" \"a\").
The . that may precede the extension must be specified explicitly.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Internal variables
;;;

;; These have docstrings so docstring-substitute will recognize them.

(defvar db-buffer nil "Buffer containing a database file being read.")

(defvar database nil "Database being read from a file; also other uses.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read a database file
;;;

(eval-when-compile (defvar edb-data-coding))

(defun db-read-database-file (db-file &optional format-file confirm)
  ;; Return the database.  The data display buffer will be the first element
  ;; of its data-display-buffers slot.
  ;;
  ;; Optional arg FORMAT-FILE, a filename, specifies the format file which
  ;; describes how database records are displayed.  If it is nil, the default
  ;; (in the default-format-file slot of the database) will be used, or one
  ;; returned by `db-file->format-file' will be used, or the user will be
  ;; prompted for one.
  ;;
  ;; If optional prefix arg CONFIRM is non-nil, then no default will be used
  ;; without confirmation from the user.
  (interactive "fDatabase file: \nP")
  (setq db-io-error-p nil)
  (let ((buf (generate-new-buffer "db-read-database-file"))
        ddb database coding)

    (set-buffer buf)
    (if (file-exists-p db-file)
        (insert-file-contents db-file nil)
      (message "New database file."))

    (setf database (or (db-read-database-internal-file-layout)
                       (make-database))
          (database-file database) db-file
          (database-modifiable-p database) (file-writable-p db-file))

    (when (or (not format-file) confirm)
      ;; We may be calling this because the argument to
      ;; db-read-database-file wasn't specified.
      (setq format-file (db-choose-format-file database format-file confirm)))

    (push (setq ddb (db-setup-data-display-buffer format-file database t))
          (database-data-display-buffers database))

    (db-in-buffer ddb
      ;; reread data if coding system is different
      (when (and (boundp 'edb-data-coding) edb-data-coding)
        (db-message "edb-data-coding: %s" edb-data-coding)
        (setq coding edb-data-coding
              buffer-file-coding-system coding)
        (with-current-buffer buf
          (unless (eq buffer-file-coding-system coding)
            (erase-buffer)
            (let ((coding-system-for-read coding))
              (insert-file-contents db-file nil))
            ;; hack! -- fixme: separate internal layout check/read, so we can
            ;;                 avoid ugliness that is the undo/redo.  --ttn
            (when (database-internal-file-layout-p database)
              (kill-buffer (pop (database-data-display-buffers database)))
              (setf database (db-read-database-internal-file-layout)
                    (database-file database) db-file
                    (database-modifiable-p database)
                    (file-writable-p db-file))
              (push (setq ddb (db-setup-data-display-buffer format-file database t))
                    (database-data-display-buffers database))
              (with-current-buffer ddb
                (setq buffer-file-coding-system coding)))))))

    (db-in-buffer ddb
      (let ((db-buffer buf))
        (run-hooks 'db-before-read-hooks)))

    (db-read-database-file-helper buf database)

    (db-in-buffer ddb
      (run-hooks 'db-after-read-hooks))

    database))


(defun db-read-database-internal-file-layout ()
  ;; If the buffer contains a database in internal file layout, read and return
  ;; the header portion (not the records).  Otherwise, return nil.
  (when (db-skip-string-forward ";; Database file written by EDB")
    (let ((here (point)))
      (emacs-lisp-mode)
      ;; Update old formats in small increments.
      (when (db-skip-string-forward "; format 0.1")
        (delete-backward-char 1)
        (insert "2")
        (forward-sexp)
        (backward-char 1)
        ;; add locals slot to database
        (insert " nil")
        (goto-char here))
      (when (db-skip-string-forward "; format 0.2")
        (delete-backward-char 1)
        (insert "3")
        (forward-sexp)
        (backward-char 1)
        (backward-sexp 1)
        (backward-char 1)
        ;; add modified-p slot to database
        (insert " nil")
        (goto-char here))
      (when (db-skip-string-forward "; format 0.3")
        (delete-backward-char 1)
        (insert "4")
        (down-list 1)
        (forward-sexp 16)
        ;; add internal-file-layout-p slot to database
        (insert " t")
        (forward-sexp 14)
        ;; add modifiable-p slot to database
        (insert " t")
        (goto-char here))
      (when (db-skip-string-forward "; format 0.4")
        (delete-backward-char 1)
        (insert "5")
        ;; change "[database" to "[cl-struct-database"
        (down-list 1)
        (insert "cl-struct-")
        (forward-sexp 17)
        ;; change "[sepinfo" to "[cl-struct-sepinfo"
        (down-list 1)
        (insert "cl-struct-")
        (up-list 1)
        (down-list 1)
        (insert "cl-struct-")
        (up-list 1)
        (down-list 1)
        (insert "cl-struct-")
        (goto-char here))
      (when (db-skip-string-forward "; format 0.5")
        (delete-backward-char 1)
        (insert "6")
        ;; Remove five slots for quotation information.
        (down-list 1)
        (forward-sexp 24)
        (kill-sexp 5)
        (goto-char here))

      ;; Don't forget to change write-database file if the format
      ;; number is updated.
      (unless (db-skip-string-forward "; format 0.6")
        (db-message "I don't know if I can read the database, but I'll try"))
      (end-of-line)
      (let ((rv (read (current-buffer))))
        (setf (database-internal-file-layout-p rv) t)
        rv))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Format file
;;;

(defun db-file->format-file (db-file)
  "Return a format file or nil."
  (db-locate-readable-file-prefer-cwd
   (file-name-nondirectory (file-name-sans-extension db-file))
   (cons (file-name-directory db-file)
         db-format-file-path)
   db-format-file-suffixes))


(defun db-locate-format-file (filename)
  ;; Return a full pathname for the format file named FILENAME, or err.
  (or (db-locate-readable-file-prefer-cwd
       (file-name-nondirectory filename)
       db-format-file-path)
      (error "I can't find a format file named %s." filename)))

(defun db-choose-format-file (database ff-default confirm)
  ;; Return a format file according to DATABASE or FF-DEFAULT.
  ;; Prompt if CONFIRM is set or if we can't get one from DATABASE alone.
  (let* ((db-file (database-file database))
         (default (or ff-default
                      (database-default-format-file database)
                      (db-file->format-file db-file))))
    (if (and default (not confirm))
        (let* ((dir (file-name-directory db-file))
               (default-directory (if dir
                                      (expand-file-name dir)
                                    default-directory)))
          (db-locate-format-file default))
      ;; Don't need `db-locate-format-file' because MUSTMATCH arg
      ;; to `read-file-name' is t.
      (expand-file-name
       (read-file-name
        (if default
            (format "Display format for %s: [default %s] "
                    (file-name-nondirectory db-file)
                    (file-name-nondirectory default))
          (format "Display format for %s: "
                  (file-name-nondirectory db-file)))
        (file-name-directory db-file)
        (when default
          (expand-file-name default (file-name-directory db-file)))
        t)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read database file:  helper functions
;;;

(defun db-read-database-file-helper (buf database)
  ;; BUF is a buffer containing the database file; DATABASE is the
  ;; database to be read into.
  ;; The local variables section of BUF is not executed, nor are the
  ;; database's auxiliary file read, nor are any database slots filled
  ;; (except for first-link).
  (db-message "Reading database...")
  (set-buffer buf)
  ;; The format and the auxiliary file should already be loaded; ie, all
  ;; special variables should be set.
  (if (database-internal-file-layout-p database)
      (let ((ls (read (current-buffer))))
        (db-message "Linking records...")
        (database-set-links-from-list database
          (mapcar 'db-new-link ls)))

    (database-io-setup database t)
    (database-delete-pre-and-post database)

    (if (database-read-record-from-region database)
        (db-read-file-custom database)
      (db-read-file-delimited database)))
  (kill-buffer nil)
  (push database db-databases)
  (db-message "Reading database...done (%d records)"
              (database-no-of-records database))
  database)


(defun database-set-links-from-list (database ls)
  ;; Take a list of links, set `database-first-link' to the first one and
  ;; set `link-next' and `link-prev' of each such that all the links are
  ;; arranged in a doubly-linked list in the order of the list.
  (setf (database-no-of-records database) (length ls))
  (let ((prev-link (car ls)))
    (setf (database-first-link database) prev-link
          ls (cdr ls))
    (dolist (this-link ls)
      (db-join-links prev-link this-link)
      (setq prev-link this-link))
    (db-join-links prev-link (database-first-link database))))


(defmacro db-reading-noninternal (condition &rest body)
  (declare (indent 1) (debug body))     ; fixme: what about condition? --ttn

  `(let* (this-record                   ; binding used by BODY
          this-link
          last-link
          first-link)

     (setq dbc-index 0)
     (while ,condition
       (incf dbc-index)
       ;; Noisy, obscures warning messages.
       (if (and db-inform-interval
                (zerop (% dbc-index db-inform-interval)))
           (db-message "Reading database...%d" dbc-index))

       ,@body                           ; uses binding `database'

       (setq this-link (db-new-link this-record))
       (unless first-link
         (setq first-link this-link
               last-link this-link))
       (db-join-links last-link this-link)
       (setq last-link this-link))

     (when first-link
       (db-join-links last-link first-link))
     (setf (database-first-link database) first-link
           (database-no-of-records database) dbc-index)))


;; fixme: integrate and refactor db-read-file-* funcs. --ttn

(defun db-read-file-custom (database)
  (let ((rrfunc (database-read-record-from-region database))
        (rsi (database-record-sepinfo database)))
    ;; Uses `database-read-record-from-region' and `sepinfo-sepfunc' from
    ;; `database-record-sepinfo'.
    ;; Check that the sepinfo has enough information.
    (if (sepinfo-sep-regexp rsi)
        (unless (sepinfo-sep-regexp-submatch rsi)
          (error "Need a submatch to go with sep-regexp `%s' of record sepinfo."
                 (sepinfo-sep-regexp rsi)))
      (unless (sepinfo-sep-function rsi)
        (let ((sep (sepinfo-sep-string rsi)))
          (when (or (not sep)
                    (equal "" sep))
            (error "You haven't specified a string, regexp, or function in record sepinfo."))
          (db-jam-si sep rsi (sepinfo-sep-string rsi) database))))

    (if (sepinfo-sep-function rsi)
        ;; Use sep-function to delimit the records

        ;; Return a pair of (end-pos . start-pos).
        (let ((delim (sepinfo-sep-function rsi))
              ;; Read-record-from-region might do insertions or deletions,
              ;; so just remembering the position isn't enough.
              next-start-marker
              (prev-end-pos 'nil)
              (end-start-pair '(nil)))
          (goto-char (point-min))
          (setq next-start-marker (point-marker))

          ;; Assume that there is at least one record in the file.
          ;; fixme: validate first. --ttn
          (db-reading-noninternal (marker-position next-start-marker)

            (setq end-start-pair (funcall delim prev-end-pos))
            (narrow-to-region (marker-position next-start-marker)
                              (car end-start-pair))
            ;; Writers of record->region often forget to do this.
            (goto-char (point-min))
            (setq this-record (funcall rrfunc))
            ;; Not `(car end-start-pair)' in case buffer was modified.
            (setq prev-end-pos (point-max))
            (widen)
            ;; In 18.55, `set-marker' can only make markers point to the
            ;; accessible portion of the buffer, so do this after widening.
            ;; (This may have changed by 18.59.)
            ;; Don't use just `(cdr end-start-pair)' because the buffer may
            ;; have been modified.
            (set-marker next-start-marker (and (cdr end-start-pair)
                                               (+ (- (cdr end-start-pair)
                                                     (car end-start-pair))
                                                  prev-end-pos)))))
      ;; Use regexp to delimit the records.
      (let ((rsep-rgx (sepinfo-sep-regexp rsi))
            (rsep-rgx-subm (sepinfo-sep-regexp-submatch
                            rsi))
            (next-start (point-min))
            ;; How many characters to skip before looking for the start of the
            ;; next record.  We can't use a position because rrfunc may insert
            ;; or delete, and we can't use markers because `narrow-to-region'
            ;; squeezes them.
            sep-len)
        ;; The caller also called `database-delete-pre-and-post'.
        (goto-char (point-min))

        (db-reading-noninternal (< next-start (point-max))

          (goto-char next-start)
          ;; re-search-forward errs if it fails
          (re-search-forward rsep-rgx)
          (setq sep-len (- (match-end rsep-rgx-subm)
                           (match-beginning rsep-rgx-subm)))
          (narrow-to-region next-start (match-beginning rsep-rgx-subm))
          (setq this-record (funcall rrfunc))
          (setq next-start (+ (point-max) sep-len))
          (widen))))))



(defun db-read-file-delimited (database)
  ;; If there are any regexps set in the sepinfos at this point, then we
  ;; assume that the programmer specified them explicitly; we assume that if
  ;; any substitutions are requested, then the programmer knows that they
  ;; won't cause any ambiguities.

  (goto-char (point-min))

  ;; We have now cleared away the pre- and post- garbage.
  (if (or (sepinfo-sep-regexp (database-field-sepinfo database))
          (sepinfo-sep-regexp (database-record-sepinfo database)))
      ;; There are regexps involved, so do no substitution on the separators,
      ;; except that the user has explicitly requested; then read the database.
      (progn
        (database-perform-substitutions database t)
        (db-read-file-delimrx database))

    ;; There are no regexps involved.
    (let* ((fsep-str (database-full-fieldsep-string database))
           (rsep-str (database-full-recordsep-string database))
           confirmation-list)

      ;; Convert the database buffer from "complex" to "simple" form.
      (let* ((fsi (database-field-sepinfo database))
             (pre-str (sepinfo-pre-first-string fsi))
             (pre-rgx (sepinfo-pre-first-regexp fsi)))
        ;; `database-delete-pre-and-post' only did the record separators;
        ;; here we do the field separators as well.
        (goto-char (point-min))
        (cond (pre-rgx
               (if (db-skip-regexp-forward pre-rgx)
                   (delete-region (point-min) (point))
                 (error "Didn't find match for regexp `%s' leading the first field."
                        pre-rgx)))
              (pre-str
               (if (db-skip-string-forward pre-str)
                   (delete-region (point-min) (point))
                 (error "Didn't find string `%s' leading the first field."
                        pre-str))))

        (goto-char (point-max))
        ;; Don't get rid of post-last-field-regexp or post-last-field-string
        ;; because we look for them at the end of every record.

        ;; Make sure that record-sepinfo-sep-string appears at the end.
        (when (sepinfo-sep-string (database-record-sepinfo database))
          (if (db-skip-string-backward
               (sepinfo-sep-string (database-record-sepinfo database)))
              (goto-char (point-max))
            (insert (sepinfo-sep-string (database-record-sepinfo database)))))
        (when pre-str
          (insert pre-str)))

      ;; We're still inside the big let.

      ;; This pre-read confirmation should be optional.  And it should be
      ;; able to deal with regexps.
      ;; When would this test fail?  Won't fsep-str and
      ;; rsep-str always be non-nil when we get here?
      (when (and fsep-str rsep-str)
        (db-message "Confirming database format...")
        (setq confirmation-list (database-confirm-fieldsep-and-recordsep
                                 fsep-str rsep-str
                                 (database-no-of-fields database) nil))
        (if (or (db-confls-fieldsep-bad-p confirmation-list)
                (db-confls-recordsep-bad-p confirmation-list))
            (progn
              (db-warning "The database file is malformed!")
              (when (db-confls-fieldsep-bad-p confirmation-list)
                (db-warning "Extra field separator `%s' found in data."
                            fsep-str))
              (when (db-confls-recordsep-bad-p confirmation-list)
                (db-warning "Extra record separator `%s' found in data."
                            rsep-str))
              ;; show the db warning buffer
              (if (yes-or-no-p "Database file is improperly formatted; try to read it anyway? ")
                  (db-message "Damaged database being read; expecting approximately %d records"
                              (db-confls-no-of-records confirmation-list))
                (kill-buffer nil)
                (error "Aborted attempt to read database.")))
          (db-message "Database looks OK from here; expecting %d records"
                      (db-confls-no-of-records confirmation-list))))

      ;; This recomputes the full-fieldsep and full-recordsep; oh well.
      ;; It also sets the io-sep variables.
      (database-substitute-for-read database)

      (db-read-file-delimstr database))))


(defun db-read-file-delimstr (database)
  ;; When we call this, the database is in the following form:
  ;; Point at start of first field of first record.
  ;; Each field, except last, is ended by actual-fieldsep.
  ;; Each record, including last, is ended by actual-recordsep.
  ;; End of last recordsep-string = eob.
  ;; Field-sep and record-sep must be strings.
  (db-message "Reading database...")

  (goto-char (point-min))
  (let* ((fsep (database-sub-fieldsep-string database))
         (rsep (database-sub-recordsep-string database))
         (flen (length fsep))
         (rlen (length rsep))
         (nfields (database-no-of-fields database))
         (max-fno (1- nfields))
         (here (point))
         fno
         end-of-rsep
         end-of-record)

    (db-reading-noninternal (not (or (eobp)
                                     ;; Does this cause any problems?
                                     ;; Special case for rsep = "\n\n",
                                     ;; extra newline at end
                                     (and (string= rsep "\n\n")
                                          (looking-at "\n\\'"))))
      (setq this-record (make-vector nfields nil))
      (if (search-forward rsep nil t)
          (setq end-of-rsep (point)
                end-of-record (- (point) rlen))
        (db-warning "Didn't find `%s' at end of last field of record %d, and I put it there myself!"
                    rsep dbc-index)
        (setq end-of-rsep (point-max)
              end-of-record (point-max)))
      (goto-char here)

      (setq fno 0)
      (while (< fno max-fno)
        ;; fixme: trap errors, maybe check for rsep in field data.  --ttn
        (if (search-forward fsep end-of-record t)
            (progn
              (record-set-field-from-index
               this-record fno (buffer-substring here (- (point) flen))
               nil)
              (setq here (point))
              (incf fno))

          (db-warning "Hit the end of the record after %d fields of record %d (didn't find fsep `%s')"
                      fno dbc-index fsep)
          (record-set-field-from-index
           this-record fno (buffer-substring here end-of-record) nil)
          (setq here end-of-record)
          (incf fno)
          (while (<= fno max-fno)
            (record-set-field-from-index
             this-record fno (recordfieldspec-default-value
                              (database-recordfieldspec database fno))
             nil)
            (incf fno))))
      ;; If there weren't too few fields, set the last one (else it's already set).
      (when (= fno max-fno)
        (when (search-forward fsep end-of-record t)
          (db-warning "Extra fields in record %d packed into the last field; beware when writing."
                      dbc-index)
          (setq db-io-error-p t))
        (record-set-field-from-index this-record max-fno
                                     (buffer-substring here end-of-record)
                                     nil))
      (goto-char end-of-rsep)
      (setq here (point)))

    ;; Convert from stored to actual format.
    (database-stored->actual database)

    ;; This function is called for side-effect, but return the database anyway.
    database))


(defun db-read-file-delimrx (database)

  (db-message "Reading database...")

  (goto-char (point-min))

  (let* ((fsi (database-field-sepinfo database))
         (rsi (database-record-sepinfo database))
         (frx (or (sepinfo-sep-regexp fsi)
                  (regexp-quote (sepinfo-sep-string
                                 (database-field-sepinfo database)))))
         (frx-subm (or (sepinfo-sep-regexp-submatch fsi)
                       0))
         (rrx (or (sepinfo-sep-regexp rsi)
                  (regexp-quote (sepinfo-sep-string rsi))))
         (rrx-subm (or (sepinfo-sep-regexp-submatch rsi)
                       0))

         ;; Do not fold into `rrx'; they may have submatches of their own.
         (pre-str (sepinfo-pre-first-string fsi))
         (pre-frx (or (sepinfo-pre-first-regexp fsi)
                      (and pre-str
                           (regexp-quote pre-str))))
         (pre-frx-subm (or (sepinfo-pre-first-regexp-submatch
                            fsi)
                           0))
         (post-str (sepinfo-post-last-string fsi))
         (post-frx (or (sepinfo-post-last-regexp fsi)
                       (and post-str
                            (regexp-quote post-str))))
         (post-frx-subm (or (sepinfo-post-last-regexp-submatch
                             fsi)
                            0))

         (nfields (database-no-of-fields database))
         (max-fno (1- nfields))
         (here (point))
         fno
         end-of-rrx
         end-of-record)

    (db-reading-noninternal (not (eobp))
      (setq this-record (make-vector nfields nil))
      (when pre-frx
        (if (db-skip-regexp-forward pre-frx)
            (progn
              (setq here (match-end pre-frx-subm))
              (goto-char here))
          (error "Didn't find pre-first stuff I expected.")))

      (setq fno 0)
      (while (< fno max-fno)
        ;; fixme: trap errors, maybe check for rrx in field data.  --ttn
        (if (re-search-forward frx nil t)
            (progn
              (record-set-field-from-index
               this-record fno
               (buffer-substring here (match-beginning frx-subm))
               nil)
              (setq here (match-end frx-subm))
              (incf fno))
          (db-warning "Hit the end of the database after %d fields of record %d (didn't find frx `%s')"
                      fno dbc-index frx)
          (setq fno max-fno)))
      (if (re-search-forward rrx nil t)
          (setq end-of-rrx (match-end rrx-subm)
                end-of-record (match-beginning rrx-subm))
        (db-warning "Didn't find `%s' at end of last field of record %d."
                    rrx dbc-index)
        (setq end-of-rrx (point-max)
              end-of-record (point-max)))
      (goto-char here)
      (when (re-search-forward frx end-of-record t)
        (db-warning "Too many fields in record %d; packing them all into the last field; beware when writing."
                    dbc-index)
        (setq db-io-error-p t))
      (record-set-field-from-index this-record max-fno
                                   (buffer-substring here end-of-record)
                                   nil)
      (goto-char end-of-rrx)
      (setq here (point)))

    (database-stored->actual database)

    database))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read database utilities
;;;


(defun database-stored->actual (&optional db)
  ;; Convert string values in a newly-read database to the actual format.  If
  ;; DB is not specified, use the value of the dynamic variable `database'.
  ;; This makes it possible to be put directly in `db-after-read-hook'.
  (unless db (setq db database))
  (let (fno-s->a-alist
        s->a
        (field-no 0)
        field-val
        (no-of-fields (database-no-of-fields db)))
    (while (< field-no no-of-fields)
      (setq s->a (recordfieldspec-stored->actual
                  (database-recordfieldspec db field-no)))
      (when s->a
        (push (cons field-no s->a) fno-s->a-alist))
      (incf field-no))
    (when fno-s->a-alist
      (db-message "Converting from stored record format...")
      (maprecords (lambda (record)
                    (dolist (s->a-pair fno-s->a-alist)
                      (setq field-no (car s->a-pair)
                            field-val (aref record field-no)
                            s->a (cdr s->a-pair))
                      (record-set-field-from-index
                       record field-no (if (stringp field-val)
                                           (funcall s->a field-val)
                                         field-val)
                       nil)))
                  db nil "Converting record format...%s")
      (db-message "Converting record format...done"))))

(defun database-delete-pre-and-post (database)
  ;; Modify current buffer according to DATABASE.
  ;; Remove pre-first-record- and post-last-record strings or regexps, then
  ;; add a field separator to the end so that every record is terminated by one.
  (let ((rsep (database-record-sepinfo database)))
    ;; This only does the RECORD separators.

    ;; Remove gubbish before first record.
    (goto-char (point-min))
    (let ((pre-str (sepinfo-pre-first-string rsep))
          (pre-rgx (sepinfo-pre-first-regexp rsep)))
      (cond (pre-rgx
             (if (db-skip-regexp-forward pre-rgx)
                 (progn
                   (goto-char (match-end (sepinfo-pre-first-regexp-submatch
                                          rsep)))
                   (delete-region (point-min) (point)))
               (error "Didn't find match for regexp `%s' leading the data."
                      pre-str)))
            (pre-str
             (if (db-skip-string-forward pre-str)
                 (delete-region (point-min) (point))
               (error "Didn't find string `%s' leading the data." pre-str)))))

    ;; Remove gubbish after last record.
    (goto-char (point-max))
    (let ((post-str (sepinfo-post-last-string rsep))
          (post-rgx (sepinfo-post-last-regexp rsep)))
      (cond (post-rgx
             (if (re-search-backward post-rgx)
                 (progn
                   (goto-char (match-beginning (sepinfo-post-last-regexp-submatch
                                                rsep)))
                   (delete-region (point) (point-max)))
               (error "Didn't find match for post-rgx `%s'."
                      post-str)))
            (post-str
             (if (search-backward post-str)
                 (delete-region (point) (point-max))
               (error "Didn't find post-str `%s'." post-str)))))
    ;; If there isn't one there already, add a final record separator so that
    ;; every record is terminated by one.
    ;; fixme: move to caller? --ttn
    (let ((sep (sepinfo-sep-string rsep)))
      (unless (db-skip-string-backward sep)
        (insert sep)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write database file
;;;


(defun db-write-1 (coding database &optional ofile)
  ;; This is called by `db-write-database-file' with the data display buffer
  ;; current, and we need the values of variables local to that buffer.
  (setq db-io-error-p nil)
  (if ofile
      (if (or (not (file-name-directory ofile))
              (file-exists-p (file-name-directory ofile)))
          (setf (database-file database) ofile)
        (error "Can't write %s: directory does not exist." ofile))
    (setq ofile (database-file database)))

  (db-in-buffer (generate-new-buffer "db-write-1")
    (buffer-disable-undo (current-buffer))
    (auto-save-mode 0)
    (if (database-internal-file-layout-p database)
        (let ((fl (database-first-link database))
              (ddbs (database-data-display-buffers database))
              (modified-p (database-modified-p database))
              (standard-output (current-buffer)))
          (setq buffer-file-coding-system coding)
          (insert ";; Database file written by EDB; format 0.6")
          (let ((coding-string (symbol-name coding)))
            (unless (string-match "^undecided-" coding-string)
              (insert " -*- coding: " coding-string "; -*-")))
          (insert "\n")
          (unwind-protect
              (progn
                (setf (database-first-link database) nil
                      (database-data-display-buffers database) nil)
                (database-set-modified-p database nil)
                (pp database))
            (setf (database-first-link database) fl
                  (database-data-display-buffers database) ddbs)
            ;; in case we abort out of the write
            (database-set-modified-p database modified-p))
          (insert "(\n")
          (maprecords 'pp database)
          (insert ") ;; thank you for using EDB\n")
          ;; This is not catching the error raised by basic-save-buffer if
          ;; the destination is not writable.
          (condition-case error
              (let ((require-final-newline nil)
                    ;; get around write-file
                    (auto-save-default nil))
                (write-file ofile)
                (database-set-modified-p database nil)
                (kill-buffer nil))
            ;; Used to be `file-error', which didn't catch the error raised
            ;; by `basic-save-buffer' if the destination is not writable.
            (error
             (setq db-io-error-p t)
             ;; This must come before the buffer is killed.
             (db-warning "Error `%s' while writing buffer %s to file %s."
                         error (buffer-name (current-buffer)) ofile)
             ;; Avoid further questions about killing the buffer.
             (set-buffer-modified-p nil)
             (kill-buffer nil))))
      ;; Don't use internal representation.
      ;;
      ;; Fine example of use of undocumented functionality
      ;; (`this-buffer' is bound by db-in-buffer).
      (db-copy-buffer-local-variables this-buffer)
      ;; We don't want to be *exactly* like the data display buffer.
      (setq major-mode 'write-database-mode
            buffer-read-only nil)
      (database-io-setup database)
      (let ((first-record t)
            (rsep (database-record-sepinfo database)))
        (when (sepinfo-pre-first-string rsep)
          (insert (sepinfo-pre-first-string rsep)))
        (if (database-write-region-from-record database)
            ;; Don't check separators for write-record-function, even for
            ;; recordsep; read-record-function's cleverness is unknown albeit
            ;; not unknowable.  Also, don't do substitution or quoting.
            (let ((record-sep-string (sepinfo-sep-string rsep))
                  (write-region-fn (database-write-region-from-record
                                    database)))
              (db-message "Writing database...")
              (maprecords (lambda (record)
                            (if first-record
                                (setq first-record nil)
                              (insert record-sep-string))
                            (funcall write-region-fn record))
                          database nil "Writing database...%d")
              (db-message "Writing database to disk..."))
          (db-write-intdelim database))
        (when (sepinfo-post-last-string rsep)
          (insert (sepinfo-post-last-string rsep)))

        ;; Add the local variables, if any, to the end.
        (when (database-file-local-variables database)
          (insert (database-file-local-variables database)))

        (condition-case error
            (let ((require-final-newline nil)
                  ;; get around `write-file'
                  (auto-save-default nil))
              (write-file ofile)
              (database-set-modified-p database nil)
              (kill-buffer nil))
          ;; Used to be `file-error', which didn't catch the error raised
          ;; by `basic-save-buffer' if the destination is not writable.
          (error
           (setq db-io-error-p t)
           ;; This must come before the buffer is killed.
           (db-warning "Error `%s' while writing buffer %s to file %s."
                       error (buffer-name (current-buffer)) ofile)
           ;; Did write-file
           (auto-save-mode 0)
           ;; Avoid further questions about killing the buffer.
           (set-buffer-modified-p nil)
           (kill-buffer nil))))))
  ;; Update.
  (when (or (db-data-display-buffer-p)
            (db-summary-buffer-p))
    (dbc-update-database-modified-p)
    (force-mode-line-update)))


(defun db-write-intdelim (database)
  ;; Insert the records, fields separated by fieldsep and records separated by
  ;; recordsep, into the current buffer; it uses delimited format.  This is
  ;; called by `db-write-1', which arranges `unwind-protect' boundaries, calls
  ;; the -internal function that uses the proper output file layout, etc.
  (let* ((previous-point (point))
         (fno 0)
         (first-record t)
         (no-of-fields (database-no-of-fields database))
         ;; wrfunc unused --ttn
         ;;(wrfunc (database-write-region-from-record database))
         (fsep (database-full-fieldsep-string database))
         (rsep (database-full-recordsep-string database))
         (sub-fsep (or (database-sub-fieldsep-string database) fsep))
         (sub-rsep (or (database-sub-recordsep-string database) rsep))
         confirmation-list)

    ;; Check the delimiters.
    (unless (database-acceptable-delimiter-p sub-fsep)
      (setq sub-fsep (database-generate-delimiter nil)))
    (unless (database-acceptable-delimiter-p sub-rsep)
      (setq sub-rsep (database-generate-delimiter nil)))

    (db-message "Writing database...")
    (maprecords
     (lambda (record)
       (setq fno 0)
       (while (< fno no-of-fields)
         (when (> fno 0) (insert sub-fsep))
         ;; This is `record-field-stored', inlined.
         (insert (db-funcall-maybe (recordfieldspec-actual->stored
                                    (database-recordfieldspec database fno))
                                   (aref record fno)))
         (incf fno))
       (insert sub-rsep))
     database nil "Writing database...%d")
    (db-message "Writing database...confirming")

    (narrow-to-region previous-point (point-max))
    (setq confirmation-list (database-confirm-fieldsep-and-recordsep
                             sub-fsep sub-rsep
                             (database-no-of-fields database)
                             (database-no-of-records database)))
    (if (or (db-confls-fieldsep-bad-p confirmation-list)
            (db-confls-recordsep-bad-p confirmation-list))
        (progn
          (when (db-confls-fieldsep-bad-p confirmation-list)
            (db-warning "Tripped over an unexpected field separator `%s' in data; trying again."
                        sub-fsep)
            (setf (database-sub-fieldsep-string database)
                  (database-generate-delimiter database t)))
          (when (db-confls-recordsep-bad-p confirmation-list)
            (db-warning "Tripped over an unexpected record separator `%s' in data; trying again."
                        sub-rsep)
            (setf (database-sub-recordsep-string database)
                  (database-generate-delimiter database t)))

          ;; We've chosen new separators; erase the work so far.
          (delete-region previous-point (point))

          ;; Call this function recursively.
          (db-write-intdelim database)
          ;; I've called this recursively; don't do any substitution
          ;; or quoting.
          )
      ;; Confirmation was OK:  correct number of field and record
      ;; separators found.

      ;; Put off adding the pre- and post- field strings until
      ;; after checking separators, as they may contain anomolous
      ;; field separators, for instance.

      ;; But do it before substitution so that all field pre- and
      ;; post- strings are treated identically.

      ;; The whole point of using io-separators is so they
      ;; appear exactly as the user specified, unaffected by
      ;; substitution.


      ;; But note that pre- and post- record strings will be added
      ;; later, after substitution.

      (database-substitute-for-write database
                                     fsep sub-fsep
                                     rsep sub-rsep)
      ;; Convert from "simple" to "complex" form.
      (goto-char (point-min))
      (let ((pre-first (sepinfo-pre-first-string (database-field-sepinfo database))))
        (when pre-first
          (insert pre-first)))

      (goto-char (point-max))
      (if (and (db-skip-string-backward (or (sepinfo-pre-first-string
                                             (database-field-sepinfo database))
                                            ""))
               (db-skip-string-backward (or (sepinfo-sep-string
                                             (database-record-sepinfo database))
                                            "")))
          (delete-region (point) (point-max))
        (error "Didn't find expected trailing junk `%s' or `%s'."
               (sepinfo-pre-first-string (database-field-sepinfo database))
               (sepinfo-sep-string (database-record-sepinfo database)))))
    (widen)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; I/O utilities
;;;

(defun database-io-setup (database &optional norx)
  ;; Some of this information may already be correctly set (especially if
  ;; we're now writing), but just in case some of the database slots have
  ;; changed since reading.

  ;; When converting strings to regexps, must be careful to watch out for
  ;; substitution and quotation; don't get fooled.

  ;; When writing to the file, we take the previous version's local
  ;; variables section verbatim.

  ;; We're only setting the regexp variables.

  ;; When reading, we'll get the variables from the database, auxiliary,
  ;; and format files anew each time anyway.

  (database-check-sepinfo (database-record-sepinfo database) database
                          "record" "\n" norx)
  (database-check-sepinfo (database-field-sepinfo database) database
                          "field" "\t" norx)
  (database-check-sepinfo (database-alternative-sepinfo database) database
                          "alternative" t norx)

  (setq db-substitution-no-no-string
        (apply 'concat
               (mapcar (lambda (pair)
                         (concat (car pair) (cdr pair)))
                       (database-substitutions database)))))


(defun database-check-sepinfo (si database name &optional sdef norx)
  ;; It's good that this setting of -regexp slots doesn't happen until the
  ;; last possible moment, when quotation-char and other variables that these
  ;; values depend on are already set to their final values.

  ;; If the sep-string slot of the sepinfo is nil, then:
  ;;  * if SDEF is nil, signal an error.
  ;;  * if SDEF is t, do nothing.
  ;;  * otherwise set the slot to SDEF.

  ;; If the string is the empty string, then we must have just set it that
  ;; way, which means that we either also just set the regexp to nil, or we
  ;; set the regexp to something we care about.  In either case don't mess
  ;; further with the regexp.

  ;; The submatches could default to 0 if they're nil; but I want to be
  ;; more paranoid than that.
  (when (and (sepinfo-pre-first-string si)
             (equal "" (sepinfo-pre-first-string si)))
    ;; pre-first-string = ""
    (setf (sepinfo-pre-first-string si) nil))
  (if (sepinfo-pre-first-regexp si)
      (unless (sepinfo-pre-first-regexp-submatch si)
        (error "Need a submatch to go with pre-first-regexp `%s' of %s."
               (sepinfo-pre-first-regexp si) name))
    (when (and (sepinfo-pre-first-string si)
               (not norx))
      (db-jam-si pre-first si (sepinfo-pre-first-string si) database)))

  ;; Don't test for sep-function because the sepinfo must be valid for
  ;; output as well as input.
  ;; On the other hand, we don't need sep-string to be set if a wrfr
  ;; function is in use, but this function doesn't do any such checks.
  (when (or (not (sepinfo-sep-string si))
            (and (equal "" (sepinfo-sep-string si))
                 (not (sepinfo-sep-regexp si))))
    ;; sep-string isn't set appropriately; it contains no information
    (cond ((eq t sdef))                 ; do nothing
          ((not sdef) (error "Sep-string must be non-empty in %s." name))
          (t (setf (sepinfo-sep-string si) sdef))))
  (if (sepinfo-sep-regexp si)
      (unless (sepinfo-sep-regexp-submatch si)
        (error "Need a submatch to go with sep-regexp `%s' of %s."
               (sepinfo-sep-regexp si) name))
    (when (and (not (sepinfo-sep-function si))
               (not norx))
      (db-jam-si sep si (sepinfo-sep-string si) database)))

  (when (and (sepinfo-post-last-string si)
             (equal "" (sepinfo-post-last-string si)))
    (setf (sepinfo-post-last-string si) nil))
  (if (sepinfo-post-last-regexp si)
      (unless (sepinfo-post-last-regexp-submatch si)
        (error "Need a submatch to go with post-last-regexp `%s' of %s."
               (sepinfo-post-last-regexp si) name))
    (when (and (sepinfo-post-last-string si)
               (not norx))
      (db-jam-si post-last si (sepinfo-post-last-string si) database))))


(defmacro db-jam-si (afrag sepinfo string database)
  ;; AFRAG is a sepinfo accessor func name fragment (symbol).
  (let ((rx-acc (intern (format "sepinfo-%s-regexp" afrag)))
        (subrx-acc (intern (format "sepinfo-%s-regexp-submatch" afrag))))
    `(let ((str ,string)
           (db ,database)
           (si ,sepinfo))
       (cond ((or (null str) (equal "" str))
              (setf (,rx-acc si) nil)
              (setf (,subrx-acc si) nil))
             (t
              (setf (,rx-acc si) (regexp-quote str))
              (setf (,subrx-acc si) 0))))))

(defun db-rfspec<-rftype (type)
  "Return the recordfieldspec associated with symbol TYPE."
  (let ((rv (gethash type db-recordfieldtypes)))
    (when rv (if (symbolp rv)
                 (db-rfspec<-rftype rv)
               rv))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Substitution
;;;

;; When substitution is called, the rendered database is always in "simple"
;; format: rendered fields are separated by sub-fieldsep, and rendered records
;; (including the last one) are followed by sub-recordsep.

;; The "complicated" format includes pre-first-field at the beginning and
;; post-last-field (without record-sep-string or pre-first-field, which are
;; the other elements of sub-recordsep) at the end.

;; Simple format is nice because it requires no special-casing at the
;; beginning (because there's no gubbish there) or at the end (because
;; what's there is exactly what's between each pair of records).

;; The caller of database-substitute-for-* arranges to convert to/from
;; simple format before/after the call.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defun database-substitute-for-read (database)
  ;; Convert field/record separators (perhaps choosing new ones), so they
  ;; won't get damaged by the substitution, and then doing the substitution.
  ;; It also must set the sub-{field,record}sep slots, because later on those
  ;; fields are slavishly followed.  We can't parse or do substitutions if
  ;; either of the separators is a regexp.
  (let* ((fsep-orig (database-full-fieldsep-string database))
         (rsep-orig (database-full-recordsep-string database)))

    (if (or (sepinfo-sep-regexp (database-record-sepinfo database))
            (sepinfo-sep-regexp (database-field-sepinfo database)))
        (setf (database-sub-recordsep-string database)
              (or (sepinfo-sep-regexp (database-record-sepinfo database))
                  (database-sub-recordsep-string database)
                  rsep-orig)
              (database-sub-fieldsep-string database)
              (or (sepinfo-sep-regexp (database-field-sepinfo database))
                  (database-sub-fieldsep-string database)
                  fsep-orig))
      (unless rsep-orig
        (error "No record separator specified."))
      (if (database-acceptable-delimiter-p rsep-orig)
          (setf (database-sub-recordsep-string database) rsep-orig)
        (db-message "Substituting record delimiter for read...")
        (setf (database-sub-recordsep-string database)
              (database-generate-delimiter database))
        (goto-char (point-min))
        (let ((s (database-sub-recordsep-string database)))
          (while (search-forward rsep-orig nil t)
            (replace-match s)))
        (setq fsep-orig (replace-regexp-in-string
                         rsep-orig (database-sub-recordsep-string database)
                         fsep-orig))
        (db-message "Substituting record delimiter for read...done"))
      (if (or (database-acceptable-delimiter-p fsep-orig)
              (database-read-record-from-region database))
          (progn
            (db-message "fsep-orig `%s' acceptable because `(or %s %s)'"
                        (string-to-list fsep-orig)
                        (database-acceptable-delimiter-p fsep-orig)
                        (database-read-record-from-region database))
            (setf (database-sub-fieldsep-string database) fsep-orig))
        (unless fsep-orig
          (error "No field separator specified."))
        (db-message "Substituting field delimiter for read...")
        (setf (database-sub-fieldsep-string database)
              (database-generate-delimiter database))
        (db-message "Substituting field delimiter for read... (`%s' for `%s')"
                    (string-to-list (database-sub-fieldsep-string database))
                    (string-to-list fsep-orig))
        (goto-char (point-min))
        (let ((s (database-sub-fieldsep-string database)))
          (while (search-forward fsep-orig nil t)
            (replace-match s)))
        (db-message "Substituting field delimiter for read...done"))
      (db-message "sub-fieldsep: %s"
                  (string-to-list (database-sub-fieldsep-string database)))

      (database-perform-substitutions database t))))


(defun database-perform-substitutions (database backward)
  (when (database-substitutions database)
    (db-message "Substituting...")
    ;; Make replacements in the current buffer according to SUBS.
    ;; SUBS is list of pairs of strings; the cdr of each pair will be
    ;; substituted for the car, in order, unless optional argument BACKWARD is
    ;; non-nil, in which case the car is substituted for the cdr and the
    ;; substitutions are done in reverse order.
    ;;
    ;; Warn if any of the
    ;; substituted-in strings already appears in the buffer; such a
    ;; situation would make substitution, then unsubstitution, not yield
    ;; a result identical to the original buffer, since all instances of
    ;; the substituted-in string will be assumed on the reverse
    ;; substitution to have been the result of replacing a
    ;; substituted-for string.
    ;;
    ;; fixme: do all checking before any substitutions are done. --ttn
    (let ((subs (database-substitutions database))
          bef aft ambiguity ambiguities)
      (when backward
        (setq subs (mapcar (lambda (pair)
                             (cons (cdr pair) (car pair)))
                           (reverse subs))))
      (dolist (sub subs)
        (setq bef (car sub)
              aft (cdr sub))
        (goto-char (point-min))

        (when (search-forward aft nil t)
          (setq ambiguity sub)
          (goto-char (point-min)))

        (while (search-forward bef nil t)
          (replace-match aft))

        ;; Don't complain if we didn't actually do any substitution.
        (when ambiguity
          (unless (= (point) (point-min))
            (push ambiguity ambiguities))
          (setq ambiguity nil)))

      (when ambiguities
        (error "Ambiguities:  %s" ambiguities)))
    (db-message "Substituting...done")))


;; Database-substitute-for-write does the substitution, then, if field
;; separators were changed in order to prevent them from getting damaged by
;; the substitution, converts them back to the user-specified strings,
;; which might contain substrings that would have been substituted for in
;; the previous operation, had we not been careful.
(defun database-substitute-for-write (database new-fsep sub-fsep
                                               new-rsep sub-rsep)
  ;; Check that there are the correct number of fieldseps and recordseps
  ;; here; if wrong number, choose new fieldsep and/or recordsep and take
  ;; it from the top.

  (database-perform-substitutions database nil)

  (unless (equal sub-fsep new-fsep)
    (goto-char (point-min))
    (while (search-forward sub-fsep nil t)
      (replace-match new-fsep)))
  (unless (equal sub-rsep new-rsep)
    (goto-char (point-min))
    (while (search-forward sub-rsep nil t)
      (replace-match new-rsep)))

  ;; Now the buffer is ready to have the preceding and trailing junk
  ;; added and to be written to disk.
  )



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Substitution Utilities
;;;

;; A "confirmation list" -- aka "conflist" -- has the form:
;;   (FIELDSEP-BAD-P RECORDSEP-BAD-P NO-OF-RECORDS)

(defsubst db-confls-fieldsep-bad-p (conflist) (car conflist))
(defsubst db-confls-recordsep-bad-p (conflist) (car (cdr conflist)))
(defsubst db-confls-no-of-records (conflist) (car (cdr (cdr conflist))))

(defun database-confirm-fieldsep-and-recordsep (fsep rsep nfields nrecords)
  ;; Return a conflist.
  ;; Assume that recordsep appears at the end of the database as well.

  ;; here "-m" means "matches"
  (let* ((fsep-m (progn (goto-char (point-min))
                        (db-how-many-string-overlapping fsep)))
         (rsep-m (progn (goto-char (point-min))
                        (db-how-many-string-overlapping rsep)))
         (goal-fsep-m (and nrecords (* nrecords (1- nfields))))
         (goal-rsep-m nrecords))
    (if (equal fsep rsep)
        (progn
          (if nrecords
              (if (= rsep-m (+ goal-rsep-m goal-fsep-m))
                  (list nil nil nrecords)
                (list t t nil))
            ;; Remember that fsep = rsep when reading this code.
            (if (zerop (% rsep-m nfields))
                (list nil nil (/ rsep-m nfields))
              (if (zerop (% (1+ rsep-m) nfields))
                  ;; A field separator at the end of the data was misinterpreted
                  ;; as a record separator, so no record separator was added.
                  (progn (goto-char (point-max))
                         (insert rsep)
                         (list nil nil (/ (1+ rsep-m) nfields)))
                (list t t (/ rsep-m nfields))))))

      ;; fsep and rsep unequal; see if one is a substring of the other.

      ;; At least one of these must be zero.
      (let ((f-in-r (db-how-many-substring-overlapping fsep rsep))
            (r-in-f (db-how-many-substring-overlapping rsep fsep)))
        (if nrecords
            (progn
              ;; At most one of these is nonzero, so cond is OK.
              (cond ((> f-in-r 0)
                     (incf goal-fsep-m (* goal-rsep-m f-in-r)))
                    ((> r-in-f 0)
                     (incf goal-rsep-m (* goal-fsep-m r-in-f))))
              (list (not (= fsep-m goal-fsep-m))
                    (not (= rsep-m goal-rsep-m))
                    nrecords))
          (incf nfields f-in-r)
          (let ((apparent-records (/ fsep-m (1- nfields))))
            (setq goal-rsep-m
                  (if (zerop apparent-records)
                      1
                    (* apparent-records
                       (1+ (* r-in-f fsep-m)))))
            (list (or
                   ;; Wrong number of fields; some record has too many or few.
                   (not (= fsep-m (* apparent-records (1- nfields))))
                   ;; too many fseps compared to rseps
                   (< rsep-m goal-rsep-m))
                  ;; too many rseps compared to fseps
                  (< goal-rsep-m rsep-m)
                  apparent-records)))))))


(defvar db-substitution-no-no-string
  ;; Global; hope we only write/read one database at a time.
  ;; It should be reset before starting to read/write a database.
  nil)

(defun database-acceptable-delimiter-p (delimiter)
  ;; Return t iff no characters of delimiter appear in
  ;; `substitution-no-no-string'.
  (when delimiter
    (let ((result t)
          (idx 0)
          (len (length delimiter)))
      (while (and result (< idx len))
        (if (db-find-char (elt delimiter idx) db-substitution-no-no-string)
            (setq result nil)
          (incf idx)))
      result)))

(defun database-generate-delimiter (database &optional checkp)
  (let ((fsep (or (database-sub-fieldsep-string database)
                  (database-full-fieldsep-string database)))
        (rsep (or (database-sub-recordsep-string database)
                  (database-full-recordsep-string database)))
        (string (make-string 1 0))
        (c 0)
        rv)
    (while (and (< c 256) (not rv))
      (aset string 0 c)
      (if (and (database-acceptable-delimiter-p string)
               (not (and checkp
                         (progn
                           (goto-char (point-min))
                           (search-forward string nil t))))
               (not (db-find-char c fsep))
               (not (db-find-char c rsep)))
          (setq rv string)
        (incf c)))
    (or rv (error "I can't find an acceptable delimiter!"))))

;;; db-file-io.el ends here
