;;; db-format.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Displaying and editing database records.

;;; Code:

(require 'db-rep)
(eval-when-compile (require 'cl))

(require 'easymenu)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

;;
;; The format -- text and displayspecs
;;

(deflocalvar dbf-displayspecs-length nil
  "The number of displayspecs in the current format.")

(deflocalvar dbf-displayspecs nil
  "An array of field specifiers, one for each field of the display format.
There are `dbf-displayspecs-length' of them.")

(deflocalvar dbf-inter-field-text nil
  "A vector with one string, the constant text that precedes that field in
the display, per displayspec, plus a final slot for trailing text.")

(deflocalvar dbf-recordindex-displayspecno-vector nil)

(defun dbf-fieldname->displayspecno (name)
  (aref dbf-recordindex-displayspecno-vector
        (fieldname->fieldnumber name dbc-database)))


;; Faces

;; There should perhaps be another face for read-only fields.

(deflocalvar dbf-default-field-face nil
  "If non-nil, the face for fields in data display buffers.
If the text already has a face, that takes precedence.
This variable is examined only when a data display buffer is being set up.")

(deflocalvar dbf-default-inter-field-face nil
  "If non-nil, the face for uneditable text in data display buffers.
If the text already has a face, that takes precedence.
This variable is examined only when a data display buffer is being set up.")

(deflocalvar dbf-selected-field-face
  "If non-nil, the face for the field being currently edited.")

(deflocalvar dbf-always-forms nil
  "Forms executed every time that the format is selected.
These forms are only executed when a different format is replaced, not
every time that a record is displayed (or even every time that
`db-change-format' is called).
See also `dbf-before-display-record-function'.")

;;
;; Location in the format (current field info)
;;

;; All of these variables include "this-" in their names.

(deflocalvar dbf-this-displayspec nil
  "The displayspec currently being operated upon, or nil.")

(deflocalvar dbf-this-field-index nil
  "The index in `dbf-displayspecs' of the current displayspec, or nil.")

(defsubst dbf-this-field-name ()
  (and dbf-this-displayspec
       (db-fname<-fno (displayspec-record-index dbf-this-displayspec)
                      dbc-database)))

(defsubst dbf-this-field-text ()
  ;; The text actually in the buffer.
  (buffer-substring dbf-this-field-beginning-pos (dbf-this-field-end-pos)))

(defsubst dbf-set-this-field-text (field-text)
  "Make the format display FIELD-TEXT in the current field."
  ;; Set the text actually in the buffer.
  (delete-region dbf-this-field-beginning-pos (dbf-this-field-end-pos))
  (goto-char dbf-this-field-beginning-pos)
  (insert field-text))

(defsubst dbf-this-field-modified-p ()
  (buffer-modified-p))
(defsubst dbf-set-this-field-modified-p (arg)
  (set-buffer-modified-p arg))

(deflocalvar dbf-this-field-beginning-pos nil
  "A position, the beginning of the current field.")

(deflocalvar dbf-this-field-end-marker (make-marker)
  "A mark one character past the end of the current field, or nil if current
field extends to end of buffer.")

(defvar dbf-moving-mark
  ;; Remember where the user just moved while possibly munging a field.
  ;; Only need one of these ever.
  (make-marker))

(defsubst dbf-point-min ()
  ;; Assume never narrowed.
  ;; fixme: remove assumption. --ttn
  1)


;;
;; The displayed record
;;

(deflocalvar dbf-this-record
  ;; Do not confuse with the record in the current link.  Real live database
  ;; records are never directly operated upon; we always munge a copy so that
  ;; the original can be restored if desired.
  nil
  "The record currently displayed and edited.  This is an honest-to-goodness
record whose slots are filled from `dbf-this-record-original' if it's modified.
The variable's value should never be set except by `db-copy-r2r'; its
slots may be freely modified, however.
This is only used if `dbf-this-record-modified-p' is t.")

(deflocalvar dbf-this-record-original nil
  "The original of `dbf-this-record'; a pointer to some poor unsuspecting
record that shouldn't be modified until everything has been checked out.
That is, when the user is setting fields, this record remains unchanged
and  dbf-this-record, a copy of the original, is munged instead.")

(deflocalvar dbf-this-record-modified-p nil
  "T if the current record has been modified, nil otherwise.
This determines which record is returned by `dbf-displayed-record':
if non-nil, then `dbf-this-record-original' has been copied to `dbf-this-record'.
It's best to use `dbf-set-this-record-modified-p' to set this variable.")

(deflocalvar dbf-set-this-record-modified-function nil
  "A function called when the current record is marked as modified.
The function takes no arguments and its return value is ignored.
It is called after `dbf-this-record-original' is copied to `dbf-this-record'
and after `dbf-this-record-modified-p' is set to t.")

(defsubst dbf-set-this-record-modified-p (arg)
  "Set the value of `dbf-this-record-modified-p' to ARG.
If ARG is non-nil and `dbf-this-record-modified-p' is nil, also do the
necessary record-copying and call `dbf-set-this-record-modified-function'."
  (cond
   ((and arg (not dbf-this-record-modified-p))
    (setq dbf-this-record-modified-p arg)
    (db-copy-r2r dbf-this-record-original dbf-this-record)
    (if dbf-set-this-record-modified-function
        (funcall dbf-set-this-record-modified-function)))
   (t
    (setq dbf-this-record-modified-p arg))))

(defsubst dbf-displayed-record ()
  "Return the record currently displayed in this data display buffer.
This is `dbf-this-record' if `dbf-this-record-modified-p' is non-nil and
`dbf-this-record-original' otherwise."
  (if dbf-this-record-modified-p
      dbf-this-record
    dbf-this-record-original))

(deflocalvar dbf-fields-displayed nil
  "A vector of one string, the displayed text for that field, per displayspec.")

(deflocalvar dbf-redisplay-entire-record-p nil
  "T if the whole record needs to be redisplayed.
This is often set by change functions.")


;;
;; Hooks
;;

;;; Minor mode hooks

(defvar db-view-mode-hooks nil
  "Function or list of functions called when Database View mode is entered.")

(defvar db-edit-mode-hooks nil
  "Function or list of functions called when Database Edit mode is entered.")

;;; Movement hooks

(deflocalvar dbf-before-display-record-function nil
  "A function called before a record is displayed.
The function takes one argument, the record.

This is a good place to put calls to `db-change-format'.  Depending on
your function's implementation, however, you may silently override any user
calls to that function.")

(deflocalvar dbf-after-display-record-hook nil
  "Hooks called after a record is displayed.
The hooks are called in the data display buffer.
Call `dbf-displayed-record' to get the just-displayed record.")

(deflocalvar dbf-enter-field-hook nil
  "A function (of no arguments) called whenever a display field is entered.
The displayspec index is `dbf-this-field-index'.")

;;; Change hooks

(deflocalvar dbf-first-change-function nil
  "A function called the first time a record field is modified, or nil.
The function takes the fieldname and the old and new values as arguments,
and returns t if the record should be redisplayed.")

(deflocalvar dbf-every-change-function nil
  "A function called whenever a record field is modified, or nil.
The function takes the fieldname and the old and new values as arguments,
and returns t if the record should be redisplayed.")

(deflocalvar dbf-change-functions nil
  "A vector of one function (or nil) per record field (not display field).
The functions take the fieldname and the old and new values as arguments,
and return t if the record should be redisplayed.
Use `dbf-set-change-function' to set the fields of this vector.")

(defun dbf-set-change-function (fieldname function)
  "Set the change function for FIELDNAME to FUNCTION in the current database.
FUNCTION takes the fieldname and the old and new values as arguments,
and returns t if the record should be redisplayed."
  (aset dbf-change-functions
        (fieldname->fieldnumber fieldname dbc-database)
        function))

(deflocalvar dbf-after-record-change-function nil
  "Function called whenever changes to a record are recorded semi-permanently
by `dbf-process-current-record-maybe'.  For convenience, the function
takes the record as an argument, which is guaranteed to be `dbf-this-record'.
Its return value is ignored.")

(deflocalvar dbf-reset-on-edit-list nil
  "An alist of (variable-name . default-value) pairs.
Every time Database Edit mode is entered, these buffer-local variables are
reset to their default values.  This is good for making sure that something
only happens once each time a record is edited.")


;;
;; The minor mode
;;

(deflocalvar dbf-minor-mode nil
  "A symbol, either 'view or 'edit.")

(deflocalvar dbf-minor-mode-name nil
  "\"View\" or \"Edit\".")

;;
;; The format
;;

;; Some variables local to the data display buffer don't need to be changed
;; when the display format changes.  The ones appearing below do.

(deflocalvar dbf-format-name nil
  "This buffer-local string names the format currently in use.
This should not be set by the user.")

(deflocalvar dbf-format-file nil
  "The format file from which the current format was built.")

(deflocalvar dbf-format-name-spec-alist nil
  "Association list of format names and format specifiers.
Each format name is an arbitrary string.
A format specifier is a filename or format file specifier, which is
a list of values for format variables.
The user sets the format specifier to a filename, and after that format file
has been read, EDB replaces the filename with a list of values for format
variables, so that the file need not be read again.

It is convenient for a database designer to set this, pre-assigning format
names to files so that the user only needs to remember the format names,
not the filenames.")

(deflocalvar dbf-format-file-spec-alist nil
  "Association list of file names and format file specifiers.
A format file specifier is a list of values for format variables.
Don't set this variable; use `dbf-format-name-spec-alist' instead.")

(defun dbf-make-format-spec ()
  ;; All of these items vary from format to format within a particular
  ;; data display buffer.
  (list
   dbf-format-file
   ;; These can vary between data display buffers which happen to be using
   ;; the same format file to specify the layout of the record's fields.
   ;; That is, these are specific to a particular data display buffer, not
   ;; to a format, because they have to do with what is actually being
   ;; displayed and/or because we might expect the user to change them
   ;; after reading in the format.  This is why we can't just associate
   ;; this information with the format file, but have to save it on a
   ;; per-data-display-buffer basis.
   dbf-summary-format
   dbf-summary-function
   dbf-fields-displayed
   dbf-field-search-defaults))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Constants
;;;

(defconst db-symbol-regexp "[a-zA-Z][-<>a-zA-Z0-9]*")

(defvar db-dt-leading-comma-optional nil
  "Non-nil if symbolic displaytypes may immediately follow a fieldname.
Nil if a comma must come between the fieldname and the displaytype symbol.
Displaytype symbols include # $ \" '.
When this is non-nil, then sometimes Emacs 19 searches are intolerably slow.")

(defconst db-dtrx (concat (if db-dt-leading-comma-optional
                              ",?" ",")
                          "\\("
                          db-symbol-regexp
                          "\\|"
                          ;; (brackets-surrounded material,
                          ;; for one-char alternative types)
                          "#\\|\\$\\|\"\\|'\\|\\[[^]]+\\]"
                          "\\)"))

(defconst db-foptrx (concat db-dtrx
                            "\\(=\\("
                            "[-<>a-zA-Z0-9]+"
                            "\\)\\)?"))
(defconst db-foptrx-symbol 1)
(defconst db-foptrx-equals 3)

(defconst db-dsrx (concat "\\("
                          (concat "\\\\" db-symbol-regexp)
                          "\\)"
                          (concat "\\(" db-foptrx "\\)*")
                          ;; possibly "\ " at the end
                          "\\(\\\\ \\)?"))
(defconst db-dsrx-fieldname 1)
(defconst db-dsrx-fieldoptions 2)
(defconst db-dsrx-content-beginning db-dsrx-fieldname)
(defconst db-dsrx-content-end 0)
;; If there was no match for the fieldoptions
(defconst db-dsrx-content-end-alt db-dsrx-fieldname)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Abstraction
;;;

;;
;; Displayspec
;;

;; Besides `record-index' all other information has to do with display only.

(defstruct (displayspec
            (:type vector)              ; explicit to support kwidx
            :named)
  record-index                          ; zero-based backref

  ;; size and shape
  indent
  min-width
  max-width
  min-height                            ; default 1
  max-height                            ; default 1
  min-bytes
  max-bytes

  ;; other display info
  truncation-display-action
  padding-action
  actual->display
  display->actual
  ;; Is this where these belong?  Well, it lets me not make a new displayspec
  ;; for them...
  match-actual->display
  match-display->actual

  ;; editing info
  truncation-editing-action
  (reachablep t))

(put 'displayspec 'kwidx                ; poor man's defsetf --ttn
     (let ((idx 0))
       (mapcar (lambda (ent)
                 (cons (intern (format ":%s" (symbol-name (car ent))))
                       (incf idx)))
               (cdr (get 'displayspec 'cl-struct-slots)))))


;;
;; Optspecinfo
;;

;; An optspecinfo tells how to interpret optional parameters to a
;; display field specification.  An optspecinfo is a three-element list of
;;  * param-name: a string
;;  * settor: function taking a displayspec and a value and setting a slot
;;  * opt-param->value: a function converting the optional parameter (that
;;    is, the string that follows the equal sign) into the actual value.

(defconst db-optspec-list
  (mapcar (lambda (x)
            `(,(nth 0 x)
              ,(flet ((dset (frag) (let ((f (intern (format "displayspec-%s"
                                                            frag))))
                                     `(lambda (ds val)
                                        (setf (,f ds) val)))))
                 ;; Settor-or-accessor is either a settor function, a
                 ;; slotname, or a list of slotnames.  In the latter two
                 ;; cases, it's first converted into a settor.
                 (let ((settor-or-accessor (nth 1 x)))
                   (cond ((functionp settor-or-accessor)
                          settor-or-accessor)
                         ((symbolp settor-or-accessor)
                          (dset settor-or-accessor))
                         (t
                          `(lambda (displayspec value)
                             ,@(mapcar (lambda (frag)
                                         `(,(dset frag) displayspec value))
                                       settor-or-accessor))))))
              ,(nth 2 x)))
          '(("indent" indent (lambda (x) t))
            ("noindent" indent (lambda (x) nil))

            ("width" (min-width max-width) db-string->number)
            ("min-width" min-width db-string->number)
            ("max-width" max-width db-string->number)
            ("length" (min-width max-width) db-string->number)
            ("min-length" min-width db-string->number)
            ("max-length" max-width db-string->number)
            ("height" (min-height max-height) db-string->number)
            ("min-height" min-height db-string->number)
            ("max-height" max-height db-string->number)
            ("bytes" (min-bytes max-bytes) db-string->number)
            ("min-bytes" min-bytes db-string->number)
            ("max-bytes" max-bytes db-string->number)

            ("trunc-display" truncation-display-action intern)
            ("truncation-display-action" truncation-display-action intern)
            ("padding-action" padding-action intern)
            ("right-justify" padding-action (lambda (&rest x) '(32 . t)))
            ("actual->display" actual->display intern)
            ("a->d" actual->display intern)
            ("display->actual" display->actual intern)
            ("d->a" display->actual intern)

            ;; match-actual->display and match-display->actual, fields 13 and 14

            ("truncation-editing-action" truncation-editing-action intern)
            ("trunc-edit" truncation-editing-action intern)
            ("reachable" reachablep (lambda (x) t))
            ("unreachable" reachablep (lambda (x) nil)))))


(defun db-pad-right (min-width rep rep-length)
  (concat rep (make-string (- min-width rep-length) 32)))

(defmacro db-display->actual-call (d->a fieldtext prev-value record fno)
  `(let ((dac-d->a ,d->a)
         (dac-fieldtext ,fieldtext))
     (if dac-d->a
         (db-vararg-call funcall 2 5
                         dac-d->a dac-fieldtext
                         ,prev-value ,record ,fno)
       dac-fieldtext)))

(defmacro db-actual->display-call (a->d fieldtext record fno)
  `(let* ((adc-a->d ,a->d)
          (adc-fieldtext ,fieldtext)
          (adc-result (if adc-a->d
                          (db-vararg-call funcall 2 4
                                          adc-a->d adc-fieldtext
                                          ,record ,fno)
                        adc-fieldtext)))
     (if (stringp adc-result)
         adc-result
       "<ERROR>")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Macros
;;;

;; If the user has deleted some of the leading spaces, they'll be restored.
;; Don't do anything about tabs, not even untabifying.

(defun db-unindentify (text)
  (let ((amt (dbf-this-field-indent)))
    (if amt
        (replace-regexp-in-string
         (concat "\n" (db-space-maybe-rx amt)) "\n" text)
      text)))

(defun db-space-maybe-rx (n)
  ;; Return a regexp matching N or fewer occurrences of the space character.
  ;; If N is nil, return the empty string, which is sometimes not a regexp you
  ;; want to search for by itself.
  (if n
      (if (> n
             ;; Emacs 19's regexp routines fix bugs in the Emacs 18 and Lucid
             ;; Emacs versions, but are sometimes much slower.  For deeply
             ;; indented fields, this can result in very slow editing.  We
             ;; disable some error-checking and correction for fields indented
             ;; more than 8 characters.
             8)
          (make-string (or n 0) 32)
        (let ((result (make-string (* 2 n) 32)))
          (setq n (1- (* 2 n)))
          (while (> n 0)
            (aset result n ??)
            (decf n 2))
          result))
    ""))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mode selection
;;;


(defvar database-view-mode-menu)        ; fixme: document. --ttn
(defvar database-edit-mode-menu)        ; fixme: document. --ttn


(defun db-view-mode (&optional arg)
  "Switch to Database View mode.
With an argument, toggle between Database View and Database Edit modes."
  (interactive "P")

  (cond ((not (db-data-display-buffer-p))
         (error "Only call this in database mode."))
        ((and arg (eq dbf-minor-mode 'view))
         (db-edit-mode))
        ;; If already in Database View mode, don't do anything.
        ((not (eq dbf-minor-mode 'view))
         (dbf-process-field-maybe t)
         (setq dbf-minor-mode 'view
               dbf-minor-mode-name "View")
         (use-local-map database-view-mode-map)
         (setq dbf-this-field-index nil
               dbf-this-displayspec nil)
         (setq buffer-read-only t)
         (goto-char (dbf-point-min))
         (dbf-set-this-field-modified-p nil)
         (easy-menu-remove database-edit-mode-menu)
         (easy-menu-add database-view-mode-menu)
         (run-hooks 'db-view-mode-hooks)
         (force-mode-line-update))))

(defun db-edit-mode (&optional arg)
  "Switch to Database Edit mode.
With an argument, toggle between Database Edit and Database View modes."
  (cond ((not (db-data-display-buffer-p))
         (error "Only call this in database mode."))
        ((and arg (eq dbf-minor-mode 'edit))
         (db-view-mode))
        (t
         (setq dbf-minor-mode 'edit
               dbf-minor-mode-name "Edit")
         (use-local-map database-edit-mode-map)
         (if (database-modifiable-p dbc-database)
             (setq buffer-read-only nil)
           (message (substitute-command-keys
                     (concat "Database is not modifiable; "
                             "change that with \\[db-toggle-modifiable-p]"))))
         (mapc (lambda (name-val)
                 (make-variable-buffer-local (car name-val))
                 (set (car name-val) (cdr name-val)))
               dbf-reset-on-edit-list)
         (easy-menu-add database-edit-mode-menu)
         (easy-menu-remove database-view-mode-menu)
         (run-hooks 'db-edit-mode-hooks)
         (force-mode-line-update))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Movement in the format
;;;

(defun db-parse-buffer-error (format-string &rest args)
  (db-emergency-restore-format)
  (db-message "%s; %s"
              "I was confused about where I was"
              "changes to the field might have been lost"))

(defun db-next-line-or-field (arg)
  "Move to ARGth next line.  If that would move out of the current field,
move to the closest field to that, but not the current one, wrapping if necessary."
  (interactive "p")
  (let ((goal-column (current-column))
        goal-line)
    ;; Determine goal line.
    (db-forward-line-wrapping arg)
    (db-jump-to-point)
    (setq goal-line (db-current-line))
    ;; Move to proper column.
    (move-to-column goal-column)
    (db-jump-to-point)
    ;; If off the goal line, move back and as near to the goal column as possible.
    (when (> (db-current-line) goal-line)
      (db-previous-field-internal 1)
      (goto-char (dbf-this-field-end-pos)))))

(defun db-move-to-field-exact (arg)
  "Move to the ARGth field in the display.  Ignores reachablep."
  (db-first-field-internal t)
  (db-next-field-internal arg t)
  (run-hooks 'dbf-enter-field-hook))

(defun db-next-field (arg)
  "Move to ARGth next reachable field, wrapping if necessary.
When called interactively, ARG defaults to 1."
  (interactive "p")
  (dbf-process-field-maybe t)
  (goto-char dbf-this-field-beginning-pos)
  (if (> arg 0)
      (db-next-field-internal arg)
    (db-previous-field-internal (- arg)))
  ;; We have just moved to a new field, which certainly isn't modified yet.
  (dbf-set-this-field-modified-p nil)
  (run-hooks 'dbf-enter-field-hook))

(defun db-next-field-internal (arg &optional exact)
  ;; Arg should be positive.  Assumes point is at the beginning of the field.
  ;; If EXACT is non-nil, reachablep is ignored.
  (while (> arg 0)
    (if (not (db-skip-string-forward
              (aref dbf-fields-displayed dbf-this-field-index)))
        (db-parse-buffer-error
         "Didn't find field %s text `%s'."
         dbf-this-field-index
         (aref dbf-fields-displayed dbf-this-field-index))
      (incf dbf-this-field-index)
      (decf arg)
      (when (= dbf-this-field-index dbf-displayspecs-length)
        (unless (db-skip-string-forward
                 (aref dbf-inter-field-text dbf-displayspecs-length))
          (db-parse-buffer-error
           "Didn't find trailing text `%s' after field %s."
           (aref dbf-inter-field-text dbf-displayspecs-length)
           (1- dbf-displayspecs-length)))
        (setq dbf-this-field-index 0)
        (goto-char (dbf-point-min)))
      (unless (db-skip-string-forward
               (aref dbf-inter-field-text dbf-this-field-index))
        (db-parse-buffer-error
         "Didn't find field separator `%s' before field %s."
         (aref dbf-inter-field-text dbf-this-field-index)
         dbf-this-field-index))
      ;; Implement reachablep.
      ;; fixme: handle infinite loop. --ttn
      (unless (or exact
                  (displayspec-reachablep
                   (aref dbf-displayspecs dbf-this-field-index)))
        (incf arg))))
  (setq dbf-this-displayspec (aref dbf-displayspecs dbf-this-field-index)
        dbf-this-field-beginning-pos (point))
  (buffer-disable-undo (current-buffer))
  (buffer-enable-undo)

  (when (looking-at (regexp-quote (aref dbf-fields-displayed
                                        dbf-this-field-index)))
    (let ((end-of-match (match-end 0)))
      (set-marker dbf-this-field-end-marker
                  (if (= end-of-match (point-max))
                      nil
                    (1+ end-of-match))
                  (current-buffer)))))

(defun db-previous-line-or-field (arg)
  "Move to ARGth previous line.  If that would move out of the current field,
move to the closest field to that, but not the current one, wrapping if necessary."
  (interactive "p")
  (let ((goal-column (current-column))
        (vacated-line (db-current-line))
        this-line)
    (db-forward-line-wrapping (- arg))
    (move-to-column goal-column)
    (db-jump-to-point)
    (setq this-line (db-current-line))
    (when (= this-line vacated-line)
      ;; We moved to a line containing no field, so db-jump-to-point
      ;; put us in the field following point; ie, one on the line in
      ;; which we started.  This is not the desired behavior.
      ;; Get to a line containing a field.
      (db-previous-field-internal 1)
      (goto-char (dbf-this-field-end-pos))
      ;; Go to the correct column.
      (move-to-column goal-column)
      ;; Avoid getting dumped back into this field.
      (goto-char (min (point) (dbf-this-field-end-pos)))
      ;; And end up there.
      (db-jump-to-point))))

(defun db-previous-field (&optional arg)
  "Move to ARGth previous reachable field, wrapping if necessary.
When called interactively, ARG defaults to 1."
  (interactive "p")
  (dbf-process-field-maybe t)
  (goto-char dbf-this-field-beginning-pos)
  (if (> arg 0)
      (db-previous-field-internal arg)
    (db-next-field-internal (- arg)))
  (dbf-set-this-field-modified-p nil)
  (run-hooks 'dbf-enter-field-hook))

(defun db-previous-field-internal (arg)
  ;; Arg should be positive.  Assume point is at the beginning of the field.
  ;; pb = previous inter-field-text beginning
  (let ((pb (marker-position dbf-this-field-end-marker)))
    (if pb
        (decf pb))
    (while (> arg 0)
      (when (db-skip-string-backward
             (aref dbf-inter-field-text dbf-this-field-index))
        (setq pb (point))
        (decf dbf-this-field-index)
        (decf arg)
        (when (< dbf-this-field-index 0)
          (setq dbf-this-field-index (1- dbf-displayspecs-length))
          (goto-char (point-max))
          (if (db-skip-string-backward (aref dbf-inter-field-text
                                             dbf-displayspecs-length))
              (setq pb (point))
            (db-parse-buffer-error
             "Didn't find trailing text `%s' after field %s."
             (aref dbf-inter-field-text dbf-displayspecs-length)
             dbf-this-field-index)))
        (unless (db-skip-string-backward
                 (aref dbf-fields-displayed dbf-this-field-index))
          (db-parse-buffer-error
           "Didn't find field %s text `%s'."
           dbf-this-field-index
           (aref dbf-fields-displayed dbf-this-field-index)))
        ;; Implement reachablep.
        ;; fixme: handle infinite loop. --ttn
        (unless (displayspec-reachablep
                 (aref dbf-displayspecs dbf-this-field-index))
          (incf arg)))
      (db-parse-buffer-error
       "Didn't find field separator `%s' before field %s."
       (aref dbf-inter-field-text dbf-this-field-index)
       dbf-this-field-index))
    (setq dbf-this-displayspec (aref dbf-displayspecs dbf-this-field-index)
          dbf-this-field-beginning-pos (point))
    (buffer-disable-undo (current-buffer))
    (buffer-enable-undo)
    (set-marker dbf-this-field-end-marker
                (and pb
                     (if (or (= 1 pb)
                             (= (point-max) pb))
                         nil
                       (1+ pb))))))

(defun db-first-field-internal (&optional exact)
  ;; Move to first field.  Optional EXACT means ignore reachability.
  (if dbf-this-field-index
      (dbf-process-field-maybe t)
    (db-edit-mode))
  (setq dbf-this-field-index 0)
  ;; We need this even if field-index was nil, because someone might have
  ;; sneakily moved point.  (In fact, this is called after point is moved
  ;; via mouse.)
  (goto-char (dbf-point-min))
  (unless (db-skip-string-forward (aref dbf-inter-field-text 0))
    (db-parse-buffer-error
     "Didn't find field separator `%s' before field %s."
     (aref dbf-inter-field-text dbf-this-field-index)
     dbf-this-field-index))
  (db-next-field-internal 0)
  ;; Implement reachablep.
  (unless (or exact
              (displayspec-reachablep
               (aref dbf-displayspecs dbf-this-field-index)))
    (db-next-field-internal 1))
  (dbf-set-this-field-modified-p nil))

(defun db-first-field ()
  "Move to first field."
  (interactive)
  (db-first-field-internal nil)
  (run-hooks 'dbf-enter-field-hook))

(defun db-last-field ()
  "Move to last field."
  (interactive)
  (db-first-field-internal nil)
  (db-previous-field 1))

(defun db-scroll-up ()
  "Like scroll-up, but also edits the nearest database field."
  (interactive)
  (scroll-up)
  (db-jump-to-point t))

(defun db-scroll-down ()
  "Like scroll-down, but also edits the nearest database field."
  (interactive)
  (scroll-down)
  (db-jump-to-point t))

(defun db-jump-to-point (&optional quietly)
  "If in a data display buffer, move to the field containing or following point.
In a summary buffer, move to the record displayed around point."
  (interactive)
  (cond ((db-data-display-buffer-p)
         (unless (and dbf-this-field-index
                      (and (<= dbf-this-field-beginning-pos (point))
                           (<= (point) (dbf-this-field-end-pos))))
           ;; moving outside current field.
           (let ((new-point (point)))
             (set-marker dbf-moving-mark (point))
             ;; Go back to where we were: if we were in a field, get back in it.
             (when dbf-this-field-index
               (goto-char dbf-this-field-beginning-pos))
             (if (and dbf-this-field-index
                      (> (marker-position dbf-moving-mark) (point)))
                 ;; We are in a field and moving forward.
                 (progn
                   (dbf-process-field-maybe t)
                   (goto-char dbf-this-field-beginning-pos))
               (db-first-field-internal nil))
             ;; If the dbf-process-field-maybe redisplays the entire record,
             ;; the marker gets wiped out (points to the beginning of the
             ;; buffer, because the buffer is cleared and refilled).
             (let ((moving-pos (marker-position dbf-moving-mark)))
               (unless (= 1 moving-pos)
                 (setq new-point moving-pos)))
             (set-marker dbf-moving-mark nil)
             (while (and (> new-point (dbf-this-field-end-pos))
                         (< dbf-this-field-index (1- dbf-displayspecs-length)))
               ;; The EXACT argument is t so we don't infinite-loop when
               ;; the last field is unreachable.
               (db-next-field-internal 1 t))
             (unless (displayspec-reachablep dbf-this-displayspec)
               ;; This message is getting wiped out by the
               ;; mouse-button-up event.  How can I fix this?
               ;; Hint:  Transposing the following two statements is
               ;; not the answer.
               (unless quietly
                 (db-message "Field `%s' is unreachable"
                             (db-fname<-fno
                              (displayspec-record-index
                               dbf-this-displayspec)
                              dbc-database)))
               (db-next-field-internal 1))

             (run-hooks 'dbf-enter-field-hook)
             ;; The max makes sure we're in a field, not beyond it.
             ;; The min is there only for the last field (because we could
             ;; be past it, in which case there's not a following field).
             (goto-char (min (max new-point dbf-this-field-beginning-pos)
                             (dbf-this-field-end-pos)))))
         ;; Check not in indentation even if didn't move to a new field.
         (when (let ((amt (dbf-this-field-indent)))
                 (and amt
                      (> amt 0)
                      (db-looking-back-at "^ +")
                      (< (current-column) amt)))
           (db-beginning-of-line-or-field)))
        ((db-summary-buffer-p)
         ;; This is wrong in the presence of hidden directory lines.
         (beginning-of-line)
         (let* ((lines (count-lines dbs-point (point)))
                (signed (if (< dbs-point (point)) lines (- lines))))
           (goto-char dbs-point)
           (dbs-next-record-ignore-hiding (/ signed dbfs-lines))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Movement within a field
;;;

;; These shouldn't be called if not on a field, so they don't check.

(defsubst dbf-this-field-end-pos ()
  (let ((pos (marker-position dbf-this-field-end-marker)))
    (if pos
        (1- pos)
      (point-max))))

(defsubst dbf-this-field-beginning-pos ()
  dbf-this-field-beginning-pos)

(defun dbf-this-field-indent ()
  (and (displayspec-indent dbf-this-displayspec)
       (if (numberp (displayspec-indent dbf-this-displayspec))
           (displayspec-indent dbf-this-displayspec)
         (save-excursion
           (goto-char dbf-this-field-beginning-pos)
           (current-column)))))

;;;
;;; Checking
;;;

(defsubst dbf-check-if-outside-field (&optional quietly)
  ;; Move point to beginning of field if it's before that.
  (when (< (point) dbf-this-field-beginning-pos)
    (goto-char dbf-this-field-beginning-pos)
    (unless quietly
      (db-message "Beginning of field")))
  ;; Move point to end of field if it's beyond that.
  (let ((end-pos (dbf-this-field-end-pos)))
    (when (> (point) end-pos)
      (goto-char end-pos)
      (unless quietly
        (dbf-inform-outside-field "End of field.")))))

(deflocalvar dbf-field-boundary-action
  ;; So keyboard macros terminate.
  'error
  "Controls action when point attempts to leave a field.
One of nil, 'message, 'beep, 'ding, 'error.
'beep and 'ding are identical and also show a message.
Having a variable is overkill, but I don't yet know what the Right Thing is.")

(defun dbf-inform-outside-field (message)
  (cond ((eq 'error dbf-field-boundary-action)
         (error message))
        ((eq 'ding dbf-field-boundary-action)
         (db-message message)
         (ding))
        ((eq 'beep dbf-field-boundary-action)
         (db-message message)
         (beep))
        ((eq 'message dbf-field-boundary-action)
         (db-message message))
        ((eq nil dbf-field-boundary-action)
         nil)
        (t
         (error "What value does this dbf-field-boundary-action value mean?  %s"
                dbf-field-boundary-action))))

;;;
;;; Movement
;;;

(defsubst db-beginning-of-field ()
  "Move to the beginning of the current field."
  (interactive)
  (goto-char dbf-this-field-beginning-pos))

(defsubst db-end-of-field ()
  "Move to the end of the current field."
  (interactive)
  (goto-char (dbf-this-field-end-pos)))

(defun db-beginning-of-line-or-field ()
  "Move to the beginning of the current line of the current field."
  (interactive)
  (beginning-of-line)
  (db-skip-regexp-forward (db-space-maybe-rx (dbf-this-field-indent)))
  (dbf-check-if-outside-field t))

(defun db-end-of-line-or-field (arg)
  "Move to the end of the current line of the current field."
  (interactive "p")
  (end-of-line arg)
  (dbf-check-if-outside-field t))

(defun db-forward-char (arg)
  "Like forward-char, but won't go outside field."
  (interactive "p")
  (if (< arg 0)
      (db-backward-char (- arg))
    (let ((indent (dbf-this-field-indent)))
      (while (> arg 0)
        (if (eobp)
            ;; This is so we get the error "End of field"
            ;; instead of "End of buffer".
            (progn
              (setq arg 0)
              (dbf-inform-outside-field "End of field."))
          (forward-char 1)
          (db-skip-regexp-forward (concat "^" (db-space-maybe-rx indent)))
          (decf arg)))
      (dbf-check-if-outside-field))))

(defun db-backward-char (arg)
  "Like backward-char, but won't go outside field."
  (interactive "p")
  (if (< arg 0)
      (db-forward-char (- arg))
    (let ((indent (dbf-this-field-indent)))
      (while (> arg 0)
        (if (bobp)
            ;; This is so we get the error "Beginning of field"
            ;; instead of "Beginning of buffer".
            (progn
              (setq arg 0)
              (dbf-inform-outside-field "Beginning of field."))
          (db-skip-regexp-backward (concat "^" (db-space-maybe-rx indent)))
          (backward-char 1)
          (decf arg)))
      (dbf-check-if-outside-field))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Editing
;;;

(defun db-delete-char (arg)
  (interactive "p")
  "Like delete-char, but won't delete outside the field."
  (delete-region (point) (progn (db-forward-char arg) (point))))

(defun db-backward-delete-char (arg)
  (interactive "p")
  "Like delete-backward-char, but won't delete outside the field."
  (delete-region (point) (progn (db-backward-char arg) (point))))

(defun db-forward-word (arg)
  "Like forward-word, but won't go outside field."
  (interactive "p")
  (forward-word arg)
  (dbf-check-if-outside-field))

(defun db-backward-word (arg)
  "Like backward-word, but won't go outside field."
  (interactive "p")
  (db-forward-word (- arg)))

(defun db-copy-region-as-kill (beg end)
  "Save the region as if killed, but don't kill it."
  (interactive "r")
  (let ((text (db-unindentify (buffer-substring beg end))))
    (if (eq last-command 'db-kill-region)
        (kill-append text (< end beg))
      (push text kill-ring)
      (when (> (length kill-ring) kill-ring-max)
        (setcdr (nthcdr (1- kill-ring-max) kill-ring) nil))))
  (setq this-command 'db-kill-region)
  (setq kill-ring-yank-pointer kill-ring))

(defun db-kill-region (beg end)
  "Kill between point and mark.
The text is deleted but saved in the kill ring.  See `kill-region' for details."
  (interactive "*r")
  (db-copy-region-as-kill beg end)
  (delete-region beg end))

(defun db-kill-word (arg)
  "Like kill-word, but won't delete outside the field."
  (interactive "p")
  (db-kill-region (point) (progn (db-forward-word arg) (point))))

(defun db-backward-kill-word (arg)
  "Like backward-kill-word, but won't delete outside the field."
  (interactive "p")
  (db-kill-word (- arg)))

(defun db-kill-line (arg)
  "Like kill-line, but won't delete outside the field."
  (interactive "p")
  (let ((here (point)))
    (db-end-of-line-or-field arg)
    (when (< (point) (dbf-this-field-end-pos))
      (db-skip-regexp-forward
       (concat "[ \t]*\n" (db-space-maybe-rx (dbf-this-field-indent)))))
    (db-kill-region here (point))))

(defun db-kill-to-end ()
  "Kill from point to the end of the current field."
  (interactive)
  (db-kill-region (point) (dbf-this-field-end-pos)))

(defun db-newline (arg)
  "Insert a newline.  Will not make the current field too tall.
If the current field's maximum height is 1 line, move to the next field instead."
  (interactive "p")
  ;; ignores the argument
  (let ((max-h (displayspec-max-height dbf-this-displayspec)))
    (if (or (not max-h)
            (< (count-lines dbf-this-field-beginning-pos
                            (dbf-this-field-end-pos))
               max-h))
        (let ((indent (dbf-this-field-indent)))
          (newline 1)
          (when indent (insert (make-string indent 32))))
      (if (= 1 max-h)
          (db-next-field 1)
        (db-message "Field is at maximum height already")))))

(defun db-open-line (arg)
  "Insert a newline and leave point before it.
Will not make the current field too tall."
  (interactive "p")
  (let ((here (point)))
    (db-newline arg)
    (goto-char here)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Value processing for fields and records
;;;

(defsubst dbf-process-field-maybe (set-text-p)
  ;; Set the value of the current record from the current field.
  ;; If arg SET-TEXT-P is non-nil, update the display as well.
  ;; Return t if field is unmodified or text is OK; nil otherwise.
  ;; May move point.
  (if dbf-this-field-index
      (if (not (dbf-this-field-modified-p))
          t
        (when (or (< (point) dbf-this-field-beginning-pos)
                  (> (point) (dbf-this-field-end-pos)))
          (db-parse-buffer-error
           "Point was outside (%d) of current field (%d - %d)."
           (point) dbf-this-field-beginning-pos (dbf-this-field-end-pos)))
        ;; This field has been modified, and point is in the field as expected.
        (if (not (equal (dbf-this-field-text)
                        (aref dbf-fields-displayed dbf-this-field-index)))
            (let* ((cur (db-display->actual-call
                         (displayspec-display->actual dbf-this-displayspec)
                         (let ((text (dbf-this-field-text)))
                           (db-unindentify text))
                         (aref (dbf-displayed-record) dbf-this-field-index)
                         (dbf-displayed-record)
                         dbf-this-field-index))
                   (idx (displayspec-record-index dbf-this-displayspec))
                   (old (aref (dbf-displayed-record) idx))
                   (saved-modified-p dbf-this-record-modified-p))
              (unless (equal cur old)
                ;; The new value is different from the old.
                (dbf-set-this-record-modified-p t)
                (record-set-field-from-index
                 dbf-this-record idx cur dbc-database)

                (when set-text-p
                  (aset dbf-fields-displayed dbf-this-field-index
                        (let ((pr (db-ds-printed dbf-this-displayspec
                                                 dbf-this-record))
                              (in (displayspec-indent dbf-this-displayspec)))
                          (if in
                              (if (numberp in)
                                  (replace-regexp-in-string
                                   "\n"
                                   (concat "\n" (make-string in 32))
                                   pr)
                                ;; Why can't I use (dbf-this-field-indent) even here??
                                (if (db-find-char ?\n pr)
                                    (error "Don't know how much to indent.")
                                  pr))
                            pr))))
                ;; No need to do redisplay before the change-hooks are
                ;; called since the user's version is already onscreen
                ;; and that will be very similar indeed to the display
                ;; text.
                (unless saved-modified-p
                  (setq dbf-redisplay-entire-record-p
                        (or (and dbf-first-change-function
                                 (funcall dbf-first-change-function
                                          (dbf-this-field-name)
                                          old cur))
                            dbf-redisplay-entire-record-p)))
                (setq dbf-redisplay-entire-record-p
                      (or (and dbf-every-change-function
                               (funcall dbf-every-change-function
                                        (dbf-this-field-name)
                                        old cur))
                          dbf-redisplay-entire-record-p))
                (setq dbf-redisplay-entire-record-p
                      (let ((change-function (aref dbf-change-functions idx)))
                        (or (and change-function
                                 (funcall change-function
                                          (dbf-this-field-name)
                                          old cur))
                            dbf-redisplay-entire-record-p))))
              ;; The text is different; the value may or may not have differed.
              ;; Display the standard representation for this value, which has
              ;; already been computed.
              (when set-text-p
                (unless (dbf-redisplay-entire-record-maybe)
                  ;; set-field-text always returns nil
                  (dbf-set-this-field-text
                   (aref dbf-fields-displayed dbf-this-field-index))))
              (dbf-set-this-field-modified-p nil))
          ;; Field is unchanged, so mark it unmodified.
          (dbf-set-this-field-modified-p nil)))
    t))


(defun dbf-redisplay-entire-record-maybe ()
  ;; If `dbf-redisplay-entire-record-p' is non-nil, redisplay current record
  ;; and return t; otherwise return nil.
  (when dbf-redisplay-entire-record-p
    (setq dbf-redisplay-entire-record-p nil)
    (db-emergency-restore-format t)
    t))


(defun dbf-process-current-record-maybe (set-text-p)
  ;; Commit changes to the record being displayed and edited.  If the current
  ;; record (see `dbf-displayed-record') is a modified copy of a database
  ;; record, this copies it back to the original database record, modifying
  ;; the database by side effect.  Return t if successful, nil otherwise.
  ;; SET-TEXT-P non-nil means to also update the display.
  (when dbc-index
    ;; Sets the field unmodified, if appropriate
    (dbf-process-field-maybe set-text-p)
    (when dbf-this-record-modified-p
      ;; Do any programmer-requested checking or postprocessing here.
      ;; This function may err, aborting out of whatever was trying to
      ;; process the current record and do something else.
      (when dbf-after-record-change-function
        (funcall dbf-after-record-change-function
                 (dbf-displayed-record)))
      (db-copy-r2r dbf-this-record dbf-this-record-original)
      (setf (link-summary dbc-link) nil)
      (dbf-update-summary-item dbc-index dbc-link)
      ;; should set link-hiddenp too.
      (dbc-set-database-modified-p t)
      (setq dbf-this-record-modified-p nil)
      (dbf-set-this-field-modified-p nil))
    ;; This function shouldn't have been called on a non-database record; how
    ;; did we get here?  It may not be the case that the info is about to be
    ;; abandoned.
    (or (not dbf-this-record-modified-p)
        (y-or-n-p "Abandon the displayed information? ")
        (error "Don't abandon displayed information."))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Undoing changes
;;;

(defun db-revert-field (&optional quietly)
  "Undo any changes made since entering this field.
Replace the onscreen text in this field with that of the underlying record.

A similar effect can be had by invoking \\[advertised-undo] multiple times."
  (interactive)

  (if (dbf-this-field-modified-p)
      (progn
        (dbf-set-this-field-text
         (aref dbf-fields-displayed dbf-this-field-index))
        (dbf-set-this-field-modified-p nil)
        (if (not quietly)
            (db-message "Reverted field `%s'" (dbf-this-field-name))))
    (if (not quietly)
        (db-message "Can't revert field `%s'; no changes since moving onto it"
                    (dbf-this-field-name)))))

(defun db-revert-record ()
  "Set the record to be the same as the corresponding one in the database.
In other words, undo any changes made since entering this record."
  (interactive)
  (db-revert-field t)
  (if dbf-this-record-modified-p
      (let ((buffer-read-only nil))
        (setq dbf-this-record-modified-p nil)
        (db-display-record (dbf-displayed-record) t)
        (when dbf-this-field-index
          (db-move-to-field-exact dbf-this-field-index))
        (db-message "Reverted record"))
    (db-message "Can't revert this record; no changes since selecting it")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Set displayspec from string
;;;


(defun db-string->displaytype (s)
  ;; Return the displaytype (a symbol) corresponding to string S.
  ;; Non-strings are returned unchanged.
  ;; This doesn't work yet for multichar alternatives.
  (if (stringp s)
      (cond ((string= s "#")
             'integer)
            ((string= s "$")
             'number)
            ((string= s "\"")
             'string)
            ((string= s "'")
             'one-line-string)
            ((string-match "^\\[.*\\]$" s)
             (list 'alternative-one-char (substring s 1 -1)))
            (t
             (intern s)))
    s))

(defun db-dspec<-dtype (type)
  "Return a copy of the displayspec corresponding to string or symbol TYPE.
Return nil if there's no corresponding displayspec."
  (let ((ds (gethash (db-string->displaytype type) db-displaytypes)))
    (when ds (if (symbolp ds)
                 (db-dspec<-dtype ds)
               (copy-displayspec ds)))))


(defun db-dspec<-string (s database)
  ;; Assume match-data set.  S is nil if from the buffer.
  (let* ((fname (intern (substring (match-string db-dsrx-fieldname s) 1)))
         (fidx (and database (fieldname->fieldnumber fname database)))
         ds)
    (when (and database (not fidx))
      (error "%s is not a field or field abbreviation."
             fname))
    (setq ds (db-dspec<-type/opts
              (database-recordfieldspec-type database fidx)
              (match-string db-dsrx-fieldoptions s)))
    (unless ds
      (error "Type %s in field %d (%s) not recognized."
             (database-recordfieldspec-type database fidx)
             fname fidx))
    (setf (displayspec-record-index ds) fidx)
    ds))


(defun db-dspec<-type/opts (type opts &optional notype-ok)
  ;; Either TYPE or OPTS (a string) must specify a type, unless
  ;; optional argument NOTYPE-OK is specified, in which case an empty
  ;; displayspec may be returned.

  ;; Ordinarily (for instance, when this is being called to parse part of a
  ;; format), NOTYPE-OK should not be specified, so that invalid
  ;; displaytypes aren't created.

  ;; A type in OPTS overrides TYPE.
  (if (not opts)
      (if type
          (or (db-dspec<-dtype type)
              (error "No such displaytype as `%s'." type))
        (make-displayspec))
    (let (ds end)
      ;; set the displayspec
      ;; note tricky sequencing
      (if (and (string-match (concat "^" db-dtrx) opts)
               (setq ds (db-dspec<-dtype
                         (db-string->displaytype (match-string 1 opts)))))
          (setq opts (substring opts (match-end 0)))
        (if type
            (setq ds (db-dspec<-dtype type))
          (error "No type specified in `%s'." opts)))

      (while (not (equal "" opts))
        (unless (string-match (concat "^" db-foptrx) opts)
          (error "`%s' isn't an optional field specification."
                 opts))
        (setq end (match-end 0))
        (let* ((opt (match-string db-foptrx-symbol opts))
               (val (match-string db-foptrx-equals opts))
               (spec (or (assoc opt db-optspec-list)
                         (error "Invalid optional field spec name or type: %s"
                                opt))))
          (funcall (nth 1 spec) ds (funcall (nth 2 spec) val)))
        (setq opts (substring opts end)))
      ds)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read a format file
;;;

(defun db-setup-data-display-buffer (format-file database new-p)
  ;; Create and return a data display buffer.  This is only called when a
  ;; brand-new data display buffer is being created, not when one is being
  ;; refreshed.  Arguments are FORMAT-FILE DATABASE NEW-P.  If NEW-P is
  ;; non-nil, then the database's auxiliary file is read and its field
  ;; variables are set.
  ;;
  ;; WARNING: If the format file's local variables set particular database
  ;; slots (such as fieldnames), and NEW-P is nil, then the database may be
  ;; left in an inconsistent state.  The "primary" format, which is read in
  ;; before the database is, should perform any such neccessary actions.
  (setq format-file (expand-file-name format-file))
  (unless (file-readable-p format-file)
    (error "Can't read format file `%s'." format-file))

  (set-buffer (db-make-data-display-buffer database new-p))
  (assert (eq dbc-database database) t)
  (setq buffer-read-only nil)           ; (database-mode) set it to t

  (funcall (if (fboundp 'crypt-insert-file-contents)
               'crypt-insert-file-contents
             'insert-file-contents)
           format-file)
  (setq dbf-format-file format-file)

  (when new-p
    (let ((aux (or (database-aux-file database)
                   (db-locate-readable-file-prefer-cwd
                    (file-name-sans-extension (database-file database))
                    (cons default-directory db-aux-file-path)
                    db-aux-file-suffixes))))
      (when aux
        ;; Note that the variable `database' is dynamically bound.
        (load-file aux))))
  (db-really-hack-local-variables)

  ;; Initialize local variables.
  ;; Assume `dbc-database' already set.
  (unless (database-no-of-fields dbc-database)
    (error "Can't tell how many fields database has."))
  (setq dbf-change-functions (make-vector (database-no-of-fields dbc-database)
                                          nil)
        dbf-this-record (make-record dbc-database))

  (when new-p
    ;; Initialize database variables.  We didn't do this earlier because
    ;; they may depend on some values set in the format file.
    (unless (database-internal-file-layout-p database)
      ;; Do this before the format is parsed but after the format's local
      ;; variables have been hacked.
      (unless (database-field-priorities database)
        (setf (database-field-priorities database)
              (let ((fno -1))
                (cons (mapcar (lambda (fieldname)
                                (incf fno)
                                (list fno))
                              (database-fieldname-alist database))
                      nil))))))

  (db-setup-ddb-parse-displayspecs database)

  (setq buffer-read-only t)
  (current-buffer))

(defun db-make-data-display-buffer (database new-p)
  ;; Create and return a data display buffer; set some vars.
  (db-in-buffer (create-file-buffer (database-file database))
    (let ((dir (file-name-directory (database-file database))))
      (when dir
        (setq default-directory (expand-file-name dir))))
    (setq dbc-database database)
    (unless new-p
      ;; These are per-data-display-buffer variables.
      (setq dbf-change-functions (make-vector (database-no-of-fields database)
                                              nil)
            dbf-this-record (make-record dbc-database)))
    (database-mode)
    (current-buffer)))


(defun db-setup-ddb-parse-displayspecs (database)
  ;; Get rid of local variables.
  (goto-char (point-max))
  (search-backward "\n\^L" (max (- (point-max) 3000) (point-min)) 'move)
  (when (search-forward
         (concat "Local"                ; use `concat' to avoid
                 " "                    ; misinterpretation
                 "Variables:")
         nil t)
    (beginning-of-line 1)
    (delete-region (point) (point-max)))
  ;; Get rid of whitespace at end of buffer.
  (goto-char (point-max))
  (re-search-backward "[^ \t\n]")
  (delete-region (match-end 0) (point-max))
  ;; Get rid of whitespace at ends of lines.
  (goto-char (dbf-point-min))
  (while (re-search-forward  "[ \t]+$" nil t)
    (replace-match ""))

  (let ((prev-end (dbf-point-min))
        (backslash-placeholder (and (goto-char (dbf-point-min))
                                    (search-forward "\\\\" nil t)
                                    ;; assume this doesn't return nil
                                    (db-unused-char-in-buffer)))
        beginning end ds
        ds-ls
        ift-ls)                         ; inter-field-text

    (when backslash-placeholder
      (setq backslash-placeholder (char-to-string backslash-placeholder))
      (goto-char (dbf-point-min))
      (while (search-forward "\\\\" nil t)
        (replace-match backslash-placeholder)))

    (setq dbf-default-summary-format nil)

    (goto-char (dbf-point-min))
    (while (re-search-forward db-dsrx nil t)
      (setq beginning (match-beginning db-dsrx-content-beginning)
            end (or (match-end db-dsrx-content-end)
                    (match-end db-dsrx-content-end-alt))
            ;; Call "internal" version of function because match-data is set.
            ;; nil as first argument means make it from the buffer.
            ds (db-dspec<-string nil database))

      ;; Fix up backslash-replacement.  The buffer is fixed up instead of
      ;; just the ift-ls because of the call to current-column.
      (when backslash-placeholder
        (save-excursion
          (save-restriction
            (narrow-to-region prev-end beginning)
            (goto-char prev-end)
            (while (search-forward backslash-placeholder nil t)
              (replace-match "\\")))))

      (push (buffer-substring prev-end beginning) ift-ls)
      ;; Match about to be deleted; we just used the old value.
      (setq prev-end beginning)

      (when (null dbf-default-summary-format)
        (setq dbf-default-summary-format (save-excursion
                                           (buffer-substring
                                            (progn (beginning-of-line 1)
                                                   (point))
                                            (progn (end-of-line 1)
                                                   (point)))))
        (when (null dbf-summary-format)
          (setq dbf-summary-format dbf-default-summary-format)))

      (delete-region beginning end)

      (when (eq t (displayspec-indent ds))
        (setf (displayspec-indent ds) (current-column)))

      (push ds ds-ls))

    ;; Fix up backslash-replacement for the post-last text.
    (when backslash-placeholder
      (save-excursion
        (save-restriction
          (narrow-to-region prev-end (point-max))
          (goto-char prev-end)
          (while (search-forward backslash-placeholder nil t)
            (replace-match "\\")))))

    (push (buffer-substring prev-end (point-max)) ift-ls)

    (setq dbf-inter-field-text (vconcat (nreverse ift-ls))
          dbf-displayspecs (vconcat (nreverse ds-ls))
          dbf-displayspecs-length (length dbf-displayspecs)
          dbf-fields-displayed (make-vector dbf-displayspecs-length nil)
          dbf-field-search-defaults (make-vector (1+ dbf-displayspecs-length) nil)))

  ;; Initialize more local variables.
  (setq dbf-recordindex-displayspecno-vector
        (make-vector (database-no-of-fields database) nil))
  (let ((fsno 0))
    (while (< fsno dbf-displayspecs-length)
      (aset dbf-recordindex-displayspecno-vector
            (displayspec-record-index (aref dbf-displayspecs fsno))
            fsno)
      (incf fsno)))

  (dbf-set-summary-format dbf-summary-format)

  (set-buffer-modified-p nil))


(defun db-additional-data-display-buffer ()
  "Create another data display buffer in which to view this database."
  (interactive)
  (dbf-process-current-record-maybe t)
  (let* ((cur (current-buffer))
         (database dbc-database)
         (new (db-make-data-display-buffer database nil)))
    (push new (database-data-display-buffers database))
    (switch-to-buffer-other-window new)

    (db-copy-buffer-local-variables cur)
    ;; Here are the trampled-on variables that we really cared about.
    (setq dbf-this-record (make-record dbc-database))
    (db-emergency-restore-format t)))


(defun db-change-format (&optional format-name filename)
  "Select and use an alternate display format to view the database.
If neither FORMAT-NAME nor FILENAME is specified (as is the case when this
is called interactively), the user is prompted for them.  In Emacs Lisp
code, if `dbf-format-name-spec-alist' has been been set, usually only one of
the arguments is specified.  If both are specified, then FORMAT-NAME
becomes a name for the format FILENAME specifies; if FORMAT-NAME is already
associated with a different format file, an error is signalled.

If the current format is unnamed, the user is prompted for a name
to give it, so that it can be conveniently restored if need be.  This
behavior is suppressed, and the record is not displayed, if the function is
not being called interactively.

The data display buffer is left in Database View mode.

Selecting the current format does not cause any work to be done.

Some databases automatically set the format of the record being displayed,
usually by setting `dbf-before-display-record-function' to a function that
overrides the format in effect when a record is about to be displayed.
This may cause this function to appear not to be doing any work.  In
actuality the format is being set, then reset."
  (interactive)

  (unless (and format-name
               (equal format-name dbf-format-name))
    ;; We're not already in the requested format
    (db-view-mode)

    ;; If neither format-name nor filename is specified, query for one of them.
    (unless (or format-name filename)
      (setq format-name
            (completing-read
             "Use which format? (? for options, RET to specify a file) "
             (cons '("") dbf-format-name-spec-alist)
             (lambda (elem)
               (stringp (car elem)))
             t))
      (when (equal "" format-name)
        (setq format-name nil
              filename (read-file-name "File for new format: " nil nil t))))

    ;; Either `format-name' or `filename' -- or possibly both,
    ;; if not called interactively -- is set.
    (when filename
      (setq filename (db-locate-format-file filename)))
    (when format-name
      (let ((spec (cdr (assoc format-name dbf-format-name-spec-alist))))
        (if spec
            ;; successful format-name
            (let ((fs-filename (if (listp spec)
                                   (car spec)
                                 spec)))
              (if filename
                  (when (and fs-filename
                             ;; This test is required for interactive
                             ;; uses of `db-change-format'.
                             (not (db-same-file-p filename fs-filename)))
                    (error "Format name %s is associated with %s, not %s."
                           format-name fs-filename filename))
                (setq filename (db-locate-format-file fs-filename))))
          ;; unsuccessful format-name
          (if filename
              (push (cons format-name filename) dbf-format-name-spec-alist)
            ;; no filename, failed format-name
            (error "`%s' is not the name of a format." format-name)))))
    ;; Filename is now set.

    ;; First save away current format.  No need to do anything with filename.
    (when (and (interactive-p)
               (not dbf-format-name)
               (y-or-n-p "Would you like to give the current format a name? "))
      (setq dbf-format-name (read-string "Name for current format: ")))
    (when dbf-format-name
      (let ((look (assoc dbf-format-name dbf-format-name-spec-alist)))
        (if look
            (setcdr look (dbf-make-format-spec))
          (push (cons dbf-format-name (dbf-make-format-spec))
                dbf-format-name-spec-alist))))

    ;; Now install the new format.
    (setq dbf-format-name format-name
          dbf-format-file filename)
    (let ((fs (cdr (assoc dbf-format-file dbf-format-file-spec-alist))))
      (if fs
          (progn
            (mapc 'eval (setq dbf-always-forms         (pop fs)))
            (setq dbf-displayspecs                     (pop fs)
                  dbf-displayspecs-length              (pop fs)
                  dbf-inter-field-text                 (pop fs)
                  dbf-recordindex-displayspecno-vector (car fs))
            (let ((spec (cdr (assoc (or dbf-format-name
                                        (intern dbf-format-file))
                                    dbf-format-name-spec-alist))))
              (setq dbf-format-file           (pop spec)
                    dbf-summary-format        (pop spec)
                    dbf-summary-function      (pop spec)
                    dbf-fields-displayed      (pop spec)
                    dbf-field-search-defaults (car spec))))
        ;; We didn't find dbf-format-file in dbf-format-file-spec-alist; we
        ;; probably didn't find more than just a filename at dbf-format-name
        ;; in dbf-format-name-spec-alist either.
        ;; This `let' is for the benefit of the new format file.
        (let ((database dbc-database)
              (buffer-read-only nil))
          (db-message "Reading format from file: %s" dbf-format-file)
          (buffer-disable-undo (current-buffer))
          (erase-buffer)
          (insert-file dbf-format-file)

          (db-really-hack-local-variables)

          (db-setup-ddb-parse-displayspecs dbc-database)

          ;; Save away the file-invariant stuff.
          (push (list dbf-format-file
                      dbf-always-forms
                      dbf-displayspecs
                      dbf-displayspecs-length
                      dbf-inter-field-text
                      dbf-recordindex-displayspecno-vector)
                dbf-format-file-spec-alist)
          ;; Install the defaults under a symbol associated with the format
          ;; file (so it's not user-accessible).
          (let ((dbf-summary-format dbf-default-summary-format)
                (dbf-summary-function (when (equal
                                             dbf-summary-format
                                             dbf-default-summary-format)
                                        dbf-summary-function)))
            (push (cons (intern dbf-format-file) (dbf-make-format-spec))
                  dbf-format-name-spec-alist))
          (erase-buffer))))

    (when (interactive-p)
      (db-display-record (dbf-displayed-record) t))))


(defun db-emergency-restore-format (&optional recompute)
  "Throw away the contents of the format buffer and redisplay the current record.
Use this if the format gets munged.
Changes made to the current field since last moving onto it may be lost.
If optional prefix argument RECOMPUTE is non-nil, the displayed text
is recomputed as well."
  (interactive "P")

  (db-display-record (dbf-displayed-record) recompute)

  (when dbf-this-field-index
    (let ((fidx dbf-this-field-index))
      (dbf-set-this-field-modified-p nil)
      (db-move-to-field-exact fidx)
      ;; If the hook changed formats, we'll be in Database View mode.
      (db-edit-mode))))

(defun dbf-set-summary-format (summary-format)
  "Specify the format used in the Database Summary buffer.
Argument SUMMARY-FORMAT is a string containing display specifications.
Call this in the data display buffer, or in a format file or auxiliary file."
  (interactive (list (read-string "Summary format: "
                                  (cons dbf-summary-format 0)
                                  nil
                                  dbf-summary-format)))
  (unless (stringp summary-format)
    (error "Argument to dbf-set-summary-format should be a string, not %s"
           summary-format))
  (when (= ?\n (elt summary-format (1- (length summary-format))))
    (setq summary-format (substring summary-format 0 -1)))
  (setq dbf-summary-format summary-format)
  (require 'db-summary)
  (dbf-set-summary-out-of-date-p)
  (setq dbf-summary-recompute-all-p t)
  ;; If the alist isn't yet set, then we're still setting up, and this will
  ;; be called later on; do nothing for now.
  (when (database-fieldname-alist dbc-database)
    (dbf-make-summary-maker summary-format dbc-database)))

(defmacro dbf-always (&rest body)
  "Execute BODY, and place its forms in `dbf-always-forms'.
They will be executed each time that this format replaces another."
  (declare (debug body))
  `(progn
     (setq dbf-always-forms (nconc dbf-always-forms ,body))
     ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Display data in a format
;;;

(defvar db-display-record-redo-inter-field-text-function nil
  "Function taking two args called in two contexts during `db-display-record'.
The args are BUF and EXTENTS-LIST.  In the first call, done right before
erasing the buffer, EXTENTS-LIST is null.  In the second call, done right
before the normal hook `dbf-after-display-record-hook' is run, EXTENTS-LIST
is a list of pairs of buffer positions in BUF that define the \"inter-field
text\", that is the text specifically not representing the data, but instead
part of the display format.")

(defun db-display-record (record &optional recompute fieldno-limit)
  ;; Why take a record as argument instead of using var `dbf-displayed-record'?
  ;; Joe Wells has used this feature, so don't remove it.
  "Display RECORD in the current buffer, which is a data display buffer.
If optional arg RECOMPUTE is non-nil, the display representations will be
computed first; RECOMPUTE is typically non-nil only the first time a record
is shown.  If optional third arg FIELDNO-LIMIT is non-nil, only
fieldnumbers strictly less than it will be displayed."
  (let ((fidx 0)
        ds
        rep
        (buffer-read-only nil)
        (is-displayed-record-p (eq record (dbf-displayed-record)))
        ;; If quitting occurs in the middle of this operation, EDB becomes
        ;; very confused.  WARNING: Inhibitting quitting is dangerous.
        (inhibit-quit t)
        ext-start
        ift-extents)
    ;; Assume `is-displayed-record' is bound.
    (db-funcall-maybe dbf-before-display-record-function record)
    ;; Allow dbf-before-display-record-function to do
    ;; dbf-set-this-record-modified-p if it wants to.
    (when is-displayed-record-p
      (setq record (dbf-displayed-record)))
    (dbc-update-database-modified-p)
    (buffer-disable-undo (current-buffer))
    (run-hook-with-args 'db-display-record-redo-inter-field-text-function
                        (current-buffer) nil)
    (erase-buffer)
    (while (< fidx dbf-displayspecs-length)
      (setq ds (aref dbf-displayspecs fidx))
      (setq ext-start (point))
      (insert (aref dbf-inter-field-text fidx))
      (push (cons ext-start (point)) ift-extents)
      (when recompute
        (aset dbf-fields-displayed fidx
              (if (and fieldno-limit (>= fidx fieldno-limit))
                  ;; fixme: handle min-height and min-bytes. --ttn
                  (make-string (or (displayspec-min-width ds) 0) 32)
                (setq rep (db-ds-printed ds record))
                (replace-regexp-in-string
                 "\n"
                 (concat "\n" (make-string (current-column) 32))
                 rep))))
      (insert (aref dbf-fields-displayed fidx))
      (incf fidx))
    (setq ext-start (point))
    (insert (aref dbf-inter-field-text fidx))
    (push (cons ext-start (point)) ift-extents)
    (dbf-set-this-field-modified-p nil)
    ;; This place is as good as any for leaving the cursor by default.
    (goto-char (dbf-point-min))
    (buffer-enable-undo (current-buffer))
    (run-hook-with-args 'db-display-record-redo-inter-field-text-function
                        (current-buffer)
                        (nreverse ift-extents))
    (run-hooks 'dbf-after-display-record-hook)
    ;; If quitting occurred while this was happening, ignore it.
    (setq quit-flag nil)))

(defun db-ds-printed (ds record)
  (let* ((ridx (displayspec-record-index ds))
         (rep (db-actual->display-call
               (displayspec-actual->display ds)
               (aref record ridx)
               record
               ridx)))
    (let ((min-h (displayspec-min-height ds))
          (max-h (displayspec-max-height ds)))
      (when (or min-h max-h)
        (let ((rep-h (1+ (db-count-newlines rep))))
          (cond ((and min-h (< rep-h min-h))
                 ;; too short
                 (setq rep
                       (concat rep
                               (make-string (- min-h rep-h)
                                            ?\n))))
                ((and max-h (> rep-h max-h))
                 ;; too tall
                 (setq rep
                       (substring rep 0
                                  (db-find-char-from-end
                                   ?\n rep
                                   (- rep-h min-h)))))))))

    ;; These conditions are much too simplistic; they only work for one-line
    ;; representations.
    (let ((rep-w (length rep))
          (min-w (displayspec-min-width ds))
          (max-w (displayspec-max-width ds)))
      (cond ((and min-w (< rep-w min-w))
             ;; The display representation is too short
             (setq rep (funcall (or (displayspec-padding-action ds)
                                    'db-pad-right)
                                min-w rep rep-w))
             (unless (= (length rep) min-w)
               (error "Padding function %s returned \"%s\", which has length %d, not %d."
                      (or (displayspec-padding-action ds)
                          'db-pad-right)
                      rep (length rep) min-w))
             (setq rep-w min-w))
            ((and max-w (> rep-w max-w))
             ;; The display representation is too long.
             (funcall (or (displayspec-truncation-display-action ds)
                          (lambda (max-w rep rep-w)
                            (put-text-property max-w rep-w
                                               'invisible t
                                               rep)))
                      max-w rep rep-w)
             ;; Assume the truncation function did the right thing.
             (setq rep-w max-w)))
      rep)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Etc.
;;;

(defun database-mode ()
  "A mode for viewing and editing formatted data; a database front end.
In Database Edit mode, fields of the database may be changed.
In Database View mode, keystrokes are bound to database commands.
Typically, if point is on a field, the buffer is in Database Edit mode; if
point is at the beginning of the buffer, the buffer is in Database View mode.
The mode line indicates which mode the buffer is in.

Database View mode key bindings:

\\{database-view-mode-map}

Database Edit mode key bindings:

\\{database-edit-mode-map}"

  (setq major-mode 'database-mode
        mode-name "Database"
        buffer-file-name nil
        mode-line-format '("-"
                           (dbc-database-modified-p "*" "-")
                           (dbf-this-record-modified-p "*" "-")
                           "%*"
                           "-Database: %17b   %[("
                           dbf-minor-mode-name
                           minor-mode-alist
                           " "
                           dbc-index-fraction
                           ")%]"
                           "---"
                           (-3 . "%p")
                           "-%-"))
  (make-local-variable 'require-final-newline)
  (setq require-final-newline nil)
  (auto-save-mode 0)
  (db-view-mode))

(defsubst db-data-display-buffer-p ()
  "T if this buffer is a database data display buffer."
  (eq major-mode 'database-mode))

;;; db-format.el ends here
