;;; db-interfa.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Commands for operating on the current database.

;;; Code:

(eval-when-compile (require 'cl))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

(deflocalvar dbc-database nil
  "The database associated with this format.
This variable is also set in the summary format.")

;;; Database modification

(deflocalvar dbc-database-modified-p nil
  "T if the database has been modified, nil otherwise.
Mirrors the value of the modified-p slot of the database.
This has to be a real variable so it can go in mode-line-format.
Set it using `dbc-set-database-modified-p'.")

(defsubst dbc-update-database-modified-p ()
  (setq dbc-database-modified-p (database-modified-p dbc-database)))
(defsubst dbc-set-database-modified-p (value)
  (database-set-modified-p dbc-database value)
  (setq dbc-database-modified-p value))

;;; Hiding

(deflocalvar dbc-hide-p nil
  "Non-nil if hiding is in effect, nil otherwise.
Use function `dbc-set-hide-p', which works in either a data display buffer or
a summary buffer and sets the variable's value in both, instead of setting
this directly.
Setting this to nil is cheaper than changing the hide function to the empty
one, since no hide bits are recomputed.
This variable is automatically set by the hiding functions.")

(defun dbc-set-hide-p (value)
  "Enable hiding if VALUE is non-nil, otherwise disable it.
This is done in both the data display buffer and summary buffer."
  (db-in-data-display-buffer
    (setq dbc-hide-p value)
    (dbf-in-summary-buffer
      (setq dbc-hide-p value))))

;;; Location in the database

(deflocalvar dbc-link nil
  "The link of the record currently being displayed, or nil.")

(deflocalvar dbc-index nil
  "The index of the record currently being displayed (and of its link), or nil.
Use `dbc-set-index' to set this value unless you know what you are doing.")

(deflocalvar dbc-index-fraction nil
  "A string of the form dbc-index/database-no-of-records.
Variables with numeric values aren't allowed in mode-line-format.
An asterisk (*) precedes dbc-index if the current record is marked.
The fraction is surrounded by square brackets if the current record is hidden.
This variable should only be set by calling `dbc-set-index'.")

;;; Movement behavior

(deflocalvar dbc-wraparound-p 'delay
  "Value t, nil, or 'delay determines whether going forward from the last
record (or backward from the first) wraps, is prohibited, or denies on the
first attempt only and then wraps.")

(deflocalvar dbf-stay-in-edit-mode-p t
  "*Whether edit mode is preserved when switching records in EDB.
Automatically becomes local to the current buffer when set in any fashion.
Only has an effect when set in an EDB data display buffer.")

;;; Etc.

(deflocalvar db-new-record-function nil
  "Function called on empty records before they're inserted in the database.
Takes two arguments, the record and the database.")

(defvar db-kill-buffer-hook-inhibit-p nil
  "Non-nil if `db-kill-buffer-hook' shouldn't do anything.")

(defvar db-mention-filename-on-save-p t
  "Non-nil if a database's print-name or filename should be mentioned when
it is saved to disk.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Keymaps
;;;

(defvar database-view-mode-map nil
  "Keymap for database data display buffer in view mode.")

(unless database-view-mode-map
  (let ((m (make-sparse-keymap)))
    (suppress-keymap m)
    (mapc (lambda (key-def)
            (define-key m (let ((key (car key-def)))
                            (if (consp key)
                                (db-meta-prefix-ify (car key))
                              key))
              (cdr key-def)))
          '(("t" . toggle-truncate-lines)

            ;; Moving around in the database
            ("n"      . db-next-record)
            ("p"      . db-previous-record)
            ("<"      . db-first-record)
            (">"      . db-last-record)
            (("<")    . db-first-record)
            ((">")    . db-last-record)
            ("j"      . db-jump-to-record)
            (" "      . db-next-screen-or-record)
            ("\177"   . db-previous-screen-or-record)
            (("n")    . db-next-record-ignore-hiding)
            (("p")    . db-previous-record-ignore-hiding)
            (("\C-n") . db-next-marked-record)
            (("\C-p") . db-previous-marked-record)

            ;; Changing to edit mode
            ("\t"   . db-first-field)
            (("\t") . db-last-field)
            ("\C-n" . db-first-field)
            ("\C-p" . db-last-field)
            ("e"    . db-first-field)
            ;; These could be db-first-field and db-last-field, but that
            ;; wouldn't fit in: nowhere else are these keystrokes
            ;; inter-field-movement keystrokes.
            ("\C-f" . undefined)
            ("\C-b" . undefined)
            ("\C-a" . undefined)
            ("\C-e" . undefined)
            ;; ("v" . db-view-mode)
            ("\C-v" . db-scroll-up)
            ;; In view-mode, we're at the top of the buffer (not after
            ;; db-next-screen).
            (("v") . db-scroll-down)

            ;; Undoing changes
            ("\C-xu" . db-revert-record)
            ;;("u"   . db-revert-record)
            ("\C-xr" . db-revert-database)

            ;; Adding and removing records
            ("a" . db-add-record)
            ("i" . db-add-record)
            ("d" . db-delete-record)
            ("k" . db-delete-record)
            ("y" . db-yank-record)
            ("o" . db-output-record-to-db)
            ("c" . db-copy-record)

            ;; Searching commands
            (("S")  . db-search)
            (("s")  . db-search)
            ("s"    . db-search)
            ("S"    . db-sort)
            ("\C-s" . db-isearch-forward)
            ("\C-r" . db-isearch-backward)

            ;; Exiting database mode
            ("q" . db-quit)
            ("x" . db-exit)

            ("m" . db-mark-record)

            ("?" . describe-mode)

            ;; Gross key bindings.
            ("O"      . db-hide-record)
            (("o")    . db-hiding-toggle)
            (("O")    . db-hiding-set)
            (("\C-o") . db-toggle-show-hidden-records)

            ("F" . dbf-set-summary-format)
            ("D" . db-summary)          ; mnemonic for Directory
            ("h" . db-summary)          ; mnemonic for Headers
            ("H" . db-summary)          ; mnemonic for Headers

            ("r"  . db-report)
            ("\r" . db-accept-record)

            ("b" . undefined)
            ("f" . undefined)
            ("g" . undefined)
            ("l" . undefined)
            ("w" . undefined)
            ("z" . undefined)))
    (setq database-view-mode-map m)))


(defvar database-edit-mode-map nil
  "Keymap for database data display buffer in edit mode.")

(unless database-edit-mode-map
  (let ((m (make-sparse-keymap)))
    ;; Obviously don't do suppress-keymap on this one; we want to be
    ;; able to edit.  The view-mode commands should be available via C-c
    ;; and many (such as next-record) available via M- commands as well,
    ;; espcially those not ordinarily bound in text mode (eg M-n and
    ;; M-p).
    (mapc (lambda (key-def)
            (define-key m
              (let ((key (car key-def)))
                (if (consp key)
                    (db-meta-prefix-ify (car key))
                  key))
              (cdr key-def)))
          '( ;; Exiting edit mode
            ("\C-c\C-c" . db-view-mode)

            ;; Undoing changes
            ("\C-xU" . db-revert-field)

            ;; Moving from record to record
            (("n") . db-next-record)
            (("p") . db-previous-record)

            ;; Moving from field to field
            ("\t"   . db-next-field)
            (("\t") . db-previous-field)
            (("<")  . db-first-field)
            ((">")  . db-last-field)
            ("\C-v" . db-scroll-up)
            (("v")  . db-scroll-down)

            ;; Movement within a field
            ("\C-n" . db-next-line-or-field)
            ("\C-p" . db-previous-line-or-field)
            ;; almost-the-same-as-before commands
            ("\C-f" . db-forward-char)
            ("\C-b" . db-backward-char)
            (("f")  . db-forward-word)
            (("b")  . db-backward-word)
            ("\C-a" . db-beginning-of-line-or-field)
            ("\C-e" . db-end-of-line-or-field)

            ;; Editing a field
            ;;insertion
            ("\r"     . db-newline)
            ("\n"     . db-newline)
            ("\C-o"   . db-open-line)
            ;;deletion
            ("\C-d"   . db-delete-char)
            ("\177"   . db-backward-delete-char)
            (("d")    . db-kill-word)
            (("\177") . db-backward-kill-word)
            ("\C-k"   . db-kill-line)
            (("k")    . db-kill-to-end)
            ("\C-w"   . db-kill-region)
            (("w")    . db-copy-region-as-kill)

            ;; Other commands
            (("s")   . db-search-field)
            ;;(("S") . db-search-field)

            ("\C-s" . db-isearch-forward)
            ("\C-r" . db-isearch-backward)

            (("?") . db-field-help)))
    (setq database-edit-mode-map m)))


;;; Bindings for both keymaps

(mapc (lambda (&optional key-def)
        (apply 'define-key database-view-mode-map key-def)
        (apply 'define-key database-edit-mode-map key-def))
      '( ;; Saving the database
        ("\C-x\C-s" db-save-database)
        ("\C-x\C-w" db-write-database-file)

        ;; Toggling modifiable-p
        ("\C-x\C-q" db-toggle-modifiable-p)

        ;; Wipe out dangerous commands
        ("\C-xn" undefined)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Menus
;;;

;; Menus by Alastair Burt <burt@dfki.uni-kl.de>,
;;          Michael Ernst <mernst@theory.lcs.mit.edu>
;;          John Overton <overton@cs.uchicago.edu>

(defvar database-view-mode-menu
  '("Database"
    "VIEW Mode:"
    ["Edit"            db-first-field                 t]
    ["Report"          db-report                      t]
    ["Summary"         db-summary                     t]
    "---"
    ["Revert  record" db-revert-record t]
    ["Accept  record" db-accept-record t]
    ["Add     record" db-add-record t]
    ["Copy    record" db-copy-record t]
    ["Delete  record" db-delete-record t]
    ["Output  record" db-output-record-to-db t]
    ["Mark    record" db-mark-record t]
    ["Hide    record" db-hide-record t]
    "----"
    ("Hiding"
     ["Hiding    on/off" db-hiding-toggle             t]
     ["Hiding hide/show (in summary)" db-toggle-show-hidden-records t]
     ["Un-hide      all" db-unhide-all                  t]
     ["Un-mark      all" db-unmark-all                  t]
     ["Mark   un-hidden" db-mark-unhidden-records      t]
     ["Hide   un-marked" db-hide-unmarked-records       t]
     )
    ("Motion"
     ["jump to"  db-jump-to-record t]
     ["first" db-first-record   t]
     ["last"  db-last-record    t]
     "---"
     ["next"               db-next-record                     t]
     ["next (screen)"      db-next-screen-or-record           t]
     ["next (marked)"      db-next-marked-record              t]
     ["next (ignore hiding)" db-next-record-ignore-hiding     t]
     "---"
     ["prev"               db-previous-record                 t]
     ["prev (screen)"      db-previous-screen-or-record       t]
     ["prev (marked)"      db-previous-marked-record          t]
     ["prev (ignore hiding)" db-previous-record-ignore-hiding t]
     )
    "----"
    ["Create Report" db-report t]
    ["Toggle Hiding" db-hiding-toggle t]
    ["Summary" db-summary t]
    "----"
    ["Edit Mode" db-first-field t]
    "----"
    ["Sort    database" db-sort t]
    ["Revert  database" db-revert-database t]
    ["Save    database"  db-save-database       t]
    ["Write   database" db-write-database-file t]
    ["Internal Layout" db-toggle-internal-file-layout t]
    "----"
    ["Quit" db-quit t]
    )
  "Menu for Database View mode.")

;; 'ignored for SYMBOL argument was giving me trouble.
;; Does this work in Lucid Emacs?
(easy-menu-define ignored
  database-view-mode-map
  "ignored-doc-string"
  database-view-mode-menu)


(defvar database-edit-mode-menu
  '("Database"
    "EDIT Mode:"
    ["View mode"        db-view-mode      t]
    ["Report"           db-report         t]
    ["Summary"          db-summary        t]
    ["Summary   subset" db-summary-subset t]
    "---"
    ["Revert    record"  db-revert-record     t]
    "---"
    ["Revert     field"  db-revert-field   t]
    ["Help    on field" db-field-help t]
    ["Search  in field" db-search-field t]
    "---"
    ("Motion"
     ["Next       field"  db-next-field     t]
     ["Prev       field"  db-previous-field t]
     ["Last       field"  db-last-field     t]
     ["First      field"  db-first-field    t]
     ["Next      record" db-next-record t]
     ["Previous  record" db-previous-record t])
    "---"
    ["Revert  database" db-revert-database t]
    ["Search  database"  db-search-field        t]
    ["Save    database"  db-save-database       t]
    ["Write   database" db-write-database-file t]
    "---"
    ["Quit" db-quit t]
    )
  "Menu for Database Edit mode.")

;; 'ignored for SYMBOL argument was giving me trouble.
;; Does this work in Lucid Emacs?
(easy-menu-define ignored
  database-edit-mode-map
  "ignored-doc-string"
  database-edit-mode-menu)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Help
;;;

(defun db-field-help ()
  "Display help for current field using the recordfieldspec help-info field.
If this is a string, display it.  If it is a form, eval it and display the
result."
  (interactive)
  (unless dbf-this-displayspec
    (error "Not on a field."))
  (let* ((fidx (displayspec-record-index dbf-this-displayspec))
         (htxt (recordfieldspec-help-info
                (database-recordfieldspec dbc-database fidx))))
    (db-best-fit-message
     (if htxt
         (if (stringp htxt)
             htxt
           (condition-case err
               (eval htxt)
             (error
              (format
               "This help form:\n\n  %s\n\nfailed with this error:\n\n%s"
               htxt err))))
       (format "No help available for `%s'."
               (db-fname<-fno fidx dbc-database)))
     " *DB Help*")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Quitting
;;;

;; These work in both the data display buffer and the summary buffer, for
;; folks who enjoy spending most of their time in the summary who will rebind
;; its keystrokes to call these functions instead of dbs-exit (for instance).

(defun db-quit ()
  "Quit editing the database for now; bury its buffers."
  (interactive)
  ;; Bury the data display and summary buffers.
  (let (dd sum)
    (db-in-data-display-buffer
     (setq dd  (current-buffer)
           sum (dbf-summary-buffer)))
    (when dd
      (delete-windows-on dd)
      (bury-buffer dd))
    (when sum
      (delete-windows-on sum)
      (bury-buffer sum))))

(defun db-exit (&optional kill)
  "Be done with the database; like `db-quit', but offers to save any changes.
With prefix argument, kills the data display buffer, and the database, if that
was its only data display buffer."
  (interactive "P")
  (db-save-database t)
  (if kill
      (db-kill-buffers)
    (db-quit)))

(defun db-kill-buffers ()
  "Kill this buffer, and the associated summary or data display buffer, if any.
If its last data display buffer is killed, the database is killed too.
Does not offer to save changes to the database or to this record; use `db-exit'
with optional argument to do so."
  (db-in-data-display-buffer
   ;; Kill this data display buffer's associated database summary buffer.
   ;; fixme: clean first? --ttn
   (dbf-in-summary-buffer
    (delete-windows-on (current-buffer))
    (let ((db-kill-buffer-hook-inhibit-p t))
      (kill-buffer nil)))
   (let ((bufs (delq (current-buffer)
                     (database-data-display-buffers dbc-database))))
     (if bufs
         (setf (database-data-display-buffers dbc-database) bufs)
       (setq db-databases (delq dbc-database db-databases))))
   (let ((db-kill-buffer-hook-inhibit-p t))
     (kill-buffer nil))))

(defun db-kill-buffer-hook ()
  ;; Does nothing if db-kill-buffer-hook-inhibit-p is non-nil.
  (cond ((and (not db-kill-buffer-hook-inhibit-p)
              (or (db-data-display-buffer-p)
                  (db-summary-buffer-p)))
         (if (or dbf-this-record-modified-p
                 (dbf-this-field-modified-p))
             (when (y-or-n-p (concat "Commit the current record before killing "
                                     (database-print-name dbc-database)
                                     "? "))
               (dbf-process-current-record-maybe t)
               ;; Ask whether to save the database.
               (db-save-database t))
           ;; Ask whether to save the database.
           (db-save-database t))
         ;; We have asked whether to save the database unless this record was
         ;; modified and the user didn't want to commit it.
         (db-kill-buffers))))

(defun db-save-some-buffers (&optional quietly exiting)
  "Save some modified databases and file-visiting buffers.
Asks user about each one.  With argument, saves all with no questions."
  (interactive "P")
  (db-save-some-databases quietly)
  (save-some-buffers quietly exiting))

;; This isn't quite right because it should modify the ???.
(defun db-save-some-databases (&optional quietly)
  "Save some modified databases.  Asks user about each one.
With argument, saves all with no questions."
  (interactive "P")
  (let (buffers)
    (dolist (db db-databases)
      (setq buffers (database-clean-data-display-buffers db))
      (when buffers
        (dolist (buf buffers)
          (db-in-buffer buf (dbf-process-current-record-maybe t)))
        (db-in-buffer (car buffers) (db-save-database (not quietly)))
        (dolist (buf buffers)
          (with-current-buffer buf
            (force-mode-line-update)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; File I/O
;;;

;;;###autoload
(defun db-find-file (filename &optional promptp)
  "Read a database from FILENAME; prompts when called interactively.
If the database file doesn't specify a format and the format file can't be
inferred from FILENAME, the user is prompted for it too.
The user is always prompted for the format if prefix argument
PROMPTP is non-nil.
If the database is already read in and PROMPTP is nil, the existing
database buffer is merely selected.
When called non-interactively, argument PROMPTP may be a string, the
name of a format file to use."
  (interactive "fDatabase file: \nP")
  ;; fixme: check whether the file is currently read in. --ttn
  (setq filename (expand-file-name filename))
  (let ((format-file (when (stringp promptp) promptp))
        database data-display-buffer)
    (when (stringp promptp)
      (setq promptp nil))

    (unless promptp
      (setq database (db-find-read-in-database filename))
      (when (and database
                 (not (database-clean-data-display-buffers database)))
        (setq database nil))
      ;; Find an appropriate data display buffer
      (when (and database format-file)
        (let ((bufs (database-data-display-buffers database))
              ddb-format-file)
          (while bufs
            (if (db-same-file-p format-file
                                (with-current-buffer (car bufs)
                                  dbf-format-file))
                (setq data-display-buffer (car bufs)
                      bufs nil)
              (setq bufs (cdr bufs)))))))
    (unless database
      ;; Either promptp is non-nil, or we couldn't find an
      ;; appropriate read-in-database.
      (setq database (db-read-database-file filename format-file promptp)))
    (unless data-display-buffer
      (setq data-display-buffer
            (car (database-clean-data-display-buffers database))))
    (switch-to-buffer data-display-buffer)
    (setq dbc-database database))
  (db-first-record))

(defun db-find-read-in-database (filename)
  ;; Return the database most recently read in from FILENAME, or nil.
  (let ((db-ls db-databases)
        result db)
    (while (and (not result) db-ls)
      (setq db (car db-ls)
            result (when (and (db-same-file-p filename (database-file db))
                              (car (database-clean-data-display-buffers db)))
                     db)
            db-ls (cdr db-ls)))
    result))

(defun database-clean-data-display-buffers (database)
  ;; Remove killed buffers from DATABASE's data-display-buffers slot.
  ;; Return a list of the remaining data display buffers.
  ;; If there are none, kill DATABASE as well.
  (let ((bufs (database-data-display-buffers database)))
    (dolist (buf bufs)
      (unless (buffer-name buf)
        (setq bufs (delq buf bufs))))
    ;; Kill the database
    (if (null bufs)
        (progn
          ;; This is a hack.  Avoid calling `dbf-process-current-record-maybe'.
          (let ((dbc-index nil)
                (dbc-database database))
            (db-save-database-helper t nil))
          (setq db-databases (delq database db-databases)))
      (setf (database-data-display-buffers database) bufs))
    bufs))

(defun db-revert-database ()
  "Replace the database with the data on disk.
This undoes all changes since the database was last saved."
  (interactive)
  (when (yes-or-no-p (format "Revert database from file %s? "
                             (database-file dbc-database)))
    (let ((database dbc-database)
          (db-buffer (generate-new-buffer "db-read-database-file")))

      (set-buffer db-buffer)
      (insert-file-contents (database-file database) nil)
      (unless (setq database (db-read-database-internal-file-layout))
        (error "File no longer contains a database"))
      (db-read-database-file-helper db-buffer database)

      (mapc (lambda (buf)
              (set-buffer buf)
              (dbc-update-database-modified-p)
              ;; abandon any changes
              (dbf-set-this-field-modified-p nil)
              (setq dbf-this-record-modified-p nil)
              (db-jump-to-record dbc-index nil))
            (database-clean-data-display-buffers database))

      (db-message "Reverted database from disk"))))


(defun db-save-database (&optional query quietly)
  "Save the database to disk in the default save file.
Any changes to the current record are processed first.
The default save file is the file it was last saved to or read from.
If optional arg QUERY is specified, the user is asked first.
Optional second arg QUIETLY suppresses messages regarding the filename."
  (interactive)
  ;; fixme: check if file is more recent than buffer. --ttn
  (db-in-data-display-buffer
    (dbf-process-current-record-maybe t)
    (db-save-database-helper query quietly)))

(defun db-save-database-helper (query quietly)
  (if dbc-database-modified-p
      (when (or (not query) (y-or-n-p (concat "Save database "
                                              (database-print-name dbc-database)
                                              "? ")))
        (db-write-database-file (database-file dbc-database) quietly))
    (unless quietly
      (let ((name (database-print-name dbc-database)))
        (when (and (stringp name)
                   (or (string= "" name)
                       (string= "Unnamed Database "
                                (substring name 0 (min 17 (length name))))))
          (setq name nil))
        (db-message "No changes need to be saved%s"
                    (if (and db-mention-filename-on-save-p name)
                        (format " in %S" name)
                      ""))))))

(defun db-write-database-file (&optional filename quietly)
  "Save the database to disk in file FILENAME; it becomes the default save file.
Any changes to the current record are processed first.
If FILENAME is not specified, the user is prompted for it.
Optional second arg QUIETLY suppresses messages regarding the filename."
  (interactive)
  ;; Do this before asking for the filename.
  (dbf-process-current-record-maybe t)
  ;; Save even if the database is not modified.
  (unless filename
    (setq filename (read-file-name
                    (format "Save database %s into file: "
                            (database-print-name dbc-database)))))
  (unless (equal filename (database-file dbc-database))
    (setf (database-file dbc-database) filename)
    ;; Rename the buffer.
    (rename-buffer (generate-new-buffer-name
                    (file-name-nondirectory filename))))
  (let ((message-filename (if db-mention-filename-on-save-p
                              (concat " to file " filename)
                            "")))
    (unless quietly (db-message "Saving database%s..." message-filename))
    (db-write-1 (db-in-data-display-buffer buffer-file-coding-system)
                dbc-database filename)
    (unless (or quietly db-io-error-p)
      (db-message "Saving database%s...done" message-filename))))

(defun db-toggle-internal-file-layout (&optional arg)
  "Toggle whether the database will be saved in EDB's internal file layout.
With a nonzero prefix argument, set it to use internal file layout.
With a zero prefix argument, set it not to use internal file layout."
  (interactive "P")
  (setf (database-internal-file-layout-p dbc-database)
        (if arg
            (not (zerop (prefix-numeric-value arg)))
          (not (database-internal-file-layout-p dbc-database))))
  (when (interactive-p)
    (message "Use of internal file layout now %sabled"
             (if (database-internal-file-layout-p dbc-database)
                 "en" "dis"))))

(defun db-toggle-modifiable-p (&optional arg)
  "Toggle whether the database may be modified by the user.
With a nonzero prefix argument, set it modifiable.
With a zero prefix argument, set it non-modifiable."
  (interactive "P")
  (let ((modifiable-p (if arg
                          (not (zerop (prefix-numeric-value arg)))
                        (not (database-modifiable-p dbc-database)))))
    (setf (database-modifiable-p dbc-database) modifiable-p)
    ;; fixme: handle mods in the presence of nil `modifiable-p'. --ttn
    (db-in-data-display-buffer
      (when (eq dbf-minor-mode 'edit)
        (setq buffer-read-only (not modifiable-p))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Record selection
;;;

;; This is the common body of `db-next-record' and `db-jump-to-record'.
(defmacro dbf-goto-record-internal (&rest record-selector-body)
  (declare (indent 0) (debug body))
  `(progn
     ;; NOTE: This line will cause problems if some hook function
     ;;       deliberately raises an error (like Joe Wells' do).
     (db-view-mode)
     (dbf-process-current-record-maybe nil)
     ,@record-selector-body
     (setq dbf-this-record-original (link-record dbc-link))
     (db-display-record (dbf-displayed-record) t)))

(defun db-jump-to-record (arg &optional respect-hiding)
  "Show the database's ARGth record.
Hiding is ignored unless optional argument RESPECT-HIDING is specified."
  ;; NOTE: Respect and ignorance have opposite nature.
  (interactive "NJump to record number: ")
  (db-in-data-display-buffer
    (dbf-goto-record-internal
      (db-select-record arg (not respect-hiding))))
  (when (db-summary-buffer-p)
    (dbs-synch-summary-with-format)))

(defun db-first-record (&optional ignore-hiding)
  "Show the database's first record.
With optional prefix argument, ignores hiding."
  (interactive "P")
  (cond ((db-data-display-buffer-p)
         (db-jump-to-record 1 (not ignore-hiding)))
        ((db-summary-buffer-p)
         ;; fixme: handle hiding. --ttn
         (goto-char (point-min))
         (db-jump-to-point))
        (t
         (error "db-first-record called in wrong context."))))

(defun db-last-record (&optional ignore-hiding)
  "Show the database's last record.
With optional prefix argument, ignores hiding."
  (interactive "P")
  (cond ((db-data-display-buffer-p)
         (db-jump-to-record (database-no-of-records dbc-database)
                            (not ignore-hiding)))
        ((db-summary-buffer-p)
         (goto-char (point-max))
         (db-jump-to-point))
        (t
         (error "db-last-record called in wrong context"))))

(defun db-next-record (arg &optional ignore-hiding markedp)
  "Go to the ARGth next record.
In that record, go to the current field, if any."
  (interactive "p")
  (when (db-summary-buffer-p)
    (dbs-synch-format-with-summary))
  (db-in-data-display-buffer
    (let ((this-field-index dbf-this-field-index))
      (dbf-goto-record-internal
        (db-select-next-record arg ignore-hiding markedp))
      ;; If in edit mode, stay in edit mode in the same field.
      (when (and this-field-index
                 dbf-stay-in-edit-mode-p)
        (db-move-to-field-exact this-field-index))))
  (when (db-summary-buffer-p)
    (let ((index (dbs-in-data-display-buffer dbc-index)))
      ;; This might not be right, depending on what records are summarized.
      (dbs-forward-record (- index dbs-index))
      (dbs-set-index index))))

(defsubst db-previous-record (arg &optional ignore-hiding markedp)
  "Go to the ARGth previous record.
In that record, go to the current field, if any."
  (interactive "p")
  (db-next-record (- arg) ignore-hiding markedp))

(defsubst db-next-record-ignore-hiding (arg)
  "Go to the ARGth next record, ignoring omissions.
That is, all records, even those which are hidden, are counted."
  (interactive "p")
  (db-next-record arg t))

(defsubst db-previous-record-ignore-hiding (arg)
  "Go to the ARGth previous record, ignoring omissions.
That is, all records, even those which are hidden, are counted."
  (interactive "p")
  (db-next-record-ignore-hiding (- arg)))

(defsubst db-next-marked-record (arg)
  "Go to the ARGth next marked record.
Hidden records are treated according to db-hide-p."
  (interactive "p")
  (db-next-record arg nil t))

(defsubst db-previous-marked-record (arg)
  "Go to the ARGth previous marked record.
Hidden records are treated according to db-hide-p."
  (interactive "p")
  (db-next-marked-record (- arg)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Moving from record to record (setting dbc-link)
;;;

;; These don't display, but they do set `dbc-link' and `dbc-index'.

;; Don't forget that when moving off a record, must check whether it has
;; been modified and, if so, call an update function.

(defun dbc-set-index (index)
  (setq dbc-index index
        dbc-index-fraction
        (let ((frac (format "%s%d/%d"
                            (if (link-markedp dbc-link) "+" "")
                            dbc-index
                            (database-no-of-records dbc-database))))
          (if (and dbc-hide-p (link-hiddenp dbc-link))
              (concat "[" frac "]")
            frac))))


(defun db-select-next-record (arg &optional ignore-hiding markedp)
  ;; Advance to the ARGth following record.  Does no display.
  (interactive "p")
  ;; fixme: check for inifinite loop. --ttn
  (let ((lil (db-next-link-and-index    ; link-index-list
              dbc-database dbc-link dbc-index
              arg (and dbc-hide-p (not ignore-hiding))
              markedp
              (or (eq dbc-wraparound-p t)
                  (and (eq dbc-wraparound-p 'delay)
                       (eq last-command 'db-next-record-failed))))))
    (setq dbc-link (car lil))
    (dbc-set-index (car (cdr lil)))
    (when (cdr (cdr lil))
      (setq this-command 'db-next-record-failed)
      (if (< (car (cdr (cdr lil))) 0)
          (db-message "First record")
        (db-message "Last record")))))

(defsubst db-select-prev-record (arg &optional ignore-hiding)
  ;; Advance to the ARGth previous record.  Does no display.
  (interactive "p")
  (db-select-next-record (- arg) ignore-hiding))

(defun db-select-first-record (&optional ignore-hiding)
  ;; Select first record.  Does no display.
  ;; If hiding is in effect, select the first unhidden record, unless
  ;; optional argument IGNORE-HIDING is non-nil.
  (interactive)
  (setq dbc-link (database-first-link dbc-database))
  (if (and dbc-hide-p (link-hiddenp dbc-link) (not ignore-hiding))
      (progn
        (setq dbc-index 1)
        (db-select-next-record 1))
    (dbc-set-index 1)))

(defsubst db-select-last-record (&optional ignore-hiding)
  ;; Select last record.  Does no display.
  ;; If hiding is in effect, select the last unhidden record, unless
  ;; optional argument IGNORE-HIDING is non-nil.
  (interactive)
  (setq dbc-link (database-first-link dbc-database)
        dbc-index 1)
  (let ((dbc-wraparound t))
    (db-select-next-record -1 ignore-hiding)))

(defun db-select-record (arg &optional ignore-hiding)
  ;; Select record ARG.  Does no display.
  ;; If record ARG is hidden, selects the first following non-hidden record,
  ;; unless optional argument IGNORE-HIDING is non-nil.
  (interactive "nRecord number: ")
  (let ((tot (database-no-of-records dbc-database)))
    (if (and (> arg 0) (<= arg tot))
        (progn
          (db-select-first-record ignore-hiding)
          (db-select-next-record (1- arg) ignore-hiding))
      (beep)
      ;; This should test on the number of unhidden records, if
      ;; hiding is on and IGNORE-HIDING is non-nil.
      (db-message "Record number %d out of range 1..%d" arg tot))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Hybrid field/record movement commands
;;;

(defun db-next-screen-or-record (arg)
  "Go to the ARGth next screenful of this display, or to the ARGth
next record, if this is the last screenful of this display.
If point is in the summary buffer and the data display buffer is not visible,
then move to the next record."
  (interactive "p")
  (cond ((db-data-display-buffer-p)
         (dbf-next-screen-or-record arg))
        ((db-summary-buffer-p)
         (let ((w (get-buffer-window dbs-data-display-buffer)))
           (if w
               (progn
                 (with-selected-window w
                   (dbf-next-screen-or-record arg))
                 (dbs-synch-summary-with-format))
             (db-next-record arg))))))

(defun dbf-next-screen-or-record (arg)
  (let ((pmax (point-max)))
    (if (= pmax (window-end))
        (db-next-record arg)
      (while (and (> arg 0) (not (= pmax (window-end))))
        (scroll-up nil)
        (decf arg)))))

(defun db-previous-screen-or-record (arg)
  "Go to the ARGth previous screenful of this display, or to the ARGth
previous record, if this is the first screenful of this display.
If point is in the summary buffer and the data display buffer is not visible,
then move to the previous record."
  (interactive "p")
  (cond ((db-data-display-buffer-p)
         (dbf-previous-screen-or-record arg))
        ((db-summary-buffer-p)
         (let ((w (get-buffer-window dbs-data-display-buffer)))
           (if w
               (progn
                 (with-selected-window w
                   (dbf-previous-screen-or-record arg))
                 (dbs-synch-summary-with-format))
             (db-previous-record arg))))))

(defun dbf-previous-screen-or-record (arg)
  (let ((pmin (point-min)))
    (if (= pmin (window-start))
        (db-previous-record arg)
      (while (and (> arg 0) (not (= pmin (window-start))))
        (scroll-down nil)
        (decf arg))
      (when (= pmin (window-start))
        (goto-char (point-min))))))


(defun db-beginning-of-field-or-record ()
  "Move to the beginning of this field; if at its beginning, to the first field."
  (interactive)
  (if (= (point) dbf-this-field-beginning-pos)
      (db-first-field)
    (db-beginning-of-field)))

(defun db-end-of-field-or-record ()
  "Move to the end of this field; if at its end, to the last field."
  (interactive)
  (if (= (point) (dbf-this-field-end-pos))
      (db-last-field)
    (db-end-of-field)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Adding and deleting records
;;;

(defun db-add-record ()
  "Add a new record to the database immediately before the current record."
  (interactive)
  (when (db-summary-buffer-p)
    (pop-to-buffer dbs-data-display-buffer))
  (let ((new (make-record dbc-database)))
    (dotimes (fno (database-no-of-fields dbc-database))
      (record-set-field-from-index
       new fno
       (recordfieldspec-default-value
        (database-recordfieldspec dbc-database fno))
       nil))
    (when db-new-record-function
      (funcall db-new-record-function new dbc-database))
    (database-add-record new dbc-database dbc-index))
  (incf dbc-index)
  ;; Why doesn't this need to be (dbc-set-modified-p t)?
  (database-set-modified-p dbc-database t)
  ;; Probably unnecessary, as `database-add-record' has done the trick.
  (dbf-set-summary-out-of-date-p)
  ;; fixme: update incrementally instead of fully. --ttn
  (when (and dbf-summary-buffer (get-buffer-window dbf-summary-buffer))
    (dbf-in-summary-buffer
     (dbs-synch-summary-with-format)))
  (db-message "Added a new record")
  (db-previous-record 1)
  ;; Begin editing the new record.  Don't use `db-edit-mode'.
  (db-first-field))

(defun db-delete-record (&optional force)
  "Remove the current record from the database.
With a prefix argument, doesn't verify."
  (interactive "P")

  (when (or force (y-or-n-p "Delete this record? "))
    (setq dbc-deleted-record (link-record dbc-link))
    (if (eq (link-next dbc-link) dbc-link)
        ;; This is the only link in the dbc-database.
        (setf (database-first-link dbc-database) nil)
      (when (eq dbc-link (database-first-link dbc-database))
        (setf (database-first-link dbc-database)
              (link-next dbc-link)))
      (db-join-links (link-prev dbc-link) (link-next dbc-link)))
    (decf (database-no-of-records dbc-database))
    ;; set some links-changed variable, or update the summary directly
    (decf dbc-index)
    (dbc-set-database-modified-p t)
    (db-message "Record deleted")
    (db-next-record 1)))

;; This should be put with the rest of the variables.
(defvar dbc-deleted-record nil
  "The record most recently deleted by \\[db-delete-record].")

(defun db-yank-record (endp)
  "Insert, and make current, the most recently deleted record.
The deleted record is inserted before the current record.
With prefix argument ENDP, insert at end of database and don't select it."
  (interactive "P")
  (unless dbc-deleted-record
    (error "No deleted record to yank."))
  (db-in-data-display-buffer
    ;; This is inelegant in the extreme, but the interaction of
    ;; `dbc-set-index' and db-{previous-next}-record and
    ;; `database-add-record' mystifies me.  --karl@owl.hq.ileaf.com (Karl Berry)
    (if endp
        (progn
          (database-add-record dbc-deleted-record dbc-database)
          ;; We go back to the current record below.
          (db-next-record 1))
      ;; Why does this progn contain only two forms?!  --ttn
      (progn
        (database-add-record dbc-deleted-record dbc-database dbc-index)
        (dbc-set-index (1+ dbc-index)))
      (db-previous-record 1)
      (dbc-set-database-modified-p t)))
  (when (db-summary-buffer-p)
    (dbs-synch-summary-with-format))
  (force-mode-line-update)
  (db-message "Record yanked"))

(defun db-copy-record (&optional arg)
  "Insert a copy of the current record in the database immediately after it.
The second of the two records is made the current record.
With a prefix argument, inserts that many copies."
  (interactive "p")
  (db-in-data-display-buffer
    (dbf-process-current-record-maybe t)
    (while (> arg 0)
      (database-add-record (copy-sequence (link-record dbc-link))
                           dbc-database dbc-index)
      (dbc-set-index (1+ dbc-index))
      (decf arg)))
  (when (db-summary-buffer-p)
    (dbs-synch-summary-with-format))
  (force-mode-line-update)
  (db-message "Record copied"))

(deflocalvar db-for-output nil
  "Default database to which to output records.")

(defun db-output-record-to-db (database)
  "Copy (output) the current record to DATABASE.
DATABASE must be read in and compatible with the current database."
  ;; Make a list of databases compatible with this one.
  (interactive
   (list
    (let ((db-alist (delq nil (mapcar
                               (lambda (db)
                                 (when (and (not (eq db dbc-database))
                                            (databases-compatible
                                             db dbc-database))
                                   (cons (or (database-print-name db)
                                             (database-file db))
                                         db)))
                               db-databases))))
      (if db-alist
          (cdr (assoc (completing-read
                       "Output record to which database (? for choices): "
                       db-alist nil t db-for-output)
                      db-alist))
        (error "No compatible databases are currently read in!")
        nil))))
  (when database
    (when (db-summary-buffer-p)
      (dbs-synch-format-with-summary))
    (db-in-data-display-buffer
      (database-add-record (link-record dbc-link) database))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sorting
;;;

(defun db-sort (&optional dont-confirm)
  "Sort the database.  With a prefix argument, don't confirm the sort order."
  (interactive "P")
  (db-in-data-display-buffer
    (dbf-process-current-record-maybe t)
    (if dont-confirm
        (progn
          (database-sort dbc-database)
          (dbf-finished-sorting))
      (database-sort-interface dbc-database))))

(defun dbf-finished-sorting ()
  ;; Need to recompute the current record's index.
  (db-lmap-macro
   (when (eq db-lmap-link dbc-link)
     (dbc-set-index db-lmap-index)
     (db-lmap-break))
   dbc-database)

  (delete-windows-on (dbf-summary-buffer))

  ;; Force summary refresh.  Set the summary buffer out of order but don't set
  ;; the must-recompute-something bit.
  (dbf-in-summary-buffer (setq dbs-no-of-records -1))

  ;; The index shown in the mode line is correct, but the database may have
  ;; been marked as modified, and that change hasn't made it to the mode line.
  (dbc-update-database-modified-p)
  (force-mode-line-update))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Editing
;;;

(defun db-field-query-replace ()
  "Replace some instances of a value in this field with some other value.
Confirms before each replacement."
  (interactive)
  (if (not dbf-this-field-index)
      (error "Call this when on a field."))
  (let* ((cur-idx  dbc-index)
         (dispspec dbf-this-displayspec)
         (fsno     dbf-this-field-index)
         (ridx     (displayspec-record-index dispspec))
         (fname    (db-fname<-fno ridx dbc-database))
         (ordfunc  (db-rs-ordfunc (database-recordfieldspec dbc-database ridx)))
         ov                             ; original value
         ov-printed
         rv                             ; replacement value
         rv-printed)
    (dbf-process-current-record-maybe nil)
    (setq ov (db-display->actual-call
              (displayspec-display->actual dispspec)
              (read-string "Query replace: ")
              ;; No previous value or record.
              nil nil
              ridx))
    (db-check-constraint ov nil ridx dbc-database)
    ;; Must keep in mind that this is not necessarily what the user typed.
    (setq ov-printed (db-actual->display-call
                      (displayspec-actual->display dispspec)
                      ov nil ridx))

    (setq rv (db-display->actual-call
              (displayspec-display->actual dispspec)
              (read-string (format "Query replace %s with: "
                                   ov-printed))
              nil nil ridx))
    (db-check-constraint rv nil ridx dbc-database)
    (setq rv-printed (db-actual->display-call
                      (displayspec-actual->display dispspec)
                      rv nil ridx))

    (maprecords (lambda (record)
                  (when (= 0 (funcall ordfunc ov (aref record ridx)))
                    (db-display-record record t)
                    (db-skip-string-forward (aref dbf-inter-field-text 0))
                    (setq dbf-this-field-index 0)
                    (db-next-field-internal fsno)
                    ;; fixme: handle strings too big for minibuffer. --ttn
                    (when (y-or-n-p (format "(%s) Replace `%s' with `%s'? "
                                            fname ov-printed rv-printed))
                      ;; It's a bit extreme that this errs if the value
                      ;; fails to meet the constraint.
                      (record-set-field-from-index record ridx rv
                                                   dbc-database))))
                dbc-database)
    (db-message "Replacement done")
    (db-jump-to-record cur-idx t)))

(defun db-accept-record ()
  "Install the current record in the database; make any changes permanent."
  (interactive)
  (dbf-process-current-record-maybe t))
(defalias 'db-commit-record 'db-accept-record)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Searching
;;;

(defun db-search-field (pattern &optional mark)
  "Search for occurrences of PATTERN in the current field of any record.
Finds the first match after the current record; wraps around automatically.
With prefix argument, marks all matches in addition to going to the first one.
If hiding is in effect, hidden records are ignored."
  (interactive
   (list (let ((fieldname (dbf-this-field-name)))
           (read-string
            (format "Search in %s for: " fieldname)
            (if dbf-this-field-index
                (aref dbf-field-search-defaults dbf-this-field-index)
              (error "Only call db-search-field when on a field in a data display buffer."))))
         current-prefix-arg))
  (if (equal "" pattern)
      (error "You didn't enter a pattern for which to search."))

  (let* ((pat (db-parse-match-pattern pattern dbf-this-displayspec))
         (pdisp (db-print-match-pattern pat dbf-this-displayspec))
         (ridx (displayspec-record-index dbf-this-displayspec))
         (rs (database-recordfieldspec dbc-database ridx))
         (this-field-index dbf-this-field-index)
         this-field
         (fname (dbf-this-field-name))
         ;; success is t if we've already found some match.
         ;; The idea is that we'll move to success-link when we're done with
         ;; the search; if success is nil then we're looking for such a link.
         ;; This is either because we haven't found one or because we have
         ;; only found one before dbc-link in the database.
         success success-link success-index
         (matches 0))
    (aset dbf-field-search-defaults dbf-this-field-index pdisp)
    (if mark
        (db-message "Marking all %s in %s..." pdisp fname)
      (db-message "Searching in %s for %s..." fname pdisp))
    (db-lmap-macro
     (progn
       (when (or mark (not success))
         (setq this-field (aref (link-record db-lmap-link) ridx))
         (when (db-match pat this-field rs)
           (if (not success)
               (setq success-link db-lmap-link
                     success-index db-lmap-index
                     success t))
           (when mark
             (incf matches)
             (setf (link-markedp db-lmap-link) t))))
       ;; We're looking for a match in some record besides the displayed
       ;; one and, preferrably, after it.  This permits the first success
       ;; succeeding the current record to overwrite the first success
       ;; preceding the current record.  This means that searches can't
       ;; abort after a success, since that success might be before the
       ;; current record.
       (when (eq db-lmap-link dbc-link)
         (setq success nil)))
     dbc-database
     dbc-hide-p)
    (if success-index
        (if (eq dbc-link success-link)
            (db-message "This record has the only match for %s" pdisp)
          ;; This takes care of committing any changes to the current record.
          (dbf-goto-record-internal
           (setq dbc-link success-link)
           (dbc-set-index success-index))
          (db-move-to-field-exact this-field-index)
          (if mark
              (progn (dbf-set-summary-out-of-date-p)
                     (db-message "Searching for %s...marked %s matches"
                                 pdisp matches))
            (db-message "Searching for %s...found" pdisp)))
      (db-message "Couldn't find a match in %s for %s" fname pdisp))))


(defun db-search ()
  "`db-search' is not yet implemented; use `db-search-field' instead.
In a future version of EDB, `db-search' will permit searching on all fields
of a record simultaneously."
  (interactive)
  (error "db-search is unimplemented; use db-search-field instead (M-s from Edit mode)."))


;; fixme: convert to after-command hooks. --ttn

(defun db-isearch-forward ()
  "Like isearch-forward, but maintains the correspondence between the format
and summary buffers."
  (interactive)
  (isearch-forward)
  (db-jump-to-point))

(defun db-isearch-backward ()
  "Like isearch-backward, but maintains the correspondence between the format
and summary buffers."
  (interactive)
  (isearch-backward)
  (db-jump-to-point))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Hiding
;;;

(defun db-mark-record (&optional arg)
  "Toggle whether the current record is marked.
With a nonzero prefix argument, set it to be marked.
With a zero prefix argument, set it to be unmarked."
  (interactive "P")
  (db-in-data-display-buffer
    (setf (link-markedp dbc-link) (if arg
                                      (not (zerop (prefix-numeric-value arg)))
                                    (not (link-markedp dbc-link))))
    (dbf-update-summary-item dbc-index dbc-link)
    (dbc-set-index dbc-index)           ; sets dbc-index-fraction
    (force-mode-line-update)))

(defun db-hide-record (&optional arg)
  "Change whether the current record is hidden.
With a nonzero prefix argument, set it to be hidden.
With a zero prefix argument, set it to be unhidden."
  (interactive "P")
  (db-in-data-display-buffer
    (setf (link-hiddenp dbc-link) (if arg
                                      (not (zerop (prefix-numeric-value arg)))
                                    (not (link-hiddenp dbc-link))))
    (if dbc-hide-p
        (dbf-update-summary-item dbc-index dbc-link)
      ;; Automatically turn on the effect of hiding.
      (dbc-set-hide-p t)
      ;; Update all marks, since potentially all have to be displayed now.
      (dbf-update-summary-marks))
    (dbc-set-index dbc-index)           ; sets `dbc-index-fraction'
    (force-mode-line-update)))

(defun db-mark-searched-records ()
  "Mark all records found in search STRING of FIELD."
  (interactive)
  (setq current-prefix-arg "1")
  (call-interactively 'db-search-field "1")
  (save-window-excursion
    (switch-to-buffer-other-window (db-summary-buffer))
    (when (db-summary-buffer-p)
      (dbs-synch-format-with-summary))))


(defun db-hide-unmarked-records ()
  "Hide all unmarked records.  Also, clear mark bits and enable hiding."
  (interactive)
  (db-in-data-display-buffer
    (db-lmap-macro
     (if (link-markedp db-lmap-link)
         (setf (link-markedp db-lmap-link) nil)
       (setf (link-hiddenp db-lmap-link) t))
     dbc-database
     t)
    (dbc-set-hide-p t)
    (dbf-update-summary-marks)
    ;; fixme: redisplay? --ttn
    ))

(defun db-mark-unhidden-records ()
  "Mark all unhidden records.  Also clears all hide bits."
  (interactive)
  (db-in-data-display-buffer
    (db-lmap-macro
     (if (link-hiddenp db-lmap-link)
         (setf (link-hiddenp db-lmap-link) nil)
       (setf (link-markedp db-lmap-link) t))
     dbc-database)
    (dbc-set-hide-p t)
    (dbf-update-summary-marks)
    ;; fixme: redisplay? --ttn
    ))

(defun db-unhide-all ()
  "Clear the hide bit of every record."
  (interactive)
  (db-in-data-display-buffer
    (db-lmap-macro
     (setf (link-hiddenp db-lmap-link) nil)
     dbc-database)
    (dbc-set-index dbc-index)           ; sets `dbc-index-fraction'
    (force-mode-line-update)
    (dbf-update-summary-marks)))

(defun db-unmark-all ()
  "Clear the mark bit of every record."
  (interactive)
  (db-in-data-display-buffer
    (db-lmap-macro
     (setf (link-markedp db-lmap-link) nil)
     dbc-database)
    (dbc-set-index dbc-index)           ; sets `dbc-index-fraction'
    (force-mode-line-update)
    (dbf-update-summary-marks)))

(defun db-hiding-toggle (&optional arg)
  "Change whether hiding is in effect.
With a nonzero prefix argument, turn hiding on.
With a zero prefix argument, turn hiding off.

This does not change the current hide-function, and a hide bit is always
computed for each record, but hide bits have no effect on any operations
if hiding is not in effect."
  (interactive "P")
  (db-in-data-display-buffer
    (dbc-set-hide-p (if arg
                        (not (zerop (prefix-numeric-value arg)))
                      (not dbc-hide-p)))
    ;; Refill summary buffer whenever displayed set of records changes,
    ;; including when switching to no hiding and showing hidden records.
    (cond
     ((not dbf-summary-show-hidden-records-p)
      ;; If the hidden records weren't being shown, the records that
      ;; should be displayed in the summary buffer just changed.
      (dbf-fill-summary-buffer-and-move-to-proper-record))
     (t
      (dbf-update-summary-marks)))
    (force-mode-line-update)
    (db-message "Hiding is now %sin effect" (if dbc-hide-p "" "not "))))

(defun db-hiding-set ()
  "Set the criteria for automatically determining whether to hide a record.
This isn't implemented yet."
  (interactive)
  (error "db-hiding-set is not yet implemented."))

(defun db-toggle-show-hidden-records (&optional arg)
  "Toggle whether hidden records are shown in the summary.
With a nonzero prefix argument, show hidden records in the summary.
With a zero prefix argument, don't show hidden records in the summary."
  (interactive "P")
  (db-in-data-display-buffer
    (setq dbf-summary-show-hidden-records-p
          (if arg
              (not (zerop (prefix-numeric-value arg)))
            (not dbf-summary-show-hidden-records-p)))
    (if dbf-summary-show-hidden-records-p
        ;; If we weren't showing hidden records, we might as well start from
        ;; scratch in filling the summary buffer.
        (dbf-fill-summary-buffer-and-move-to-proper-record)
      (dbf-in-summary-buffer
        (let ((buffer-read-only nil))
          (goto-char (point-min))
          (delete-matching-lines "^.\\[")
          (dbs-move-to-proper-record))))
    (if dbf-summary-show-hidden-records-p
        (db-message "Hidden records will now be shown")
      (db-message "Hidden records will not now be shown"))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reporting
;;;

(defun db-report (filename &optional markedp)
  "Create a report according to format specifications in FILENAME.
Prefix argument MARKEDP, if non-nil, means report on only marked records.
If hiding is in effect, hidden records are not reported upon.
When called interactively, prompt for FILENAME."
  (interactive "fReport format file: \nP")
  (dbf-process-current-record-maybe t)
  (let ((database dbc-database)
        fmt rfunc)
    (save-window-excursion
      (set-buffer (get-buffer-create " *Database work buffer*"))
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert-file-contents filename)
      (setq fmt (buffer-substring (point-min) (point-max))))
    (let ((lasfl (db-format->lines/sforms fmt dbc-database nil nil t)))
      (setq rfunc `(lambda (formatted-record)
                     (insert ,@(cdr lasfl)))))
    (let ((hide-p dbc-hide-p))
      (switch-to-buffer (get-buffer-create "*Database Report*"))
      (setq buffer-read-only nil)
      (erase-buffer)
      (db-lmap-macro
       (when (or (not markedp) (link-markedp db-lmap-link))
         (funcall rfunc (link-record db-lmap-link)))
       database
       hide-p)
      (goto-char (point-min)))))

;;; db-interfa.el ends here
