;;; db-rep.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Representation and basic operations for
;; database, link, recordfieldspec objects.

;;; Code:


(require 'db-util)
(eval-when-compile (require 'cl))
(provide 'db-rep)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Database abstraction
;;;

;; A database is just a doubly-linked circular list, with some supporting
;; information.

(defvar databases-made 0)

(defstruct database
  (print-name (format "Unnamed Database %d" (incf databases-made)))

  first-link                            ; pointer to a link
  ;; this could be "size" or "length"
  no-of-records                         ; integer.  First link is numbered 1.

  file                                  ; filename
  file-local-variables                  ; string, the local variables section
                                        ; of the file from which this db came.
  aux-file                              ; filename
  data-display-buffers                  ; list of buffers
  default-format-file                   ; filename

  ;; This does not appear to be used anywhere at all!
  hide-functions

  ;; field information
  no-of-fields                          ; integer
  fieldnames                            ; this is repeated in the recordpsecs
  fieldname-alist                       ; associate NAME with NUMBER
  recordfieldspecs                      ; vector of symbols or recordfieldspecs
                                        ; if symbol, search db-recordfieldtypes

  field-priorities                      ; maybe call this order-fields instead
  hidden-to-end-p                       ; boolean

  ;; For file i/o
  internal-file-layout-p
  (record-sepinfo (make-sepinfo))
  (field-sepinfo (make-sepinfo))
  (alternative-sepinfo (make-sepinfo))
  read-record-from-region
  write-region-from-record
  sub-fieldsep-string
  sub-recordsep-string

  substitutions                         ; associate ACTUAL with STORED strings

  modified-p
  modifiable-p

  locals                                ; associate SYMBOL with VALUE
  )

;;; Accessors

(defun database-set-modified-p (db val)
  (unless (eq val (database-modified-p db))
    (setf (database-modified-p db) val)
    ;; Reflect the change in each data display buffer.
    (mapc (lambda (ddb)
            (with-current-buffer ddb
              (setq dbc-database-modified-p val)
              (set-buffer-modified-p val)
              (force-mode-line-update)))
          (database-clean-data-display-buffers db))))

;;; Non-primitive accessors

(defsubst database-empty-p (database)
  (null (database-first-link database)))

;;; Database-local variables

(defun database-make-local (symbol database &optional value)
  "Declare a database-local variable named by SYMBOL for DATABASE.
Each such variable should only be declared once.
If optional argument VALUE is specified, the variable is set to it."
  (let ((lookup (assq symbol (database-locals database))))
    (if lookup
        (error "%s is already defined as a local variable in %s."
               symbol (database-print-name database))
      (push (cons symbol value) (database-locals database)))))

(defun database-set-local (symbol database value &optional no-error)
  "Set the value of database-local variable SYMBOL, in DATABASE, to VALUE.
SYMBOL must have been declared by a previous call to `database-make-local'
unless optional argument NO-ERROR is supplied, in which case the function
does that automatically, if necessary."
  (let ((lookup (assq symbol (database-locals database))))
    (if lookup
        (setcdr lookup value)
      (if no-error
          (database-make-local symbol database value)
        (error "%s is not a database-local variable for %s."
               symbol (database-print-name database))))))

(defun database-get-local (symbol database &optional no-error)
  "Return the value of database-local variable SYMBOL for DATABASE.
If SYMBOL was not declared by a previous call to `database-make-local',
an error is signalled unless optional argument NO-ERROR is non-nil,
in which case nil is returned."
  (let ((lookup (assq symbol (database-locals database))))
    (cond (lookup
           (cdr lookup))
          (no-error
           nil)
          (t
           (error "%s is not a database-local variable for %s."
                  symbol (database-print-name database))))))

(defun database-local-p (symbol database)
  "Return non-nil if SYMBOL is a database-local variable for DATABASE."
  (assq symbol (database-locals database)))


;;; Non-primitive setters

(eval-when-compile (defvar db-default-field-type))

(defun database-set-fieldnames-to-list (database fspecs &optional dtype)
  "Set DATABASE's fieldnames and record field types according to FSPECS.
But do nothing if DATABASE's `fieldnames' slot is already set.

FSPECS is a list, each element of which can either be a single symbol,
the field name, or a cons (NAME . RECORDFIELDTPYE).  Optional third arg
DTYPE specifies the default recordfieldtype for single-symbol elements.
If DTYPE is not given, use the value of `db-default-field-type' (NOTE:
this variable is deprecated) if bound, or `string' otherwise."
  (unless (database-fieldnames database)
    (let* ((len (length fspecs))
           (fno 0)
           (rs  (make-vector len nil))
           (vec (make-vector len nil))
           ;; We used to deflocalvar `db-default-field-type', but that was a
           ;; bug; a field's type is a property of the field, not of its
           ;; display, not of a buffer.  Generally, the concepts of "local"
           ;; and "default" together are difficult to explain w/o detailed
           ;; knowledge of the implementation and its precise timing of
           ;; events.  Better to specify the default together with the
           ;; non-defaults, such as is now possible w/ DTYPE, and let this
           ;; function worry about making the correct associations.
           ;;
           ;; For EDB 1.x, although we are obligated to check for use of this
           ;; var, we are free to drop the deflocalvar, as long as we explain
           ;; ourselves clearly in the documentation.
           (default-type (cond (dtype)
                               ((and (boundp 'db-default-field-type)
                                     db-default-field-type))
                               (t 'string))))
      (setf (database-fieldname-alist database)
            (mapcar (lambda (fname)
                      (let ((type (if (consp fname)
                                      (prog1 (cdr fname)
                                        (setq fname (car fname)))
                                    default-type)))
                        (if (gethash type db-recordfieldtypes)
                            (aset rs fno type)
                          (error "bad type %s" type)))
                      (aset vec fno fname)
                      (cons fname (prog1 fno (incf fno))))
                    fspecs))
      (setf (database-no-of-fields database) len
            (database-fieldnames database) vec
            (database-recordfieldspecs database) rs))))

;;; Not quite so basic functions.

(defun database-full-fieldsep-string (database)
  ;; Return the string that really separates database record fields.
  (if (database-write-region-from-record database)
      nil
    (sepinfo-sep-string (database-field-sepinfo database))))

(defun database-full-recordsep-string (database)
  ;; Return the string that really separates database records.
  (let ((rsep (database-record-sepinfo database)))
    (if (database-write-region-from-record database)
        (sepinfo-sep-string rsep)
      (let ((fsep (database-field-sepinfo database)))
        (concat (sepinfo-post-last-string fsep)
                (sepinfo-sep-string rsep)
                (sepinfo-pre-first-string fsep))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Link abstraction
;;;

(defstruct (link (:constructor db-make-link))
  prev
  next
  markedp
  hiddenp
  summary
  record)

(defsubst link-set-record (link result)
  ;; There is hidden work happening here.
  ;; This should also set link-hiddenp.
  (setf (link-record link) result
        (link-summary link) nil))

(defsubst db-join-links (link1 link2)
  ;; Place LINK1 and LINK2 in a prev-next relationship.
  (setf (link-next link1) link2
        (link-prev link2) link1))

(defun db-new-link (record)
  (let ((result (db-make-link)))
    (link-set-record result record)
    result))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Recordfieldspec abstraction
;;;

(defstruct (recordfieldspec
            (:type vector)              ; explicit to support kwidx
            :named)

  ;; datatype information
  type                                  ; e.g. 'string
  default-value
  common-form-function
  merge-function

  order-fn                              ; (A B) => {-1,0,1}
  sort-fn                               ; (A B) => bool
  match-function                        ; (PATTERN OBJ) => bool

  help-info                             ; perhaps should be help-string;
                                        ; or should be more complicated.

  actual->stored
  stored->actual

  ;; customizations
  change-hook                           ; not currently used
  constraint-function)

(put 'recordfieldspec 'kwidx            ; poor man's defsetf --ttn
     (let ((idx 0))
       (mapcar (lambda (ent)
                 (cons (intern (format ":%s" (symbol-name (car ent))))
                       (incf idx)))
               (cdr (get 'recordfieldspec 'cl-struct-slots)))))


(defun db-rs-sortfunc (rs &optional reversep)
  "Return a sort function for records described by recordfieldspec RS.
If optional argument REVERSEP is non-nil, then the sort function goes in
the opposite order.
If the sort-fn slot of the appropriate recordfieldspec of `database' doesn't
contain one, one is made up on the fly from the order-fn slot.
If the order-fn slot is also empty, the resulting function always returns
nil, indicating that it is not the case that the first argument is less
than the second."
  (let ((sort-fn (recordfieldspec-sort-fn rs)))
    (if sort-fn
        (if reversep
            `(lambda (value1 value2)
               ;; Maintain arg order and invert the result, so as
               ;; to properly handle the case: (eq value1 value2).
               (not (,sort-fn value1 value2)))
          sort-fn)
      (let ((order-fn (recordfieldspec-order-fn rs)))
        (if order-fn
            `(lambda ,(if reversep '(value2 value1) '(value1 value2))
               (= -1 (funcall (function ,order-fn) value1 value2)))
          'nil-function)))))

(defun db-rs-ordfunc (rs &optional reversep)
  "Return an order function for records described by recordfieldspec RS.
If optional argument REVERSEP is non-nil, then the order function goes in
the opposite order.
If the order-fn slot of the appropriate recordfieldspec of `database' doesn't
contain one, one is made up on the fly from the sort-fn slot; `equal'
is used to determine whether two records are equal.
If the sort-fn slot is also empty, the resulting function always
returns 0, indicating equality."
  (let ((order-fn (recordfieldspec-order-fn rs)))
    (if order-fn
        (if reversep
            `(lambda (value1 value2)
               (,order-fn value2 value1))
          order-fn)
      (let ((sort-fn (recordfieldspec-sort-fn rs)))
        (if sort-fn
            `(lambda ,(if reversep '(value2 value1) '(value1 value2))
               (cond ((equal value1 value2)
                      0)
                     ((funcall (function ,sort-fn) value1 value2)
                      -1)
                     (t
                      1)))
          (lambda (value1 value2) 0))))))


(defun database-recordfieldspec (database sel)
  "Return the recordfieldspec of DATABASE corresponding to SEL.
If SEL is a number, it is taken as the record index.
If it is a symbol, it names a field in the record.
Dereferences any symbol found in the recordfieldspecs slot of DATABASE."
  (let ((rs (aref (database-recordfieldspecs database)
                  (cond ((integerp sel) sel)
                        ((symbolp sel) (fieldname->fieldnumber sel database))
                        (t (error "Bad sel: %s" sel))))))
    (cond ((symbolp rs)
           (db-rfspec<-rftype rs))
          ((recordfieldspec-p rs)
           rs)
          (t
           (error "database-recordfieldspec:  rs = %s" rs)))))

(defun database-set-recordfieldspec (database sel rs)
  "Set the recordfieldspec of DATABASE corresponding to SEL to RS.
If SEL Is a number, it is taken as the record index.
If it is a symbol, it names a field in the record.
Use this to redefine, on a per-field basis, subfields of the recordfieldspec."
  (aset (database-recordfieldspecs database)
        (cond ((integerp sel) sel)
              ((symbolp sel) (fieldname->fieldnumber sel database))
              (t (error "Bad sel: %s" sel)))
        (cond ((symbolp rs)
               (db-rfspec<-rftype rs))
              ((recordfieldspec-p rs)
               rs)
              (t
               (error "database-set-recordfieldspec: Bad spec: rs= %s" rs)))))

(defun database-recordfieldspec-type (database idx)
  "Return the type of the recordfieldspec of DATABASE corresponding to IDX,
a record index."
  (let ((rs (aref (database-recordfieldspecs database) idx)))
    (cond ((symbolp rs)
           rs)
          ((recordfieldspec-p rs)
           (recordfieldspec-type rs))
          (t
           (error "database-recordfieldspec:  rs = %s" rs)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Records
;;;

;; Abstraction

(defsubst make-record (database)
  "Return a record with number of fields specified by argument DATABASE."
  (make-vector (database-no-of-fields database) nil))

(defun db-copy-r2r (source target)
  "Copy the field values of the SOURCE record to the TARGET record."
  (dotimes (fno (length source))
    (aset target fno (aref source fno))))

;;; Fieldnames and record fieldnumbers

(defsubst fieldname->fieldnumber (fieldname database)
  "Given a FIELDNAME and DATABASE, return a record fieldnumber.
Do not be fooled into thinking this is a format fieldnumber."
  (cdr (assq fieldname (database-fieldname-alist database))))

(defsubst db-fname<-fno (fieldnumber &optional database)
  "Given a record FIELDNUMBER and DATABASE, return a record fieldname.
If DATABASE is not specified, the value of `dbc-database' is used.
The first argument is not a format fieldnumber."
  (aref (database-fieldnames (or database dbc-database)) fieldnumber))

;;; Retrieving field values

(defsubst record-field-from-index (record fieldno)
  "Return from RECORD the value of the FIELDNOth field."
  (aref record fieldno))

(defsubst record-field (record fieldname database)
  "Return from RECORD the field with name FIELDNAME.  Third argument is DATABASE."
  (let ((fno (fieldname->fieldnumber fieldname database)))
    (if fno
        (record-field-from-index record fno)
      (error "No %s field in current record." fieldname))))

;;; Checking constraints

(defun db-check-constraint (value rec idx database)
  (let ((func (recordfieldspec-constraint-function
               (database-recordfieldspec database idx))))
    (when (and func (not (funcall func value rec idx database)))
      (error "The value `%s' does not satisfy the constraint for field %s."
             value (db-fname<-fno idx database)))))

;;; Setting field values

(defsubst record-set-field-from-index (record fieldno value database)
  "Set, in RECORD, the FIELDNOth field to VALUE.
Checks field constraints first if DATABASE is non-nil."
  (when database
    (db-check-constraint value record fieldno database))
  (aset record fieldno value))

(defun record-set-field (record fieldname value database &optional nocheck)
  "Set, in RECORD, field FIELDNAME to VALUE.  Fourth argument is DATABASE.
Check constraints first unless optional fifth argument NOCHECK is non-nil."
  (let ((fno (fieldname->fieldnumber fieldname database)))
    (if fno
        (record-set-field-from-index
         record fno value (and (not nocheck) database))
      (error "No %s field in current record." fieldname))))


;;; Setting fields in dbf-this-record

(defsubst dbf-this-record-set-field (fieldname value)
  "Set field with name FIELDNAME in `dbf-this-record' to VALUE.
Causes the entire record to be redisplayed pretty soon.
You may want to use `dbf-displayed-record-set-field' instead."
  ;; fixme: check dbf-this-record-modified-p first. --ttn
  (record-set-field dbf-this-record fieldname value dbc-database)
  (setq dbf-redisplay-entire-record-p t))

;;; The displayed record

(defsubst dbf-displayed-record-field (fieldname)
  "Return the value of the field named FIELDNAME from the displayed record."
  (record-field (dbf-displayed-record) fieldname dbc-database))

(defun dbf-displayed-record-set-field (fieldname value)
  "Set field with name FIELDNAME in displayed record to VALUE.
Cause the entire record to be redisplayed soon."
  ;; Make sure displayed-record = this-record.
  (dbf-set-this-record-modified-p t)
  (dbf-this-record-set-field fieldname value))

(defsubst dbf-displayed-record-set-field-and-redisplay (fieldname value)
  "Set field with name FIELDNAME in displayed record to VALUE.
Cause the entire record to be redisplayed immediately."
  ;; Is this correct?  Maybe displayed-record != this-record.
  (dbf-this-record-set-field fieldname value)
  (dbf-redisplay-entire-record-maybe))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sepinfo abstraction
;;;

;; This tells how a list of information appears in a file.

(defstruct sepinfo
  pre-first-string
  pre-first-regexp
  pre-first-regexp-submatch
  sep-string
  sep-regexp
  sep-regexp-submatch
  sep-function                          ; return (end-pos . next-start-pos)
                                        ; takes prev-end-pos as an argument
                                        ; next-start-pos is nil for last record
  post-last-string
  post-last-regexp
  post-last-regexp-submatch)


(defun db-make-n-line-sep-function (n)
  "Return a sep-function useful when all records have exactly N lines on disk."
  `(lambda (prev-end)
     (forward-line ,n)
     (cons (point) (unless (eobp) (point)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Moving from record to record
;;; NOTE: "Moving" has nothing to do with the display.
;;;

(defun db-next-link-and-index (database link idx n hidep markedp wraparoundp)
  ;; Arguments DATABASE LINK IDX N HIDEP MARKP WRAPAROUNDP.
  ;; Return a list of (link index) for Nth successor of LINK, whose index in
  ;; the database is IDX.  N may be negative.  If MARKEDP is non-nil, find the
  ;; Nth marked successor of LINK.  If HIDEP is non-nil, skip hidden links.
  ;; If WRAPAROUNDP is nil, stop at the first or last candidate link (ie,
  ;; properly marked and/or non-hidden, or LINK itself if no such
  ;; encountered); if the end of the database stops the search in this way,
  ;; the returned list also contains a third element, the number of elements
  ;; yet to go.
  (let* ((ln1 (database-first-link database))
         ;; We need to remember recent link and index if HIDEP or MARKEDP,
         ;; in case we examine every link.
         (rln link)                     ; recent link
         (ridx idx)                     ; recent index
         (fln (and wraparoundp          ; final link
                   (if (> n 0)
                       (link-prev ln1)
                     ln1))))
    ;; fixme: add infinite loop check. --ttn
    (while (and (> n 0) (not (eq link fln)))
      (setq link (link-next link))
      (incf idx)
      (if (or hidep markedp)
          (unless (or (and hidep (link-hiddenp link))
                      (and markedp (not (link-markedp link))))
            (setq rln link
                  ridx idx)
            (decf n))
        (decf n)))
    (while (and (< n 0) (not (eq link fln)))
      (setq link (link-prev link))
      (decf idx)
      (if (or hidep markedp)
          (unless (or (and hidep (link-hiddenp link))
                      (and markedp (not (link-markedp link))))
            (setq rln link
                  ridx idx)
            (incf n))
        (incf n)))
    (let (rv-ln rv-idx)
      (if (and (not (zerop n)) (or hidep markedp))
          (setq rv-ln rln rv-idx ridx)
        (setq rv-ln link rv-idx idx))
      (list rv-ln (let* ((tot (database-no-of-records database))
                         (remainder (% rv-idx tot)))
                    (if (zerop remainder) tot remainder))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Manipulating database records
;;;

(defun database-add-record (record database &optional location)
  ;; Add RECORD to DATABASE.  If optional third argument LOCATION is a number,
  ;; insert immediately before that index; if it is nil, insert at the end; if
  ;; it is t, insert in order.
  (let ((ln (db-new-link record)))
    (if (database-empty-p database)
        (progn
          (db-join-links ln ln)
          (setf (database-first-link database) ln))
      (let* ((aft (if (numberp location)
                      (car (db-next-link-and-index
                            database (database-first-link database)
                            1 (1- location) nil nil nil))
                    (database-first-link database)))
             (bef (link-prev aft)))
        (db-join-links ln aft)
        (db-join-links bef ln)
        (when (equal 1 location)
          (setf (database-first-link database) ln))))
    (incf (database-no-of-records database))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Mapping over a database
;;;

;; Mapping functions dynamically bind `db-lmap-link' and `db-lmap-index'.

(defvar db-lmap-link nil
  "The current link in a call to `db-lmap' or `db-lmap-macro'.")
(defvar db-lmap-index nil
  "The current index in a call to `db-lmap' or `db-lmap-macro'.")

(defmacro db-lmap-break ()
  "Cause the db-lmap loop to quit after executing the current iteration.
This is not a nonlocal exit!  It sets a flag which prevents future iterations.
\(Actually, it sets variable `db-lmap-link'.\)"
  `(setq db-lmap-link (link-prev first-link)))

(defun db-lmap (func database &optional hide message accumulate)
  "Apply FUNC to every link in DATABASE.
If optional third arg HIDE is non-nil, apply FUNC only to unhidden links.
If optional fourth arg MESSAGE is non-nil, it should be a format string
containing one numeric \(%d\) specifier.  That message will be issued every
`db-inform-interval' links.
If optional fifth arg ACCUMULATE is non-nil, return a list of the results;
otherwise return nil.

In the body, variable `db-lmap-index' is bound to the index of the link being
operated upon, and `db-lmap-link' is the argument to FUNC.
The loop may be short-circuited (aborted) by calling `db-lmap-break'.
To avoid the per-link function call overhead, use `db-lmap-macro' instead."
  (let* ((ln1 (database-first-link database))
         (db-lmap-link ln1)
         (db-lmap-index 1)
         v results donep)
    (setq message (and db-inform-interval message))
    (when ln1
      (while (not donep)
        (unless (and hide (link-hiddenp db-lmap-link))
          (setq v (funcall func db-lmap-link))
          (when accumulate (push v results)))
        (when (and message (zerop (% db-lmap-index db-inform-interval)))
          (db-message message db-lmap-index))
        (setq db-lmap-link (link-next db-lmap-link))
        (incf db-lmap-index)
        (setq donep (eq db-lmap-link ln1))))
    (when accumulate
      (nreverse results))))

(defmacro db-lmap-macro (lmap-body database &optional hide message)
  "Execute LMAP-BODY for each link in DATABASE, and return nil.
If optional third arg HIDE is non-nil, execute LMAP-BODY only for unhidden links.
If optional fourth arg MESSAGE is non-nil, it should be a format string
containing one numeric \(%d\) specifier.  That message will be issued every
`db-inform-interval' links.

In the body, variable `db-lmap-link' is bound to the link being operated upon,
and `db-lmap-index' is bound to its index.
The loop may be short-circuited (aborted) by calling `db-lmap-break'.
Speed demons should call this instead of `db-lmap' to avoid a function call
overhead per link."
  (declare
   ;; Second arg `sexp' because the macro uses its unevalled value.
   (debug (form sexp &optional form form)))
  `(let* (,@(when hide `((hidep ,hide)))
          ,@(when message
              `((message-evalled (and db-inform-interval ,message))))
          (first-link (database-first-link ,database))
          (db-lmap-link first-link)
          (db-lmap-index 1)
          (lmap-not-done t))
     (when first-link
       (while lmap-not-done
         ,(if hide
              `(unless (and hidep
                            (link-hiddenp db-lmap-link))
                 ;; Body is a single form
                 ,lmap-body)
            lmap-body)
         ,@(when message
             '((when (and message-evalled
                          (zerop (% db-lmap-index db-inform-interval)))
                 (db-message message-evalled db-lmap-index))))
         (setq db-lmap-link (link-next db-lmap-link))
         (incf db-lmap-index)
         (when (eq db-lmap-link first-link)
           (setq lmap-not-done nil))))
     nil))

(fset 'maprecords-break 'db-lmap-break)


(defun maprecords (maprecords-func database &optional hide message accumulate)
  ;; The formal parameter name MAPRECORDS-FUNC is different than that of
  ;; `db-lmap' because if they're the same, then when the function created here
  ;; is run and FUNC is looked up, the nearest dynamically enclosing binding
  ;; of FUNC will be returned.  That won't be the one we want.
  "Apply MAPRECORDS-FUNC to every record in DATABASE.  Return nil.
If optional third arg HIDE is non-nil, apply MAPRECORDS-FUNC only to
unhidden records.
If optional fourth arg MESSAGE is non-nil, it should be a format string
containing one numeric \(%d\) specifier.  That message will be issued every
`db-inform-interval' records.
If optional fifth arg ACCUMULATE is non-nil, return a list of the results;
otherwise return nil."
  (db-lmap (lambda (ln)
             (funcall maprecords-func (link-record ln)))
           database hide message accumulate))

;;; db-rep.el ends here
