;;; db-search.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;;; Code:

(provide 'db-search)


(deflocalvar dbf-field-search-defaults nil
  "Vector of defaults for field search.
It is one element longer than the number of fields; the element indexed by
dbf-displayspecs-length is the default for a search over all fields.")

(defvar database-search-mode-map (copy-keymap database-edit-mode-map)
  "Keymap for database data display buffer in edit mode.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Searching primitives
;;;

;;;###autoload
(defun db-parse-match-pattern (string ds)
  ;; These connectives and prefixes are listed in order of precedence.
  ;; [That order is rather unconventional.  Also, no grouping. --ttn]
  ;; The latter prefixes used to be preceded by [ \t]* and followed by [ \t]+;
  ;; for instance, (defvar dbm-<-prefix "^[ =t]*<[ \t]+").
  (let ((c-and "[ \t]+and[ \t]+")
        (c-or  "[ \t]+or[ \t]+")
        (p-not "^[ \t]*not[ \t]+")
        (p-<   "^<")
        (p->   "^>")
        (p-=   "^="))
    (flet ((try (x) (string-match x string))
           (sub (x) (db-parse-match-pattern x ds))
           (p1o (x) (db-display->actual-call (displayspec-display->actual ds)
                                             x nil nil nil))
           (b () (substring string 0 (match-beginning 0)))
           (e () (substring string (match-end 0))))
      (cond ((try c-and) `(db-match-and ,(sub (b)) (sub (e))))
            ((try c-or)  `(db-match-or  ,(sub (b)) (sub (e))))
            ;; no infix connectives in string
            ((try p-not) `(db-match-not ,(sub (e))))
            ((try p-<)   `(db-match-<   ,(p1o (e))))
            ((try p->)   `(db-match->   ,(p1o (e))))
            ((try p-=)   `(db-match-=   ,(p1o (e))))
            (t (db-display->actual-call
                (or (displayspec-match-display->actual ds)
                    (displayspec-display->actual ds))
                string
                nil nil nil))))))

;;;###autoload
(defun db-print-match-pattern (pat ds)
  (flet ((sub (x) (db-print-match-pattern x ds))
         (p1o (x) (db-actual->display-call (displayspec-actual->display ds)
                                           x nil nil)))
    (let* ((type (when (listp pat) (car pat)))
           (one (and type (cadr pat)))
           (two (and type (caddr pat))))
      (case type
        ;; extra space around AND to emphasize its low precedence
        (db-match-and (concat (sub one) "  AND  " (sub two)))
        (db-match-or  (concat (sub one)   " OR "  (sub two)))
        (db-match-not (concat "NOT " (sub one)))
        (db-match-<   (concat   "< " (p1o one)))
        (db-match->   (concat   "> " (p1o one)))
        (db-match-=   (concat   "= " (p1o one)))
        ;; pat was not a list or the type wasn't recognized
        (t (db-actual->display-call
            (or (displayspec-match-actual->display ds)
                (displayspec-actual->display ds))
            pat
            nil nil))))))

;;;###autoload
(defun db-match (pat targ rs)
  (flet ((m (x) (db-match x targ rs))
         (rso (x) (funcall (db-rs-ordfunc rs) x targ))
         (rsm (x) (funcall (recordfieldspec-match-function rs) x targ)))
    (if (listp pat)
        (let ((one (cadr pat))
              (two (caddr pat)))
          (case (car pat)
            (db-match-and (and (m one) (m two)))
            (db-match-or  (or  (m one) (m two)))
            (db-match-not (not (m one)))
            (db-match-<   (=  1 (rso one)))
            (db-match->   (= -1 (rso one)))
            (db-match-=   (=  0 (rso one)))
            (t            (rsm pat))))
      (rsm pat))))

;;; db-search.el ends here
