;;; db-sort.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Ordering-fields and field-priorities refer to the same thing.
;; Order funcs preferred over sort funcs.

;;; Code:


(provide 'db-sort)


(defvar db-converted-ofields)           ; used dynamically


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

(defvar db-sort-modifies-p nil
  "*If non-nil, then sorting a database marks it as modified too.")

;; Field priorities is a cons of two lists; the first list is those that
;; will be used, and the the second set is the ignored fields.  Ignorance
;; is only in the user interface; EDB always remembers everything.

;; Each list consists of (fieldnumber . order-info) pairs such that every
;; field is accounted for exactly once.  The order-info is 'increasing,
;; 'decreasing, or a cons of (type . value) where type is 'order-function or
;; 'sort-function.

;; Priorities are initially associated w/ the database.  On first use
;; it becomes format-local.

(deflocalvar dbf-field-priorities nil
  "The list of field priorities for this database in this data display buffer.
If non-nil, overrides the database's field-priorities slot.")
(deflocalvar dbf-hidden-to-end-p nil
  "The default, local to this data display buffer, for the hidden-to-end-p
database slot.  Only used if  dbf-field-priorities  is non-nil.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sort
;;;

;; fixme: Implement "hidden at end" as synthetic field (attribute).  --ttn

;;;###autoload
(defun database-sort (database &optional sorter hidden-records-at-end-p)
  "Sort and return DATABASE, which is also side-effected.  SORTER is
either a field priorities list or a function which takes two records
as arguments and returns t if r1 < r2."
  (db-message "Sorting...")
  (unless (database-empty-p database)
    (let (db-converted-ofields)

      (unless (functionp sorter)
        (setq db-converted-ofields (db-convert-ordering-fields
                                    sorter database)
              sorter 'db-record-lessp))
      (database-set-links-from-list database
        (sort (db-lmap 'identity database nil nil t)
              (lambda (l1 l2)
                (funcall sorter
                         (link-record l1)
                         (link-record l2)))))))
  (when db-sort-modifies-p
    (database-set-modified-p database t)
    (force-mode-line-update))
  (db-message "Sorting...done")
  database)


(defun db-convert-ordering-fields (ofields database)
  (let (f fno sortinfo type value)
    (mapcar (lambda (spec)
              (setq f (car spec)
                    fno (if (integerp f)
                            f
                          (fieldname->fieldnumber f database))
                    sortinfo (cdr spec))
              (cons fno
                    (if (atom sortinfo)
                        (db-rs-ordfunc (database-recordfieldspec database fno)
                                       (eq sortinfo 'decreasing))
                      (setq type (car sortinfo)
                            value (cdr sortinfo))
                      (cond ((eq 'order-function type)
                             value)
                            ((eq 'sort-function type)
                             ;; fixme: lame. --ttn
                             (error "Trying to order, but only sort function provided for field %s (%s)."
                                    fno
                                    (if (symbolp f)
                                        f
                                      (db-fname<-fno f database))))
                            (t (error "Unrecognized type in %s" sortinfo))))))
            (car (or ofields
                     (database-field-priorities database))))))


(defun db-make-ordering-fields-canonical (ofields database)
  ;; Make sure that every car is a number (not a fieldname) and
  ;; that every field is represented.
  (let* ((nfields (database-no-of-fields database))
         (seen (make-vector nfields nil))
         fno sortinfo value
         (check-field
          (lambda (spec)
            (setq fno (if (numberp (car spec))
                          (car spec)
                        (fieldname->fieldnumber (car spec) database))
                  sortinfo (cdr spec))
            (cond ((not fno)
                   (db-warning "%s is not a field in the database; ignoring it."
                               (car spec))
                   nil)
                  ((aref seen fno)
                   (db-warning "%s already appears; ignoring subsequent entries."
                               (db-fname<-fno fno database))
                   nil)
                  (t
                   (aset seen fno t)
                   (cons fno
                         (cond ((eq sortinfo 'decreasing)
                                'decreasing)
                               ((atom sortinfo)
                                (when (and sortinfo (not (eq sortinfo 'increasing)))
                                  (db-warning "Unrecognized ordering symbol %s being changed to 'increasing."
                                              sortinfo))
                                'increasing)
                               ;; sortinfo is a list
                               ((or (eq (car sortinfo) 'order-function)
                                    (eq (car sortinfo) 'sort-function))
                                (unless (functionp (cdr sortinfo))
                                  (db-warning "%s is claimed to be a %s but doesn't look like a function to me."
                                              (cdr sortinfo) (car sortinfo)))
                                sortinfo)
                               (t (error "Unrecognized type in %s" sortinfo))))))))
         (sig-fields (mapcar check-field (car ofields)))
         (nonsig-fields (mapcar check-field (cdr ofields)))
         missing-fields)
    (setq fno 0)
    (while (< fno nfields)
      (unless (aref seen fno)
        (db-warning "%s doesn't appear; adding it."
                    (db-fname<-fno fno database))
        (push (cons fno 'increasing) missing-fields))
      (setq fno (1+ fno)))
    (cons sig-fields
          (if missing-fields
              (append nonsig-fields missing-fields)
            nonsig-fields))))


;; Uses the `db-converted-ofields' dynamic variable.
(defun db-order-records (r1 r2)
  ;; Return -1, 0, or 1 depending on whether records R1 and R2 are <, =, or >.
  (let ((ofields db-converted-ofields)
        (result 0)
        fno ofunc)
    (while (and ofields (zerop result))
      (setq fno (car (car ofields))
            ofunc (cdr (car ofields))
            ofields (cdr ofields)
            result (funcall ofunc (aref r1 fno) (aref r2 fno))))
    result))


(defun db-record-lessp (r1 r2)
  ;; Return t if record R1 < R2.  Use `db-order-records'.
  (= -1 (db-order-records r1 r2)))


(defun database-sorted-p (database &optional sorter)
  ;; Return t if DATABASE is sorted, nil otherwise.
  ;; Optional argument SORTER is used as the function.
  (db-with-sorter database
    (let ((rv t)
          (flp t))
      ;; Don't do anything for the first link.
      ;; If some other link is less than its predecessor, complain.
      (db-lmap-macro
       (if flp
           (setq flp nil)
         (when (funcall sorter
                        (link-record db-lmap-link)
                        (link-record (link-prev db-lmap-link)))
           (setq rv nil)
           (db-lmap-break)))
       database)
      rv)))

(defun database-ordered-p (database &optional orderer)
  ;; Return t if DATABASE is ordered, nil otherwise.
  ;; Optional argument ORDERER is used as the ordering function.
  (db-with-orderer database
    (let ((rv t)
          (flp t))                      ; first link
      (db-lmap-macro
       (if flp
           (setq flp nil)
         (when (= 1 (funcall orderer
                             (link-record (link-prev db-lmap-link))
                             (link-record db-lmap-link)))
           ;; fixme: short-circuit. --ttn
           (setq rv nil)))
       database)
      rv)))


(defmacro db-with-orderer (database &rest body)
  (declare (indent 1) (debug (sexp body)))
  `(let (db-converted-ofields)
     (unless orderer                    ; dynamically bound
       (setq orderer 'db-order-records
             db-converted-ofields
             ;; "stupid way to call this function" (why? --ttn)
             (db-convert-ordering-fields nil ,database)))
     ,@body))

(defmacro db-with-sorter (database &rest body)
  (declare (indent 1) (debug (sexp body)))
  `(let (db-converted-ofields)
     (if (not sorter)                   ; dynamically bound
         (setq sorter 'db-record-lessp
               db-converted-ofields
               ;; "stupid way to call this function" (why? --ttn)
               (db-convert-ordering-fields nil ,database)))
     ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Sorting interface
;;;

;; Commands:
;;  cursor movement
;;  kill a line
;;  yank a line
;;  sort on this field only
;;  accept this ordering, sort with it, and make it the default
;;  accept this ordering and sort with it, but don't make it the default
;;  toggle ordering direction (increasing vs decreasing) or -to-end-p
;;  input custom ordering info:  a list or a function name

;;; The buffer looks like:
;; ==== Significant fields:
;; foo
;; bar
;; baz
;; ==== Nonsignificant fields:
;; bum
;; but
;; ==== Hidden records to end:  nil

;;;###autoload
(defun database-sort-interface (database)
  ;; Assume current-buffer is data display buffer.
  (let ((canon (db-make-ordering-fields-canonical
                (database-field-priorities database)
                database))
        (ddb (current-buffer)))
    (switch-to-buffer (get-buffer-create
                       (concat "Ordering info for "
                               (database-print-name database))))
    (database-sort-interface-mode)

    (setq dbsi-data-display-buffer ddb
          dbsi-database database
          dbsi-hidden-to-end-p (database-hidden-to-end-p database)
          dbsi-killed-fields nil
          dbsi-sig-fields (cons nil (car canon))
          dbsi-nonsig-fields (cons nil (cdr canon)))

    ;; Fill the buffer with info.
    (let (buffer-read-only)
      (erase-buffer)
      (insert "==== Significant fields:\n")
      (mapc 'dbsi-format (cdr dbsi-sig-fields))
      (insert "==== Nonsignificant fields:\n")
      (mapc 'dbsi-format (cdr dbsi-nonsig-fields))
      (insert "==== Hidden records to end:  "
              (if dbsi-hidden-to-end-p "Yes" "No ")))
    (goto-char (point-min))
    (db-message "%s, %s, %s, %s, %s"
                "u = use"
                "RET = also make default"
                "! = this field only"
                "q = abort"
                "? = help")))

(defun database-sort-interface-mode ()
  "Sort interface buffer for choosing fields upon which to sort a database.
\\<database-sort-mode-map>
Reorder the fields with the following commands:
  \\[dbsi-kill-line]    Kill field under point.
  \\[dbsi-yank-line]    Yank most recently killed field.
Specify how a particular field should be ordered with these commands:
  \\[dbsi-increasing]   Sort the field under point in increasing order.
  \\[dbsi-decreasing]   Sort the field under point in decreasing order.
  \\[dbsi-ordering-function]    Specify an ordering function for the field at point.
  \\[dbsi-sorting-function]     Specify a sorting function for the field at point.
Change whether hidden records go at the end of the sorted order with:
  \\[dbsi-toggle-hidden-to-end] Toggle the value of  dbsi-hidden-to-end-p.
Exit the sort interface with the following commands:
  \\[dbsi-use-ordering-make-database-default]   Use this ordering, and make it the default for this database.
  \\[dbsi-use-ordering-make-buffer-default]     Use this ordering, and make it the default in this buffer only.
  \\[dbsi-quit-clear-buffer-default]    Clear the default sort in this buffer, and abort the sort.
  \\[dbsi-use-ordering] Use this ordering, but don't make it the default.
  \\[dbsi-this-field-only]      Sort only according to the field at point.
  \\[dbsi-quit] Abort the sort; don't change the default ordering criterion.

More specifically:
\\{database-sort-mode-map}"
  (setq major-mode 'database-sort-interface-mode)
  (setq mode-name "Database Sort Interface")

  (set-buffer-modified-p nil)
  (setq buffer-read-only t)

  (auto-save-mode 0)
  (setq buffer-file-name nil)

  (setq truncate-lines t)

  (use-local-map database-sort-mode-map)

  (setq mode-line-format
        '("---Database Sort: %17b ["
          (-3 . "%p")
          "]-%-")))


(deflocalvar dbsi-data-display-buffer)
(deflocalvar dbsi-database nil)
(deflocalvar dbsi-killed-fields nil)
;; These are lists whose first element is ignored.
(deflocalvar dbsi-sig-fields nil)
(deflocalvar dbsi-nonsig-fields nil)
(deflocalvar dbsi-hidden-to-end-p nil
  "The hidden-to-end slot of the database will be set from this variable.")

(defvar database-sort-mode-map nil
  "Keymap for database data display buffer in view mode.")

(unless database-sort-mode-map
  (let ((m (make-sparse-keymap)))
    (suppress-keymap m)
    (mapc (lambda (key-def)
            (define-key m (car key-def) (cdr key-def)))
          '(("\C-k"     . dbsi-kill-line)
            ("\C-y"     . dbsi-yank-line)
            ("!"        . dbsi-this-field-only)
            ("\r"       . dbsi-use-ordering-make-database-default)
            ("\C-c\C-c" . dbsi-use-ordering-make-database-default)
            ("A"        . dbsi-use-ordering-make-buffer-default)
            ("U"        . dbsi-use-ordering-make-buffer-default)
            ("c"        . dbsi-quit-clear-buffer-default)
            ("a"        . dbsi-use-ordering)
            ("u"        . dbsi-use-ordering)
            ("t"        . dbsi-toggle-hidden-to-end)
            ("o"        . dbsi-ordering-function)
            ("s"        . dbsi-sorting-function)
            ("i"        . dbsi-increasing)
            ("d"        . dbsi-decreasing)
            ("q"        . dbsi-quit)
            ("?"        . describe-mode)))
    (setq database-sort-mode-map m)))


;; PC is previous-cons
(defmacro db-with-pc (&rest body)
  (declare (indent 0) (debug body))
  `(let ((pc (dbsi-prev-cons))
         (buffer-read-only nil))
     (if pc
         (progn ,@body)
       (error "Can't insert here."))))

;; PC is previous-cons; TI is this-item.
(defmacro db-with-pc-and-ti (&rest body)
  (declare (indent 0) (debug body))
  `(let* ((pc (dbsi-prev-cons))
          (ti (car (cdr pc)))
          (buffer-read-only nil))
     (if ti
         (progn ,@body)
       (error "No field on current line."))))

;;; Determining where we are.

;; mernst sez: I have no idea whether this is correct; it's a crock;
;;             just test it 'til it's right.
;; ttn sez: where is the test?

(defun dbsi-prev-cons ()
  ;; actually 1 less than the line number
  (let ((lno (save-excursion
               (beginning-of-line)
               (count-lines (point-min) (point)))))
    (cond ((zerop lno)
           nil)
          ((<= lno (length dbsi-sig-fields))
           (nthcdr (1- lno) dbsi-sig-fields))
          (t
           (decf lno (1- (length dbsi-sig-fields)))
           (when (< lno (length dbsi-nonsig-fields))
             (nthcdr lno dbsi-nonsig-fields))))))


;; Insert item and trailing newline.
(defun dbsi-format (item)
  (insert (format "%-16s    " (db-fname<-fno (car item) dbsi-database)))
  (if (atom (cdr item))
      (insert (format "%s" (cdr item)))
    (insert (format "%s [%s]" (cdr (cdr item)) (car (cdr item)))))
  (insert "\n"))

(defun dbsi-reformat (item)
  (beginning-of-line)
  (kill-line 1)
  (dbsi-format item))

;;; Moving fields around.

(defun dbsi-kill-line ()
  "Kill field on current line, placing it in the sort interface kill stack."
  (interactive)
  (db-with-pc-and-ti
    (push ti dbsi-killed-fields)
    (setcdr pc (cdr (cdr pc)))
    (beginning-of-line)
    (kill-line 1)))

(defun dbsi-yank-line ()
  "Yank most recently killed (lifo ordering) field, inserting it before point.
This removes the field from the sort interface kill stack."
  (interactive)
  (db-with-pc
    (if dbsi-killed-fields
        (let ((ti (car dbsi-killed-fields)))
          (beginning-of-line)
          (setcdr pc (cons ti (cdr pc)))
          (dbsi-format ti)
          (previous-line 1)
          (setq dbsi-killed-fields (cdr dbsi-killed-fields)))
      (error "Nothing to yank."))))

;;; Changing info about hiding.

(defun dbsi-toggle-hidden-to-end ()
  "Toggle the boolean value of `dbsi-hidden-to-end-p'.
This controls whether hidden records should all be placed at the end
of the sorted order or should be sorted according to the same criteria as
non-hidden records."
  (interactive)
  (save-excursion
    (let (buffer-read-only)
      (setq dbsi-hidden-to-end-p (not dbsi-hidden-to-end-p))
      (goto-char (point-max))
      (backward-delete-char 3)
      (insert (if dbsi-hidden-to-end-p "Yes" "No ")))))

;;; Changing

(defun dbsi-increasing ()
  "Specify that the field at point should use an increasing ordering."
  (interactive)
  (db-with-pc-and-ti
    (setcdr ti 'increasing)
    (dbsi-reformat ti)))

(defun dbsi-decreasing ()
  "Specify that the field at point should use a decreasing ordering."
  (interactive)
  (db-with-pc-and-ti
    (setcdr ti 'decreasing)
    (dbsi-reformat ti)))

(defun dbsi-ordering-function ()
  "Specify an ordering function for the field at point.
An ``ordering function'' returns -1, 0, or 1 depending on whether its first
argument is less than, equivalent to, or greater than its second argument."
  (interactive)
  (db-with-pc-and-ti
    (setcdr ti (cons 'order-function
                     (read-from-minibuffer "Ordering function: " nil nil t)))
    (dbsi-reformat ti)))

(defun dbsi-sorting-function ()
  "Specify a sorting function for the field at point.
A ``sorting function'' returns t if its first argument is less
than its second argument and nil otherwise."
  (interactive)
  (db-with-pc-and-ti
    (setcdr ti (cons 'sort-function
                     (read-from-minibuffer "Sorting function: " nil nil t)))
    (dbsi-reformat ti)))

;;; Quitting

(defun dbsi-quit ()
  "Abort the sort and exit the sort interface."
  (interactive)
  (let ((ddb dbsi-data-display-buffer))
    (kill-buffer nil)
    (switch-to-buffer ddb)
    (db-message "Aborting sort")))

(defun dbsi-quit-clear-buffer-default ()
  "Clear the default sort order for this buffer and exit the sort interface
without sorting.
In the future, the default sort order will come from the database."
  (interactive)
  (let ((ddb dbsi-data-display-buffer))
    (kill-buffer nil)
    (switch-to-buffer ddb)
    (setq dbf-field-priorities nil)
    (db-message "Reset buffer-local default sort order; didn't sort")))

;;; Specifying an ordering.

(defun dbsi-killed-fields-to-end-maybe ()
  (when (cdr dbsi-killed-fields)
    (if (y-or-n-p
         "Put killed but not yanked fields at end of nonsignificant list? ")
        (nconc dbsi-nonsig-fields (cdr dbsi-killed-fields))
      (error "There are still killed fields."))))

(defun dbsi-use-ordering-make-database-default ()
  "Use the current ordering to sort, and make it the default for future sorts
of this database.
The user is warned if there are killed, non-yanked fields."
  (interactive)
  (dbsi-killed-fields-to-end-maybe)
  (with-current-buffer dbsi-data-display-buffer
    (setq dbf-field-priorities nil))
  (setf (database-field-priorities dbsi-database)
        (cons (cdr dbsi-sig-fields) (cdr dbsi-nonsig-fields))
        (database-hidden-to-end-p dbsi-database)
        dbsi-hidden-to-end-p)
  (dbsi-use-ordering (database-field-priorities dbsi-database)))

(defun dbsi-use-ordering-make-buffer-default ()
  "Use the current ordering to sort, and make it the default for future sorts
in this data display buffer only.
The user is warned if there are killed, non-yanked fields."
  (interactive)
  (dbsi-killed-fields-to-end-maybe)
  (let ((ordering (cons (cdr dbsi-sig-fields) (cdr dbsi-nonsig-fields))))
    (with-current-buffer dbsi-data-display-buffer
      (setq dbf-field-priorities ordering
            dbf-hidden-to-end-p dbsi-hidden-to-end-p))
    (dbsi-use-ordering ordering)))

(defun dbsi-use-ordering (&optional ordering)
  "Use the current ordering for this sort only."
  (interactive)
  (database-sort dbsi-database
                 (or ordering (cons (cdr dbsi-sig-fields) nil))
                 dbsi-hidden-to-end-p)
  (let ((ddb dbsi-data-display-buffer))
    (kill-buffer nil)
    (switch-to-buffer ddb)
    (dbf-finished-sorting)))


;; If sorting on a field, don't even use ordering-fields; get the field's
;; sorting function and call database-sort with just that (with a wrapper
;; so that it takes a record as an argument).  This is why we want to
;; permit (but not encourage, particularly) sort functions in recordfieldspecs.


(defun dbsi-this-field-only ()
  "Sort according to only the field at point.
All editing of other fields is ignored."
  ;; This is similar to, but different from, what's done in
  ;; db-convert-ordering-fields.
  (interactive)
  (dbsi-use-ordering
   (db-with-pc-and-ti
     (let* ((fno (car ti))
            (spec (cdr ti))
            (sort-function
             (if (atom spec)
                 (db-rs-sortfunc (database-recordfieldspec dbsi-database fno)
                                 (eq spec 'decreasing))
               (let ((func (cdr spec)))
                 (case (car spec)
                   (order-function (lambda (item1 item2)
                                     (= -1 (funcall func item1 item2))))
                   (sort-function func)
                   (t (error "unkown spec type: %s" (car spec))))))))
       `(lambda (r1 r2)
          (funcall ',sort-function
                   (aref r1 ,fno)
                   (aref r2 ,fno)))))))

;;; db-sort.el ends here
