;;; db-summary.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Patterned in part after rmail-new-summary.

;;; Code:


(provide 'db-summary)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;

;;;
;;; Hooks
;;;

(defvar database-summary-mode-hooks nil
  "Function or list of functions run when switching to Database Summary mode.")


;;;
;;; Summary variables
;;;

;; All these are expendible; authoritative values are copied in.

(deflocalvar dbs-data-display-buffer nil
  "The buffer of the format for which this buffer is a summary.")

(deflocalvar dbs-index nil
  "The index of the record summary at point.
Used in determining whether the data display buffer and its summary are in synch.
Don't set this variable directly; use `dbs-set-index' instead.")

(defsubst dbs-set-index (index)
  ;; Set `dbs-index' to INDEX and `dbs-index-fraction' appropriately.
  (setq dbs-index index
        dbs-index-fraction
        (format "%d/%d" dbs-index dbs-no-of-records)))

(deflocalvar dbs-no-of-records nil
  "The number of records in the database when this summary was made.")

(deflocalvar dbs-point nil
  "The beginning of the current record.")

(deflocalvar dbs-index-fraction nil
  "Like `dbc-index-fraction', for the benefit of the mode line.")

(deflocalvar dbs-recompute-p nil
  "T if some summary information is out of date, nil otherwise.
This is usually set to t when some link-summary is set to nil.")



;;;
;;; Format variables related to the summary
;;;

;; Summary buffer does not follow observer protocol;
;; multiple summary buffers unsupported and unsupportable.


(deflocalvar dbf-summary-format nil
  "A string in the same format as the format-file.
Use `dbf-set-summary-format' to set it.")

;; Needn't be buffer-local.
(defvar dbf-default-summary-format nil)

(deflocalvar dbf-summary-function nil
  "Function which inserts summary information for a single record in
the summary buffer; it takes the record as its argument.")

(deflocalvar dbf-summary-buffer nil
  "The summary buffer associated with this format.")

(deflocalvar dbf-summary-show-hidden-records-p t
  "Nil if hidden records should be hidden from the summary, t otherwise.")

(deflocalvar dbf-summary-recompute-all-p nil
  "T if every record summary in this buffer should be recomputed.")


;;;
;;; Variables in both the format and summary buffers
;;;

;; These variables are too important to be kept only in the summary buffer,
;; which may disappear at any time, but are often used by it and so should
;; be handy.

;; This doesn't depend on the field values in the individual records
;; because db-format->lines/sforms errs if min-height is not
;; equal to max-height (unless variable-height is set).  That makes
;; determining which summary point is in, and getting to a particular
;; summary, much easier.
(deflocalvar dbfs-lines nil
  "The (constant) number of screen lines occupied by each record summary.
This variable is computed automatically from the summary format.
It has a value in both the summary and data display buffers.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Macros for working in the correct buffer
;;;

(defsubst dbf-summary-buffer ()
  (and (bufferp dbf-summary-buffer)
       (buffer-name dbf-summary-buffer)
       dbf-summary-buffer))

(defmacro dbf-in-summary-buffer (&rest body)
  ;; Execute BODY in the summary buffer, if it exists.
  (declare (indent 0) (debug body))
  `(when (dbf-summary-buffer)
     (db-in-buffer dbf-summary-buffer
       (progn
         ,@body))))

(defmacro dbs-in-data-display-buffer (&rest body)
  ;; Execute BODY in the data display buffer (which always exists).
  (declare (indent 0) (debug body))
  `(db-in-buffer dbs-data-display-buffer
     ,@body))

(defsubst db-data-display-buffer ()
  "Return the database data display buffer associated with the current buffer,
which must be either a summary buffer or a data display buffer."
  (cond ((db-summary-buffer-p)
         dbs-data-display-buffer)
        ((db-data-display-buffer-p)
         (current-buffer))
        (t
         (error "Neither in format nor summary buffer"))))

(defsubst db-summary-buffer ()
  ;; Return the database summary buffer associated with the current buffer,
  ;; which must be either a summary buffer (which is returned) or a data
  ;; display buffer.  Return nil if there is no associated summary buffer.
  (cond ((db-summary-buffer-p)
         (current-buffer))
        ((db-data-display-buffer-p)
         (dbf-summary-buffer))
        (t
         (error "Neither in format nor summary buffer"))))

(defmacro db-in-data-display-buffer (&rest body)
  (declare (indent 0) (debug body))
  `(db-in-buffer (db-data-display-buffer)
     ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Creating the summary
;;;


(defun db-summary ()
  "Display a summary (or directory) of all database records according to
the value by `dbf-set-summary-format'.  The summary appears in a separate buffer.
When called from the summary buffer, this updates the summary."
  (interactive)

  (db-in-data-display-buffer

    (cond ((database-empty-p dbc-database)
           (delete-windows-on (dbf-summary-buffer))
           (db-message "Database is empty"))
          (t
           (let ((ddb (current-buffer)))
             (unless dbf-summary-function
               (dbf-set-summary-format dbf-summary-format))
             (unless (dbf-summary-buffer)
               (setq dbf-summary-buffer
                     (db-in-buffer (generate-new-buffer
                                    (format "%s-summary" (buffer-name ddb)))
                       (setq dbs-data-display-buffer ddb
                             dbc-database (with-current-buffer ddb
                                            dbc-database))
                       (database-summary-mode)
                       (current-buffer))))
             ;; Avoid a check by not using `dbf-in-summary-buffer'.
             (when (with-current-buffer dbf-summary-buffer
                     (dbs-out-of-date-p))
               (dbf-fill-summary-buffer))
             (pop-to-buffer dbf-summary-buffer)
             (setq dbs-data-display-buffer ddb)
             ;; Go to proper line.
             (dbs-move-to-proper-record))))))


;; This is spelled out instead of being db-summary-mode because it's a
;; major mode, while db-edit-mode and db-view-mode are minor modes.
(defun database-summary-mode ()
  "Summary buffer for database mode.
Most keystrokes perform the same function they do in the data display buffer.

Key bindings:

\\{database-summary-mode-map}"

  (setq major-mode 'database-summary-mode)
  (setq mode-name "Database Summary")

  (set-buffer-modified-p nil)
  (setq buffer-read-only t)

  (auto-save-mode 0)
  (setq buffer-file-name nil)

  (use-local-map database-summary-mode-map)

  (setq mode-line-format
        (dbs-in-data-display-buffer
          (list
           (format "-----Database: %17s   %%[(Summary" (buffer-name))
           'minor-mode-alist
           " "
           'dbs-index-fraction
           ")%]---"
           '(-3 . "%p")
           "-%-")))

  (run-hooks 'database-summary-mode-hooks)

  ;; Force an update.
  (setq dbs-no-of-records -1))


(defsubst db-summary-buffer-p ()
  "T if this buffer is a database summary buffer."
  (eq major-mode 'database-summary-mode))


(defun db-summary-subset ()
  "Make EDB summary-listing based on hits of STRING search, of FIELD."
  (interactive)
  (db-mark-searched-records)
  (db-hide-unmarked-records)
  (db-toggle-show-hidden-records 0)
  (db-summary)
  (db-hiding-toggle 0)
  (db-hiding-toggle 1))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Filling the summary
;;;

(defun dbs-insert-link-summary (link mhp)
  ;; mhp -- mark-hidden-records-p
  (let ((one (if (link-markedp link) "+" " "))
        (two (if (and mhp (link-hiddenp link)) "[" " "))
        (rest (substring (link-summary link) 2)))
    (insert one two rest)))


(defun dbf-fill-summary-buffer ()
  ;; fixme: check database nonempty. --ttn
  (let ((sum dbf-summary-function)
        (lines dbfs-lines)
        (hide (and (not dbf-summary-show-hidden-records-p)
                   dbc-hide-p))
        (mhp dbc-hide-p)
        (redo-p dbf-summary-recompute-all-p))
    (dbf-in-summary-buffer
      (let (buffer-read-only)
        (erase-buffer)
        (db-message "Computing summary...")
        (db-lmap-macro
         (progn
           (when (or redo-p (not (link-summary db-lmap-link)))
             (setf (link-summary db-lmap-link)
                   (funcall sum (link-record db-lmap-link))))
           (dbs-insert-link-summary db-lmap-link mhp))
         dbc-database hide "Computing summary...%d")
        (db-message "Computing summary...done")
        ;; get rid of last newline.
        (backward-delete-char 1)
        (set-buffer-modified-p nil)
        (setq dbs-no-of-records (database-no-of-records dbc-database))
        ;; mernst sez: What is this doing?  lines is bound to dbfs-lines
        ;;             above, and hasn't been changed (except perhaps
        ;;             dynamically??)  since.
        ;; ttn sez: wtf?
        (setq dbfs-lines lines)
        (setq dbs-recompute-p nil)
        ;; fixme: move to proper record, taking into account hidden, etc. --ttn
        (setq dbs-index 0)))
    (setq dbf-summary-recompute-all-p nil)))

(defun dbf-fill-summary-buffer-and-move-to-proper-record ()
  (when (dbf-summary-buffer)
    (dbf-fill-summary-buffer)
    (dbf-in-summary-buffer
      (dbs-move-to-proper-record))))

(defun dbf-update-summary-marks ()
  ;; Update just the marked and hidden summary markings efficiently.
  (dbf-in-summary-buffer
    (let ((opoint (point))
          (hidep (not (or dbf-summary-show-hidden-records-p
                          (not dbc-hide-p))))
          line)
      (unwind-protect
          (let (buffer-read-only)
            (goto-char (point-min))
            (db-lmap-macro
             (progn
               (delete-char 1)
               (insert (if (link-markedp db-lmap-link) "+" " "))
               (backward-char 1)
               (setq line 0)
               ;; Each summary item spans exactly dbfs-lines screen lines.
               (while (< line dbfs-lines)
                 (forward-char 1)
                 (delete-char 1)
                 (insert (if (and dbc-hide-p
                                  (link-hiddenp db-lmap-link))
                             "["
                           " "))
                 (forward-line 1)
                 (incf line)))
             dbc-database
             hidep))
        (goto-char opoint)))))


(defun dbf-update-summary-item (index &optional link)
  ;; Update just changes to one record in the summary efficiently.
  (unless link
    (setq link (car (db-next-link-and-index
                     dbc-database (database-first-link dbc-database)
                     1 (1- index) nil nil nil))))
  (let ((sum dbf-summary-function))
    (when (or (not (link-hiddenp link))
              dbf-summary-show-hidden-records-p
              (not dbc-hide-p))
      (dbf-in-summary-buffer
        (let ((orig dbs-index))
          (unwind-protect
              (let (buffer-read-only)
                (or (link-summary link)
                    (setf (link-summary link)
                          (funcall sum (link-record link))))
                (dbs-move-to-proper-record index)
                ;; assuming at beginning of line
                (delete-region (point)
                               (progn
                                 (forward-line dbfs-lines)
                                 (point)))
                (dbs-insert-link-summary link dbc-hide-p))
            ;; save old line and column instead.
            (dbs-move-to-proper-record orig))))
      ;; fixme: update cached summary. --ttn
      )))


(defun dbf-make-summary-maker (summary-format database)
  (let ((lasfl (db-format->lines/sforms summary-format database 2 t nil)))
    (setq dbfs-lines (car lasfl)
          dbf-summary-function `(lambda (formatted-record)
                                  (concat ,@(cdr lasfl))))))


(defun db-format->lines/sforms (format database indent nlp vheightp)
  ;; Take a format and return a cons of two values: a number and a list.
  ;; The list is list of forms which, when evaluated with variable
  ;; `formatted-record' bound, evaluate to strings; these can be used as
  ;; argumentes to concat, insert, etc.  The number is the number of lines
  ;; occupied by the items when inserted.
  ;;
  ;; Signal an error if any displayspec has nonequal min-height and
  ;; max-height, unless VHEIGHTP is non-nil, in which case the number
  ;; returned is a minimum.  NLP non-nil means add a newline.
  (let ((ph (and (string-match "\\\\\\\\" format) ; backslash place holder
                 (db-unused-char-in-string format)))
        (lines 0)
        results beg end ds minh maxh)
    (when (and indent (> indent 0))
      (setq format (concat (make-string indent 32)
                           (replace-regexp-in-string
                            "\n"
                            (concat "\n" (make-string indent 32))
                            format))))
    (when ph
      (setq format (replace-regexp-in-string
                    (regexp-quote "\\\\")
                    (char-to-string ph)
                    format)))

    (while (string-match db-dsrx format)
      (setq beg (match-beginning db-dsrx-content-beginning)
            end (or (match-end db-dsrx-content-end)
                    (match-end db-dsrx-content-end-alt))
            ds (db-dspec<-string format database)
            minh (displayspec-min-height ds)
            maxh (displayspec-max-height ds))
      ;; fixme: should not be necessary. --ttn
      (unless minh
        (setf minh (setf (displayspec-min-height ds) 1)))
      (unless (or vheightp maxh)
        (setf maxh (setf (displayspec-max-height ds) minh)))
      (if (or vheightp (= minh maxh))
          (incf lines (1- minh))
        (error "Min-height %s must equal max-height %s in summary displayspec %s."
               minh maxh ds))
      (unless (zerop beg)
        (let ((literal (substring format 0 beg)))
          (when ph
            (subst-char-in-string ph ?\\ literal t))
          (push literal results)
          (incf lines (db-count-newlines literal))))
      (push `(db-ds-printed ,ds formatted-record) results)
      (setq format (substring format end)))
    (when nlp (setq format (concat format "\n")))
    (unless (equal "" format)
      (when ph
        (subst-char-in-string ph ?\\ format t))
      (push format results))
    (incf lines (db-count-newlines format))
    (cons lines (nreverse results))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Synching the format and summary buffers
;;;

(defsubst dbs-in-synch-p ()
  (= dbs-index (dbs-in-data-display-buffer dbc-index)))

(defsubst dbs-out-of-date-p ()
  (or dbs-recompute-p
      (not (= dbs-no-of-records (database-no-of-records dbc-database)))))

(defsubst dbf-set-summary-out-of-date-p ()
  (dbf-in-summary-buffer
    (setq dbs-recompute-p t)))

(defun dbs-synch-format-with-summary ()
  ;; Ensure that the data display and summary buffers
  ;; have the same current record.
  (if (dbs-out-of-date-p)
      (dbs-synch-summary-with-format)
    (unless (dbs-in-synch-p)
      (dbs-in-data-display-buffer
        (db-select-record (with-current-buffer dbf-summary-buffer
                            dbs-index))))))

(defun dbs-synch-summary-with-format ()
  (when (dbs-out-of-date-p)
    (dbs-in-data-display-buffer
      (dbf-fill-summary-buffer)))
  ;; If we just did the above, it will clearly be out of synch.
  ;; But it might be even if it wasn't out of date.
  (unless (dbs-in-synch-p)
    (dbs-move-to-proper-record)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Moving about
;;;

(defsubst dbs-goto-nth-summary (n)
  ;; NOTE: If hidden records aren't shown in the summary (and that should be
  ;;       an option), then this is wrong.  And in that case it's better to do
  ;;       relative than absolute motion.
  (goto-line (1+ (* dbfs-lines (1- n))))
  (setq dbs-point (point)))

(defun dbs-move-to-proper-record (&optional index)
  "Move point to the summary of the record shown in the format or to INDEX."
  ;; Make no assumptions about dbs-index.
  (if (dbs-in-data-display-buffer (or (not dbc-hide-p)
                                      dbf-summary-show-hidden-records-p))
      (let ((index (or index (dbs-in-data-display-buffer dbc-index))))
        (dbs-goto-nth-summary index)
        (dbs-set-index index))
    (let ((prev 0)
          (last nil)
          (pidx (or index               ; proper index
                    (dbs-in-data-display-buffer dbc-index))))
      (db-lmap-macro
       (if (<= db-lmap-index pidx)
           (setq prev (1+ prev)
                 last db-lmap-index)
         ;; Past it but still haven't found a nonhidden link.
         (unless last
           (setq prev 1
                 last db-lmap-index)))
       dbc-database
       t)
      ;; If there are no displayed records at all, this will fail.
      (unless (= last pidx)
        (db-message "Record %s does not appear in the summary buffer"
                    pidx))
      (dbs-goto-nth-summary prev)
      (dbs-set-index last))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Movement commands
;;;

(defsubst dbs-forward-record (arg)
  ;; Move point forward ARG records in the summary buffer, and set dbs-point.
  (goto-char dbs-point)
  (db-forward-line-wrapping (* dbfs-lines arg))
  (setq dbs-point (point)))

(defun dbs-next-record-ignore-hiding (arg)
  "Go to the ARGth next record, ignoring hiding.
That is, all records, even those which are hidden, are counted."
  (interactive "p")
  (if (not (dbs-in-data-display-buffer dbf-summary-show-hidden-records-p))
      (db-next-record-ignore-hiding arg)
    (dbs-synch-format-with-summary)
    (dbs-in-data-display-buffer
      (db-next-record-ignore-hiding arg))
    (dbs-forward-record arg)
    (dbs-set-index (dbs-in-data-display-buffer dbc-index))))

(defun dbs-previous-record-ignore-hiding (arg)
  "Go to the ARGth previous record, ignoring hiding.
That is, all records, even those which are hidden, are counted."
  (interactive "p")
  (dbs-next-record-ignore-hiding (- arg)))

(defun dbs-scroll-up ()
  (interactive)
  (scroll-up)
  (db-jump-to-point))

(defun dbs-scroll-down ()
  (interactive)
  (scroll-down)
  (db-jump-to-point))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Summary Mode commands
;;;

(defvar database-summary-mode-map nil
  "Keymap for database summary buffer.")

(unless database-summary-mode-map
  (let ((m (make-sparse-keymap)))
    (suppress-keymap m)
    (mapc (lambda (key-def)
            (define-key m
              (let ((key (car key-def)))
                (if (consp key)
                    (db-meta-prefix-ify (car key))
                  key))
              (cdr key-def)))
          '(("t" . toggle-truncate-lines)

            ;; Moving around in the database
            ("n"      . db-next-record)
            ("p"      . db-previous-record)
            ("\C-n"   . db-next-record)
            ("\C-p"   . db-previous-record)
            ("<"      . db-first-record)
            (">"      . db-last-record)
            (("<")    . db-first-record)
            ((">")    . db-last-record)
            ("j"      . db-jump-to-record)
            (" "      . db-next-screen-or-record)
            ("\177"   . db-previous-screen-or-record)
            (("n")    . dbs-next-record-ignore-hiding)
            (("p")    . dbs-previous-record-ignore-hiding)
            (("\C-n") . db-next-marked-record)
            (("\C-p") . db-previous-marked-record)

            ;; Exiting summary mode
            ("e" . dbs-edit)
            ("v" . dbs-view)
            ("q" . dbs-exit)
            ("x" . dbs-exit)

            ;; Adding and removing records
            ("a" . db-add-record)
            ("i" . db-add-record)
            ("d" . dbs-delete-record)
            ("k" . dbs-delete-record)
            ("o" . dbs-output-record-to-db)
            ("c" . db-copy-record)

            ;; Searching commands
            ("s"    . db-search)
            ;;("S"  . db-incremental-search)
            ("\C-s" . db-isearch-forward)
            ("\C-r" . db-isearch-backward)

            ;; Everything else
            ("?"      . describe-mode)
            ("O"      . db-hide-record)
            (("o")    . db-hiding-toggle)
            (("O")    . db-hiding-set)
            (("\C-o") . db-toggle-show-hidden-records)
            ("g"      . db-summary)
            ("h"      . db-summary)
            ("D"      . db-summary)
            ("m"      . db-mark-record)
            ("r"      . db-report)
            ("\C-xr"  . db-revert-database)
            ("\C-v"   . dbs-scroll-up)
            (("v")    . dbs-scroll-down)

            ("\C-x\C-q" . db-toggle-modifiable-p)

            ("\C-c\C-c" . dbs-exit)

            ("b"   . undefined)
            ("f"   . undefined)
            ("l"   . undefined)
            ;;("u" . db-revert-record)
            ("w"   . undefined)
            ("y"   . undefined)
            ("z"   . undefined)))
    (setq database-summary-mode-map m)))


(defun dbs-view ()
  "Manipulate this record in the data display buffer in View mode."
  (interactive)
  (pop-to-buffer dbs-data-display-buffer)
  (db-view-mode))

(defun dbs-edit ()
  "Manipulate this record in the data display buffer in Edit mode."
  (interactive)
  (pop-to-buffer dbs-data-display-buffer)
  (when (eq dbf-minor-mode 'view)
    (db-first-field)))

(defun dbs-exit ()
  "Exit the summary buffer."
  (interactive)
  ;; This is oh-so-very-crude.
  (let ((buf dbs-data-display-buffer))
    (delete-windows-on (current-buffer))
    (switch-to-buffer buf)))

(defun dbs-delete-record (&optional force)
  "Delete the current record from the database.
With a prefix argument, doesn't verify."
  (interactive "P")
  (when (or force (y-or-n-p "Delete this record? "))
    (dbs-in-data-display-buffer
      (db-delete-record t))
    ;; hope we're at the beginning of the record
    (let ((buffer-read-only nil))
      (kill-line dbfs-lines)
      (when (eobp)
        (goto-char (point-min))))
    (decf dbs-no-of-records)
    (db-message "Record deleted")
    (dbs-set-index (dbs-in-data-display-buffer dbc-index))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Menus
;;;

(defvar database-summary-mode-menu
  '("Database"
    "SUMMARY Mode:"
    ["Update summary"     db-summary t]
    ["Report"             db-report  t]
    "----"
    ("Motion"
     ["jump  to record"  db-jump-to-record t]
     ["last     record"  db-last-record    t]
     ["first    record"  db-first-record   t]
     "---"
     ["next"               db-next-record                 t]
     ["next (screen)"      db-next-screen-or-record       t]
     ["next (marked)"      db-next-marked-record          t]
     ["next (ingore hiding)" db-next-record-ignore-hiding t]
     "---"
     ["prev"               db-previous-record                 t]
     ["prev (screen) "     db-previous-screen-or-record       t]
     ["prev (marked)"      db-previous-marked-record          t]
     ["prev (ingore hiding)" db-previous-record-ignore-hiding t]
     "---"
     ["Isearch Backward" db-isearch-backward t]
     ["Isearch Forward" db-isearch-forward t]
     )
    "----"
    ["View record" dbs-view t]
    ["Edit record" dbs-edit t]
    ["Delete Record" dbs-delete-record t]
    ["Add Record" db-add-record t]
    ["Mark Record" db-mark-record t]
    ["Hide Record" db-hide-record t]
    "----"
    ("Hiding"
     ["Hiding    on/off" db-hiding-toggle             t]
     ["Hiding hide/show (in summary)" db-toggle-show-hidden-records t]
     ["Un-hide      all" db-unhide-all                  t]
     ["Un-mark      all" db-unmark-all                  t]
     ["Mark   un-hidden" db-mark-unhidden-records      t]
     ["Hide   un-marked" db-hide-unmarked-records       t]
     )
    "----"
    ["Sort     database" db-sort                t]
    ["Revert   database" db-revert-database t]
    ["Save     database" db-save-database       t]
    ["Write    database" db-write-database-file t]
    "----"
    ["Quit" dbs-exit t]
    )
  "Menu for Database Summary mode.")

;; 'ignored for SYMBOL argument was giving me trouble.
;; Does this work in Lucid Emacs?
(easy-menu-define ignored
  database-summary-mode-map
  "ignored-doc-string"
  database-summary-mode-menu)

;;; db-summary.el ends here
