;;; db-tagged.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael A. Patton <MAP@lcs.mit.edu>
;;	Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Provide support for files of tagged fields, such as
;;  Name:  John Doe
;;  Age:   42
;;  Salary:22000


;; This file adds to EDB basic support for parsing and generating
;; files where records consist of a group of lines with tags and
;; values.  All functions and variables defined here start with the
;; prefix "db-tagged-".

;; The basic way to use this is to call `db-tagged-setup' with a list
;; describing the fields.  There are hooks to parse out and generate
;; possible added entries that don't fit the basic model.  There are
;; variables to customize all the separator strings and regexps for
;; complete parsing.  These should be set in the aux file after calling
;; `db-tagged-setup'.

;; The field info passed to `db-tagged-setup' is a list with one entry
;; per field.  Each entry is a three-element list: the field name, the
;; tag used to identify it in the file, and a short description for
;; documentation (not presently used).  The field name can be a two
;; element list with a field name and type to use types other than the
;; default.  Two values in the tag position are special, nil means that
;; it will be a computed field and is never stored in the database (and
;; thus doesn't need a tag), and the empty string marks a field where
;; "comments" are collected, comments consist of lines starting with the
;; separator, i.e. empty tags.

;; The default is records separated by blank lines, tags separated from
;; fields by ":", white space around the separator is not significant on
;; input, and that the separator should be followed by one tab on
;; output, and continuation lines start with whitespace.  All of these
;; can be customized.

;;; Code:


(require 'database)
(eval-when-compile (require 'cl))
(provide 'db-tagged)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variables
;;;


(defvar database)                       ; used dynamically


;;; These are the variables used for customization, these are actually
;;; database-local variables and not real variables.  The defvars are
;;; mainly to get documentation somewhere...the values are used as
;;; defaults when initializing the database.

(defvar db-tagged-tag-chars "-_A-Za-z"
  "The characters that are allowed in field tags, in a form suitable for
placing inside [] in a regular expression.")

(defvar db-tagged-pre-tag ""
  "The string that precedes each tag.
Used only if  db-tagged-pre-tag-regexp  or  db-tagged-pre-tag-output
nil \(depending on whether the record is being read or written\).")

(defvar db-tagged-pre-tag-regexp nil
  "A regexp for what precedes each tag on its line when parsing.
It must NOT contain any grouping constructs.")

(defvar db-tagged-pre-tag-output nil
  "The fixed string to place before each tag on output.")

(defvar db-tagged-separator ":"
  "The string that separates field names from values.
Used only if `db-tagged-separator-regexp' or `db-tagged-separator-output' is
nil \(depending on whether the record is being read or written\).")

(defvar db-tagged-separator-regexp nil
  "A regexp for the separator between field names and values when parsing.")

(defvar db-tagged-separator-output nil
  "The separator between field names and values on output.")

(defvar db-tagged-continuation "\t"
  "The string that marks (the beginning of) a continuation line.
Used only if `db-tagged-continuation-regexp' or `db-tagged-continuation-output'
is nil \(depending on whether the record is being read or written).")

(defvar db-tagged-continuation-regexp "[ \t]+"
  "A regexp for a continuation line in a value when parsing.")

(defvar db-tagged-continuation-output nil
  "The fixed string to use (before) continuing values on output.")

(defvar db-tagged-rrfr-hooks ()
  "Hooks run on each database record before tagged parse.")

(defvar db-tagged-wrfr-before-hooks ()
  "Hooks run before each tagged write of a database record.
The record is bound to the dynamic variable  record, and point is where the
record will be inserted in the buffer.")

(defvar db-tagged-wrfr-after-hooks ()
  "Hooks run after each tagged write of a database record.
The record is bound to the dynamic variable  record, and point is immediately
after the file representation of the record.")

(defvar db-tagged-converted-p ()
  "Non-nil if `database-stored->actual' has already been run.")

(defvar db-tagged-default-field nil
  "A fieldname (symbol) for the field indicated by an illegal or empty tag.
\(For instance, you might set it to 'comments.\)
Nil if those values should be discarded.")


;;; Abstraction of the tagged field specification.
(defsubst db-tagged-spec-name (tfs) (car tfs))
(defsubst db-tagged-spec-tag (tfs) (car (cdr tfs)))
(defsubst db-tagged-spec-help (tfs) (car (cdr (cdr tfs))))


;;;###autoload
(defun db-tagged-setup (fspecs &rest attrs)
  "Ready the database to read files in tagged format.
Argument FSPECS is a list of tagged-field specifications, one
for each field in a database record.  Each tagged-field specification is a
three-element list of the field name \(a symbol\), the tag used to identify
it in the file \(a string\), and a brief help string.
Instead of a symbol, the tagged-field name may be a cons of the
field name and its type.
To indicate that a field is never found in the input file \(typically
because it is computed on the fly\), use nil for its tag.
ATTRS is a sequence of alternating keywords and values specifying
overriding attributes.

Note: Do not call `database-set-fieldnames-to-list' if you call this function."
  ;; First, try to detect whether this database was once in tagged file
  ;; layout, but is now in internal file layout.
  (if (db-tagged-p database)
      (unless (database-get-local 'db-tagged-converted-p database t)
        (database-set-local 'db-tagged-converted-p database t))
    (database-make-local 'db-tagged-field-spec database fspecs)
    (let ((vstems '(rrfr-hooks
                    wrfr-before-hooks
                    wrfr-after-hooks
                    tag-chars
                    pre-tag
                    pre-tag-regexp
                    pre-tag-output
                    separator
                    separator-regexp
                    separator-output
                    continuation
                    continuation-regexp
                    continuation-output
                    converted-p)))
      (mapc (lambda (vstem)
              (let ((var (intern (format "db-tagged-%s" vstem))))
                (database-make-local var database (symbol-value var))))
            vstems)

      (database-set-fieldnames-to-list database
        (mapcar 'db-tagged-spec-name fspecs)
        ;; Most fields are strings (or missing, here represented as nil)
        'nil-or-string)

      ;; Insert the help strings into *copies* of the recordfieldspecs in
      ;; order to not redefine the prototype fieldtypes.
      (dolist (fspec fspecs)
        (let ((fn (db-tagged-spec-name fspec))
              (help (db-tagged-spec-help fspec))
              fnum rs ohelp)
          (when (consp fn)
            (setq fn (car fn)))
          (setq fnum (fieldname->fieldnumber fn database)
                rs (copy-recordfieldspec
                    (database-recordfieldspec database fnum))
                ohelp (recordfieldspec-help-info rs))
          (setf (recordfieldspec-help-info rs) (if ohelp
                                                   (concat help "\n" ohelp)
                                                 help))
          (database-set-recordfieldspec database fnum rs)))
      (setf (database-read-record-from-region database) 'db-tagged-rrfr
            (database-write-region-from-record database) 'db-tagged-wrfr
            (sepinfo-sep-string (database-record-sepinfo database)) "\n\n"
            (sepinfo-post-last-string (database-record-sepinfo database)) "\n")
      (add-hook 'db-before-read-hooks 'db-tagged-before-read)
      (add-hook 'db-after-read-hooks 'db-tagged-database-stored->actual)

      ;; Override the defaults.
      (while attrs
        (let ((nam (intern (substring (symbol-name (car attrs)) 1)))
              (val (cadr attrs)))
          (unless (memq nam vstems)
            (error "No such db-tagged attribute: %s" nam))
          (database-set-local (intern (format "db-tagged-%s" nam))
            database val))
        (setq attrs (cddr attrs))))))

(defsubst db-tagged-p (database)
  "Non-nil if DATABASE is in tagged file layout."
  (database-get-local 'db-tagged-field-spec database t))

(defun db-tagged-database-stored->actual (&optional db)
  "Like `database-stored->actual', but only in tagged file layout."
  (when (and (db-tagged-p (or db database))
             (not (database-get-local 'db-tagged-converted-p database t)))
    (database-stored->actual db)
    (database-set-local 'db-tagged-converted-p database t)))

(defun db-tagged-before-read ()
  (when (db-tagged-p database)
    (unless (database-get-local 'db-tagged-pre-tag-regexp database)
      (database-set-local 'db-tagged-pre-tag-regexp database
        (regexp-quote
         (database-get-local 'db-tagged-pre-tag database))))
    (unless (database-get-local 'db-tagged-pre-tag-output database)
      (database-set-local 'db-tagged-pre-tag-output database
        (database-get-local 'db-tagged-pre-tag database)))
    (unless (database-get-local 'db-tagged-separator-regexp database)
      (database-set-local 'db-tagged-separator-regexp database
        (concat "[ \t]*"
                (regexp-quote
                 (database-get-local 'db-tagged-separator database))
                "[ \t]*")))
    (unless (database-get-local 'db-tagged-separator-output database)
      (database-set-local 'db-tagged-separator-output database
        (concat (database-get-local 'db-tagged-separator database)
                "\t")))
    (unless (database-get-local 'db-tagged-continuation-regexp database)
      (database-set-local 'db-tagged-continuation-regexp database
        (regexp-quote
         (database-get-local 'db-tagged-continuation database))))
    (unless (database-get-local 'db-tagged-continuation-output database)
      (database-set-local 'db-tagged-continuation-output database
        (database-get-local 'db-tagged-continuation database)))))


;; The buffer is narrowed to the region which is one record.
(defun db-tagged-rrfr ()
  (let* ((rv (make-record database))
         (crx (database-get-local 'db-tagged-continuation-regexp database))
         (trx (concat
               "^" (database-get-local 'db-tagged-pre-tag-regexp database)
               "\\(["
               (database-get-local 'db-tagged-tag-chars database)
               "]*\\)"
               (database-get-local 'db-tagged-separator-regexp database)))
         nam fld prev)
    (let ((hooks (database-get-local 'db-tagged-rrfr-hooks database)))
      (run-hooks 'hooks))
    (goto-char (point-min))
    (while (not (eobp))
      (cond ((looking-at trx)
             (setq nam (match-string 1)
                   fld (let ((r nil))
                         (dolist (x (database-get-local
                                     'db-tagged-field-spec database))
                           (when (equal nam (car (cdr x)))
                             (setq r (car x))))
                         (if (consp r)
                             (car r)
                           r)))
             (end-of-line)
             (let ((val (buffer-substring (match-end 0) (point))))
               ;; Why isn't this (forward-char 1)?
               (forward-line 1)
               (while (looking-at crx)
                 (end-of-line)
                 (setq val (concat val "\n"
                                   (buffer-substring (match-end 0)
                                                     (point))))
                 (forward-line 1))
               (when (null fld)
                 ;; Should allow an escape hook here, and provide a
                 ;; generic one that adds to an alist-like entry
                 (db-message "Invalid field name `%s'" nam)
                 (setq fld db-tagged-default-field))
               (setq prev (record-field rv fld database))
               (when fld
                 (record-set-field rv fld
                                   (if (null prev)
                                       val
                                     (concat prev "\n" val))
                                   database))))
            (t
             (error "At char %d, %s." (point)
                    "didn't find either a tag or the end of the record"))))
    rv))

(defun db-tagged-wrfr (record)
  (let ((hooks (database-get-local 'db-tagged-wrfr-before-hooks database)))
    (run-hooks 'hooks))
  (let ((sep (database-get-local 'db-tagged-separator-output database))
        (contin (database-get-local 'db-tagged-continuation-output database)))
    (mapcar (lambda (def)
              (let ((nam (db-tagged-spec-name def))
                    (tag (db-tagged-spec-tag def)))
                ;; Handle case of typed field
                (when (consp nam)
                  (setq nam (car nam)))
                (unless (null tag)
                  ;; Fields without tags are computed, so don't get stored
                  (let* ((fno (fieldname->fieldnumber nam database))
                         (a->s (recordfieldspec-actual->stored
                                (database-recordfieldspec database fno)))
                         (val (db-funcall-maybe a->s (aref record fno))))
                    (unless (or (null val) (equal "" val))
                      (let ((i 0)
                            (cnt (if (equal "" tag)
                                     sep
                                   contin))
                            j)
                        (insert tag sep)
                        (while (setq j (string-match "\n" val i))
                          (insert (substring val i j) "\n" cnt)
                          (setq i (+ j 1)))
                        (insert (substring val i) "\n")))))))
            (database-get-local 'db-tagged-field-spec database)))
  (delete-char -1)                      ; HACK: punt last newline;
                                        ; it will be added back later.
  (let ((hooks (database-get-local 'db-tagged-wrfr-after-hooks database)))
    (run-hooks 'hooks)))

;;; db-tagged.el ends here
