;;; db-two-dbs.el --- part of EDB, the Emacs database

;; Copyright (C) 2004,2005 Thien-Thi Nguyen

;; Author: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Support for actions on two databases.

;;; Code:


(eval-when-compile (require 'cl))
(provide 'db-two-dbs)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Process two databases
;;;

;;;###autoload
(defun db-process-two-databases (db1 db2 lone1 lone2 corr &optional orderer)
  (db-with-orderer db1
    (unless (database-ordered-p db1 orderer)
      (db-message "Sorting %S..." (database-print-name db1))
      (database-sort db1 orderer))
    (unless (database-ordered-p db2 orderer)
      (db-message "Sorting %S..." (database-print-name db2))
      (database-sort db2 orderer))

    ;; Perhaps check for identical keys here.

    (db-message "Databases are properly ordered")
    (let* ((fl1 (database-first-link db1))
           (ln1 fl1)
           (fl2 (database-first-link db2))
           (ln2 fl2)
           (done1 (database-empty-p db1))
           (done2 (database-empty-p db2))
           r1 r2 r-order)
      (while (not (or done1 done2))
        (setq r1 (link-record ln1)
              r2 (link-record ln2)
              r-order (funcall orderer r1 r2))
        (cond ((= -1 r-order)
               (funcall lone1 r1)
               (setq ln1 (link-next ln1))
               (when (eq ln1 fl1) (setq done1 t)))
              ((= 1 r-order)
               (funcall lone2 r2)
               (setq ln2 (link-next ln2))
               (when (eq ln2 fl2) (setq done2 t)))
              ((= 0 r-order)
               (funcall corr r1 r2)
               (setq ln1 (link-next ln1)
                     ln2 (link-next ln2))
               (when (eq ln1 fl1) (setq done1 t))
               (when (eq ln2 fl2) (setq done2 t)))
              (t
               (error "Bad result %s from orderer." r-order))))
      (while (not done1)
        (funcall lone1 (link-record ln1))
        (setq ln1 (link-next ln1))
        (when (eq ln1 fl1) (setq done1 t)))
      (while (not done2)
        (funcall lone2 (link-record ln2))
        (setq ln2 (link-next ln2))
        (when (eq ln2 fl2) (setq done2 t)))
      nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Merge
;;;

;; Want to display both records in full and, for each differing field,
;; ask how to set the merged database: from one, from the other, or by
;; entering something particular.  It should also be possible to edit
;; the merged record before proceeding to the next one.

;; Should check for the consistency of the two databases.  Should permit
;; conversion for one or both databases as part of this procedure.

;;;###autoload
(defun db-merge ()
  "Merge two databases chosen with completion among read-in databases."
  (interactive)
  (let ((avail (mapcar (lambda (database)
                         (cons (database-print-name database)
                               database))
                       db-databases))
        db1 db2 db3 buf1 buf2 buf3)
    (cond ((< (length avail) 2)
           (error "db-merge requires there to be at least two read-in databases."))
          ((= 2 (length avail))
           (setq db1 (car db-databases)
                 db2 (car (cdr db-databases))))
          (t
           ;; Could check that a data display buffer exists.
           (setq db1 (assoc (completing-read
                             "First database to merge?  [? for options] "
                             avail nil t nil)
                            avail))
           (setq avail (delq db1 avail))
           (setq db1 (cdr db1))
           ;; fixme: err if not compatabile.
           (setq db2 (cdr (assoc (completing-read
                                  "Second database to merge?  [? for options] "
                                  avail nil t nil)
                                 avail)))))
    (setq buf1 (car (database-clean-data-display-buffers db1))
          buf2 (car (database-clean-data-display-buffers db2)))
    (setq db3 (let ((result (copy-database db1)))
                (callf2 concat "Copy of " (database-print-name db1))
                (setf (database-first-link result) nil
                      (database-no-of-records result) 0
                      ;; Should this go after db-choose-format-file?
                      (database-file result) (concat (database-file db1) "-COPY"))
                (setf (database-data-display-buffers result)
                      (list (db-setup-data-display-buffer
                             (db-choose-format-file result nil nil)
                             result
                             t)))
                (push result db-databases)
                result)
          buf3 (car (database-clean-data-display-buffers db3)))
    (setf (database-print-name db3)
          (concat "Merge of `" (database-print-name db1)
                  "' and `" (database-print-name db2) "'"))
    (let ((idx1 (with-current-buffer buf1 dbc-index))
          (idx2 (with-current-buffer buf2 dbc-index)))
      (delete-other-windows)
      (switch-to-buffer buf1)
      (split-window-vertically (/ (window-height) 2))
      (other-window 1)
      (switch-to-buffer buf2)
      (split-window-vertically)
      (other-window 1)
      (switch-to-buffer buf3)
      (set-buffer buf1)
      (db-process-two-databases
       db1 db2
       (lambda (r) (database-add-record r db3))
       (lambda (r) (database-add-record r db3))
       (lambda (r1 r2)
         (database-add-record (db-merge-records r1 r2 db3 buf1 buf2 buf3)
                              db3)))
      (db-in-buffer buf1 (db-jump-to-record idx1))
      (db-in-buffer buf2 (db-jump-to-record idx2))
      (db-in-buffer buf3 (db-jump-to-record 1))))
  (message "Done merging."))


(defun db-merge-records (r1 r2 database buf1 buf2 buf3)
  (if (equal r1 r2)
      r1
    (set-buffer buf1)
    (db-display-record r1 t)
    (set-buffer buf2)
    (db-display-record r2 t)
    (set-buffer buf3)
    (let ((r3 (make-record database))
          (fno 0)
          (ds1-all (with-current-buffer buf1 dbf-displayspecs))
          (ds2-all (with-current-buffer buf2 dbf-displayspecs))
          v1 v2 ds1 ds2 rs)
      (while (< fno (database-no-of-fields database))
        (setq ds1 (aref ds1-all fno)
              ds2 (aref ds2-all fno)
              rs (database-recordfieldspec database fno)
              v1 (db-funcall-maybe (recordfieldspec-common-form-function rs)
                                   (aref r1 fno))
              v2 (db-funcall-maybe (recordfieldspec-common-form-function rs)
                                   (aref r2 fno)))
        (if (equal v1 v2)
            (record-set-field-from-index r3 fno v1 nil)
          (let ((fieldname (db-fname<-fno fno database)))
            ;; Problem: nil might not be valid in some places.
            ;; For now, ignore that.
            (db-display-record r3 t fno)
            ;; Does no constraint checking; is this the right thing to do?
            ;; Help for one-char-question is not entirely satisfactory here.
            (record-set-field-from-index
             r3 fno
             (let ((fname (db-fname<-fno fno database)))
               (cond ((y-or-n-p
                       (format "Use first value for %s field? [%s] "
                               fname
                               (db-actual->display-call
                                (displayspec-actual->display ds1)
                                v1
                                nil nil)))
                      v1)
                     ((y-or-n-p
                       (format "Use second value for %s field? [%s] "
                               fname
                               (db-actual->display-call
                                (displayspec-actual->display ds2)
                                v2
                                nil nil)))
                      v2)
                     (t
                      (if (equal (displayspec-display->actual ds1)
                                 (displayspec-display->actual ds2))
                          (db-display->actual-call
                           (displayspec-display->actual ds1)
                           (read-from-minibuffer
                            (format "Enter value for %s field: " fname))
                           nil nil nil)
                        (error "displayspecs have different display->actual")))))
             nil)))
        (incf fno))
      r3)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Compare
;;;

;;;###autoload
(defun databases-compatible (db1 db2)
  "Return t if the database records have the same field names and type,
nil otherwise."
  ;; Should eventually check types as well.
  (let ((rv t)
        (fno 0)
        (count1 (length (database-recordfieldspecs db1)))
        (count2 (length (database-recordfieldspecs db2)))
        rs1 rs2)
    (when (= count1 count2)
      (while (and rv (< fno count1))
        (setq rs1 (database-recordfieldspec db1 fno)
              rs2 (database-recordfieldspec db2 fno)
              ;; used to also check recordfieldspec-name here
              rv (eq (recordfieldspec-type rs1)
                     (recordfieldspec-type rs2)))
        (incf fno))
      rv)))

(defun database-compare-hack (db1 db2)
  ;; Check that fieldnames and types are the same.
  (unless (databases-compatible db1 db2)
    (error "Incompatible databases."))
  (let* ((name1 (or (database-print-name db1) "First Database"))
         (name2 (or (database-print-name db2) "Second Database"))
         (loners1 (get-buffer-create (concat "Loners for " name1)))
         (loners2 (get-buffer-create (concat "Loners for " name2)))
         (discrep (get-buffer-create (concat "Discrepancies between "
                                             name1 " and " name2))))
    (dolist (buf (list loners1 loners2 discrep))
      (db-in-buffer buf (erase-buffer)))
    (flet ((print-r (r db)
                    (let ((names (database-fieldnames db)))
                      (princ "\n")
                      (dotimes (fno (database-no-of-fields db))
                        (princ (format "%s:  %s\n" (aref names fno)
                                       (aref r fno))))))
           (print-l1 (r) (db-in-buffer loners1 (print-r r db1)))
           (print-l2 (r) (db-in-buffer loners2 (print-r r db2))))
      (db-process-two-databases
       db1 db2 'print-l1 'print-l2
       (lambda (r1 r2)
         ;; We already know that order-function considers the two
         ;; records the same; now we need to check whether any of
         ;; their fields differ and if so, report it.
         (unless (equal r1 r2)
           (db-in-buffer discrep
             (let (fname ofunc v1 v2)
               (princ "\n")
               (dotimes (fno (database-no-of-fields db1))
                 (setq fname (db-fname<-fno fno db1)
                       ofunc (db-rs-ordfunc (database-recordfieldspec db1 fno))
                       v1    (aref r1 fno)
                       v2    (aref r2 fno))
                 (cond ((equal v1 v2)
                        (princ (format "  %s:  %s\n" fname v1)))
                       ((and ofunc (zerop (funcall ofunc v1 v2)))
                        (princ (format ". %s:  %s\n. %s:  %s\n"
                                       fname v1 fname v2)))
                       (t
                        (princ (format "* %s:  %s\n* %s:  %s\n"
                                       fname v1 fname v2)))))))))))))

;;; db-two-dbs.el ends here
