;;; edb-t-human-names.el

;; Copyright (C) 2005 Thien-Thi Nguyen

;; Authors: Michael Ernst <mernst@theory.lcs.mit.edu>

;; EDB is distributed under the terms of the GNU General Public License.

;; EDB is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY.  No author or distributor accepts responsibility to anyone
;; for the consequences of using it or for whether it serves any particular
;; purpose or works at all, unless he says so in writing.  Refer to the GNU
;; General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute EDB, but
;; only under the conditions described in the GNU General Public License.  A
;; copy of this license is supposed to have been given to you along with EDB
;; so you can know your rights and responsibilities.  It should be in a file
;; named COPYING.  If not, write to the Free Software Foundation, Inc., 675
;; Mass Ave, Cambridge, MA 02139, USA for a copy.  Among other things, the
;; copyright notice and this notice must be preserved on all copies.

;;; Commentary:

;; Names for human beings
;;
;; These variables and functions used to be in db-types.el,
;; without the "edb-t-human-names:" prefix.
;;
;; TODO: Internationalize (remove/reduce English bias).

;;; Code:

;;
;; Last names
;;

(defun edb-t-human-names:order-last-names (n1 n2)
  "Return -1, 0, or 1 depending on whether last name N1 is lexographically
less than, equal to, or greater than N2."
  ;; Note stringency of equality test:  capitalization and spacing matter.
  (cond ((equal n1 n2)
         0)
        ((string-lessp (edb-t-human-names:canonicalize-name n1)
                       (edb-t-human-names:canonicalize-name n2))
         -1)
        (t
         1)))

(defun edb-t-human-names:canonicalize-name (s)
  ;; Remove spaces and quotation marks, and ignore capitalization.
  (let ((result ""))
    (while (string-match "[ ']+" s)
      (setq result (concat result (substring s 0 (match-beginning 0)))
            s (substring s (match-end 0))))
    (downcase (concat result s))))


;;
;; First names
;;

(defconst edb-t-human-names:nicknames
  '(("Abraham" "Abe")
    ("Charles" "Charlie")
    ("David" "Dave")
    ("Elizabeth" "Beth" "Liz")
    ("Gerald" "Gerry")
    ("James" "Jamey" "Jamie")
    ("Leonard" "Lenny")
    ("Michael" "Mike")
    ("Richard" "Dick" "Dicky" "Rick" "Ricky" "Rico")
    ("Robert" "Bob")
    ("Theodore" "Ted"))
  "List of nicknames which are not prefixes of the full name.
It is by no means comprehensive.")

(defun edb-t-human-names:same-first-name-p (n1 n2)
  (or (edb-t-human-names:nicknamep n1 n2)
      (edb-t-human-names:nicknamep n2 n1)))

(defun edb-t-human-names:order-first-names (n1 n2)
  (let ((m1 (db-string-split-first-word n1))
        (m2 (db-string-split-first-word n2)))
    (cond ((or
            (edb-t-human-names:same-first-name-p (car m1) (car m2))
            (edb-t-human-names:same-first-name-p n1 n2))
           0)
          ((string-lessp (car m1) (car m2))
           -1)
          (t
           1))))

(defun edb-t-human-names:nicknamep (nickname fullname)
  (or (string-match (concat "^" (regexp-quote nickname)) fullname)
      (member nickname (cdr (assoc (car (db-string-split-first-word fullname))
                                   edb-t-human-names:nicknames)))))


;;
;; Full names
;;

(defun edb-t-human-names:standardize-name (name)
  ;; remove leading honorifics
  (when (string-match (concat "^\\(\\(Dr\\|Mrs?\\|Prof\\|1?Lt\\)\\.?"
                              "\\|Miss\\|Ensign\\|Doctor\\) ") name)
    (setq name (substring name (match-end 0))))
  ;; remove trailing gubbish
  (setq name (substring name
                        0 (string-match ",? \\(PhD\\|MD\\|Esq\\.?\\|Esquire\\)$"
                                        name)))
  ;; add periods after initials
  (while (string-match "\\(^\\| \\)[A-Z]\\([ ,]\\|$\\)" name)
    (setq name (concat (substring name 0 (match-beginning 2))
                       "."
                       (substring name (match-beginning 2)))))
  name)

(defvar edb-t-human-names:jr-assoc-list ; preferred suffixes
  '(("3d" . "3rd")))

(defun edb-t-human-names:name->name-jr (name)
  ;; Return (suffixless-name suffix).
  (if (string-match ",? +\\(jr\\.?\\|sr\\.?\\|3r?d\\|[45]th\\|I+\\)$" name)
      (list (substring name 0 (match-beginning 0))
            (let ((jr (substring name (match-beginning 1))))
              (or (cdr (assoc (downcase jr) edb-t-human-names:jr-assoc-list))
                  jr)))
    (list name "")))

(defun edb-t-human-names:name->first-last-jr (fullname)
  ;; Return (fname lname jr).
  (let* ((jr (edb-t-human-names:name->name-jr
              (edb-t-human-names:standardize-name fullname)))
         (ls (db-string-split-last-word (car jr) "\\(La\\|de\\) [A-Z-]+")))
    (append ls (cdr jr))))


(provide 'edb-t-human-names)

;;; edb-t-human-names.el ends here
